/*
** --------------------------------------------------------------------------
**  MDA and HGC chipset driver
** --------------------------------------------------------------------------
**
**  Copyright (C) 1997      Jochen Hoenicke, 
**			    [jochen.hoenicke@informatik.uni-oldenburg.de]
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2, or (at your option)
**  any later version.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; see the file COPYING.  If not, write to
**  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
*/
#define	MAINTAINER	"Jochen Hoenicke, [jochen.hoenicke@informatik.uni-oldenburg.de]"

#undef 	CURSOR_HW
#define DISPLAY_NUMBER	1	/* change to 0 if you don't have a VGA */
#define HGC_TEXT_LEN	0x1000	/* On my card only 4096 Bytes are usable 
				 * in text mode */


/*
**     GGI includes
*/

#define DRIVER_VERSION	"v 0.0.9.3"

#include <linux/kernel.h>
#include <linux/ioport.h>
#include <ggi/module.h>
#include "Hercules/hgc.h"
#include "hgc_io.inc"

static struct kgi_chipset *chipinfo;
static struct kgi_chipset ci[2] = {
#define CI_MDA 0
        {
                "IBM",  		/* manufacturer         */
                "MDA",			/* model                */
                GM_TEXT16,              /* supported modes      */
                0,
                {720, 350},		/* maximum resolution   */
                0x1000,			/* 4k addressable RAM size */
                {16000000, 17000000}	/* DCLK limits            */
        }, 
#define CI_HGC 1
        {
                "Hercules",  		/* manufacturer         */
                "HGC",			/* model                */
                GM_TEXT16,              /* supported modes      */
                0,
                {720, 350},		/* maximum resolution   */
                0x8000,			/* 32k addressable RAM size */
                {16000000, 17000000}	/* DCLK limits            */
        }
};

static struct statetable	/* this table holds the default bits */
    default_state =		/* for text & graphics mode, will be */
{				/* filled in during mode_init        */
    0x09, 0x00,			/* MISC */
    {
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* CRT */
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff,
    }
};

struct statetable reset_state= {
        0x09, 0x00,
        {
                0x61, 0x50, 0x52, 0x0f, 0x19, 0x06, 0x19, 0x19,
                0x02, 0x0d, 0x0b, 0x0c, 0x00, 0x00, 0xff, 0xff
        } 
};

struct kgi_mode default_kgi_mode = {
	{ 1,				/* need one frame	*/
	  { 720, 350 }, { 720, 350 },	/* visible/virtual	*/
	  { 0, 0 },			/* size in mm		*/
	  GT_TEXT16,			/* mode			*/
	  { 9, 14 }			/* text size		*/
	},

	{ 720, 720, 738, 873, 873, 873,  1 },	/* horiz.timing	*/
	{ 350, 350, 364, 370, 370, 370,  0 },	/* vert. timing	*/
	{ 1, 1 },			/* no magnification	*/
	0,	/* !!! check */		/* special capabilities	*/
	16257000,			/* dot clock in Hz	*/
	{ 1, 1 },			/* clock rate to program*/
	{ 1, 1 },			/* transfer rate to dac	*/

	/* !!! init the reset (to 0 ?) */
}; 

extern int display_number;

static unsigned long fb_text_len;
static unsigned long fb_text;

static int detect_HGC(void)
{
        unsigned char old, New;
        int y;

        old = inbCRT(0x0f);
        outb_p(0xa5, CRTDm);
        New = inb_p(CRTDm);
        outb_p(old, CRTDm);
	if (New != 0xa5) {
                printk(KERN_WARNING "detect_HGC: No MDA/HGC detected.\n");
                return -1;
        }

        for (y=1; y; y--) {
                while (((old = inb(STATm)) & (STAT_HSYNC|STAT_VSYNC)) == 0) {}
                if (old & STAT_VSYNC) 
                        break;
                while (((old = inb(STATm)) & (STAT_HSYNC|STAT_VSYNC))
                       == STAT_HSYNC) {}
                if (old & STAT_VSYNC) 
                        break;
        }
        fb_text = 0xb0000;
        if (old & STAT_VSYNC) {
                chipinfo = &ci[CI_HGC];

                fb_text_len = HGC_TEXT_LEN;
                printk("HGC detected.\n");
        } else {
                chipinfo = &ci[CI_MDA];
                fb_text_len = 0x1000;
                printk("MDA detected.\n");
        }
        return EOK;
}

/*
**  GGI interface
*/

#include "hgc_text16.inc"

#ifdef DEBUG
static void printk_mode (struct kgi_mode *mode, char *msg)
{
	#define M(X)	mode->X
	#define R(X)	mode->request.X
	#define T(X)	M(X.width), M(X.blankstart), M(X.syncstart),\
			M(X.syncend), M(X.blankend), M(X.total), M(X.polarity)

	printk(KERN_DEBUG "%s\n",msg);
	printk(KERN_DEBUG "mode = {\n" 
               KERN_DEBUG "   { %i,  /* frames */\n"
               KERN_DEBUG "   { %i, %i},  /* visible */\n",
			  R(frames), R(visible.x),R(visible.y));
	printk(KERN_DEBUG "   { %i, %i},   /* virtual */\n",
                          R(virt.x),R(virt.y));
	printk(KERN_DEBUG "   { %i, %i},   /* size [mm] */\n",
                          R(size.x),R(size.y));
	printk(KERN_DEBUG "   0x%x,\n", R(graphtype));
	printk(KERN_DEBUG "   { %i, %i}},   /* textgrid */\n",
                          R(text.x),R(text.y));
	printk(KERN_DEBUG "   { %i, %i, %i, %i, %i, %i, %i}, /* x timing */\n",
                          T(x));
	printk(KERN_DEBUG "   { %i, %i, %i, %i, %i, %i, %i}, /* y timing */\n",
                          T(y));
	printk(KERN_DEBUG "   { %i, %i},   /* magnify */\n",
                          M(magnify.x), M(magnify.y));
	printk(KERN_DEBUG "   %x,          /* flags   */\n", M(flags));
	printk(KERN_DEBUG "   %i,          /* dclk    */\n", M(dclk));
	printk(KERN_DEBUG "   { %i, %i},   /* clk     */\n", 
                          M(clk.mul), M(clk.div));
	printk(KERN_DEBUG "   { %i, %i}    /* lclk    */\n"
	       KERN_DEBUG "}\n",
                          M(lclk.mul), M(lclk.div));

	#undef M
	#undef R
	#undef T
}
#endif /* #ifdef DEBUG */

static void chipset_setup_dpy(struct kgi_display *dpy,struct kgi_mode *mode)
{
    switch (mode->request.graphtype) {
    case GT_TEXT16:
	dpy->set_font=NULL;
	dpy->set_origin=text16_set_origin;
	dpy->set_split=text16_set_split;
	dpy->set_frame=text16_set_frame;
	dpy->show_cursor=text16_show_cursor;
	dpy->hide_cursor=text16_hide_cursor;
	dpy->undo_cursor=text16_undo_cursor;
	dpy->show_pointer=text16_show_pointer;
	dpy->hide_pointer=text16_hide_pointer;
	dpy->undo_pointer=text16_undo_pointer;
	dpy->fb.size.x = mode->request.virt.x / mode->request.text.x;
	dpy->fb.size.y = mode->request.virt.y / mode->request.text.y;
	dpy->fb.bpp    = MAP_16_BPP;
	dpy->fb.pbuf   = phys_to_virt(fb_text);
	dpy->fb.zbuf   = NULL;
	dpy->fb.ops    = &text16_color_io;
	dpy->driver_data=text16_data_realloc(dpy->driver_data);
	if (dpy->driver_data==NULL)
	    INTERNALERROR;
	break;
    default:
	INTERNALERROR;
    }
    dpy->mode=(*mode);
    dpy->fb.dpy=dpy;
}

static void chipset_setup_mode(struct kgi_mode *mode)
{
  switch (mode->request.graphtype) {
    case GT_TEXT16:
      mode->fb     = fb_text;
      mode->fb_size= fb_text_len;
      mode->fsize  = (mode->request.virt.x/mode->request.text.x)*
		     (mode->request.virt.y/mode->request.text.y)*2;
      mode->mem    = &text16_mono_mem;
      mode->io     = &text16_mono_io;
      break;
    default:
      INTERNALERROR;
  }
}

static int chipset_get_reset_mode(struct kgi_display *dpy)
{
   if (dpy->defmode==NULL)
     dpy->defmode=&default_kgi_mode.request;
   else
     (*dpy->defmode)=default_kgi_mode.request;

   dpy->mode = default_kgi_mode;

#ifdef DEBUG
   printk_mode(&(dpy->mode),"Default mode:");
#endif

   return EOK;
}
   
static int chipset_init(struct kgi_display *dpy)
{
    TRACE(printk("Hercules Chipset Driver " DRIVER_VERSION "\n"));
    display_number= DISPLAY_NUMBER;

    if (!detect_HGC()) {
	if (!check_region(HGC_IOBASE, HGC_IOSIZE))
	    request_region(HGC_IOBASE, HGC_IOSIZE, "kgi-chipset-hgc");
	else
	    printk(KERN_WARNING"chipset_init: failed to allocate IO region.\n");

	text16_data_init();
        return chipset_get_reset_mode(dpy);
    } 

    return -E(CHIPSET,FAILED);
}

static void chipset_reset(struct kgi_display *dpy)
{
        restore_state(&reset_state);
        if (dpy) {
                dpy->mode=default_kgi_mode;
                (*dpy->defmode)=default_kgi_mode.request;
        }
}

static void chipset_done(void)
{
        chipset_reset(NULL);
        release_region(HGC_IOBASE, HGC_IOSIZE);
}

static void adjust_timing(struct kgi_mode *mode)
{
        register int foo = (mode->request.graphtype == GT_TEXT16)
                ? mode->request.text.x : 16;

#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

        ADJUST(width);
        ADJUST(blankstart);
        ADJUST(syncstart);
        ADJUST(syncend);
        ADJUST(blankend);
        if (mode->x.total % foo)
                mode->x.total += foo - (mode->x.total % foo);

#undef ADJUST
}

static int chipset_check_mode(struct kgi_display *dpy,
                              struct kgi_mode *mode,
                              enum kgi_tm_cmd cmd)
{
    int bpp = 0;
    int textx,texty;
    textx = mode->request.text.x;
    texty = mode->request.text.y;

    switch (mode->request.graphtype) {
	case GT_1BIT:
	case GT_TEXT16:
	    break;		/* Not used in textmode */
	default:
	    TRACE(printk("%s: graphtype %i not supported.\n", __FILE__, \
			 mode->request.graphtype));
	    return -E(CHIPSET,FAILED);
    }

#define XVIRTUAL (mode->request.virt.x)
#define YVIRTUAL (mode->request.virt.y)
#define XVISIBLE (mode->request.visible.x)
#define YVISIBLE (mode->request.visible.y)

    switch (cmd) {
	case CMD_PROPOSE:
	    if (mode->request.graphtype == GT_TEXT16) {

		mode->magnify.x = mode->magnify.y = 1;
		if (textx != 9 || texty > 16 || texty < 1) {
		    TRACE(printk("%s: invalid char size (%i,%i).\n", 
				 __FILE__, textx,texty));
		    return -EINVAL;
		}

		mode->request.visible.x -=
		    (mode->request.visible.x % textx);
		mode->request.visible.y -=
		    (mode->request.visible.y % texty);

		if (XVIRTUAL==0 && YVIRTUAL==0)
		  return -EINVAL;

		if (XVIRTUAL==0) 
		  XVIRTUAL = XVISIBLE;

		if (YVIRTUAL==0) 
		  YVIRTUAL = fb_text_len / ((XVIRTUAL/textx) * 2) * texty;

		if ((2 * (XVIRTUAL / textx)
		     * (YVIRTUAL / texty)) > fb_text_len) {
		    TRACE(printk("%s: not enough text memory.\n", __FILE__));
		    return -ENOMEM;
		}
	    } else {		/* mode->request.graphtype == GT_TEXT16 */
                    return -EINVAL;
	    }

	    mode->x.width = mode->request.visible.x;
	    mode->y.width = mode->request.visible.y;

	    adjust_timing(mode);
	    return EOK;

	case CMD_LOWER:

	    if (mode->dclk < chipinfo->dclk.min) {
		TRACE(printk("%s: DCLK too low (%i).\n", __FILE__, mode->dclk));
		return -E(CHIPSET,FAILED);
	    }
	    mode->dclk = (mode->dclk > chipinfo->dclk.max)
		? chipinfo->dclk.max : mode->dclk;

	    adjust_timing(mode);
	    return EOK;

	case CMD_RAISE:

	    if (mode->dclk > chipinfo->dclk.max) {
		TRACE(printk("%s: DCLK too high (%i).\n", \
			     __FILE__, mode->dclk));
		return -E(CHIPSET,FAILED);
	    }
	    mode->dclk = (mode->dclk < chipinfo->dclk.min)
		? chipinfo->dclk.min : mode->dclk;

	    adjust_timing(mode);
	    return EOK;

	case CMD_CHECK:
    	    if (mode->lclk.mul != 1) {
                    TRACE(printk("%s: invalid lclk multiplier %d.\n",
                                 __FILE__,mode->lclk.mul));
                    return -E(CHIPSET,FAILED);
	    }
	    if (mode->lclk.div != 1) {
	    	    TRACE(printk("%s: invalid CLK to LCLK divider %d.\n",
                                 __FILE__,mode->lclk.div));
                    return -E(CHIPSET,FAILED);
	    }
            if (XVIRTUAL != XVISIBLE) {
                    TRACE(printk("%s: virt.x != visible.x.\n",
                                 __FILE__));
                    return -E(CHIPSET,FAILED);
            }
	    if (mode->request.graphtype == GT_TEXT16) {
		mode->flags |= FL_CHIP_COLORTEXT;
		if ((mode->magnify.x != 1) 
                    ||(mode->magnify.y != 1)	/*driver */
		    ||(textx != 9)	/* hardware */
		    ||((2 * (XVIRTUAL / textx)
                          * (YVIRTUAL / texty)) > fb_text_len)
		    ||(mode->dclk > chipinfo->dclk.max)		/* hardware */
		    ||(mode->dclk < chipinfo->dclk.min)) {	/* hardware */
                        TRACE(printk("%s: hw check failed\n", __FILE__));
                        return -E(CHIPSET,FAILED);
		}
	    } else {
		if (((XVIRTUAL * YVIRTUAL * bpp / 8) > chipinfo->ramsize)
		    || (mode->magnify.x != 1) || (mode->magnify.y != 1)
		    || (mode->clk.mul != 1) || (mode->clk.div != 1)
		    || (mode->dclk < chipinfo->dclk.min)
		    || (chipinfo->dclk.max < mode->dclk)
		    || (XVIRTUAL != 720))
		    return -E(CHIPSET,FAILED);
	    }

	    chipset_setup_mode(mode);
	    return EOK;
	    
	default:
	    INTERNALERROR;
	    return -E(CHIPSET,FAILED);
    }

#undef XVIRTUAL
#undef YVIRTUAL
#undef XVISIBLE
#undef YVISIBLE
}

static void chipset_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
    struct statetable setup = default_state;
    int DCLKdiv, cheight;

#if 0
    if (mode->request.graphtype == GT_1BIT) {
	DCLKdiv = 16;
        cheight = 4;
        setup.CTRL |= CTRL_ENB_GRAPHICS;
	set
    } else
#endif
    {	
        ASSERT (mode->request.graphtype == GT_TEXT16);

	DCLKdiv = 9;
        cheight = mode->request.text.y;
#ifdef CURSOR_HW
        setup.CRT[0x0a] = cheight-3;
        setup.CRT[0x0b] = cheight-2;
#endif
    }

    setup.CRT[0x08] = CR08_NON_INTERLACED;
    setup.CRT[0x09] = cheight - 1;

    /*
    **        setup horizontal timing
    */

    setup.CRT[0x00] = mode->x.total / DCLKdiv;
    setup.CRT[0x01] = mode->x.width / DCLKdiv;
    setup.CRT[0x02] = mode->x.syncstart / DCLKdiv;
    setup.CRT[0x03] = mode->x.syncend / DCLKdiv - setup.CRT[0x02];

    /*
    **        setup vertical timing
    */

    setup.CRT[0x04] = mode->y.total / cheight - 1;
    setup.CRT[0x05] = mode->y.total % cheight;
    setup.CRT[0x06] = mode->y.width / cheight;
    setup.CRT[0x07] = mode->y.syncstart / cheight - 1;

    /*
    **        Ok. We'll do the programming now.
    */
#ifdef DEBUG
    printk_mode(mode,"Set mode:");
#endif /* DEBUG */

    kgim_monitor_set_mode(dpy,mode);
    kgim_ramdac_set_mode(dpy,mode);
    kgim_clock_set_mode(dpy,mode);
    restore_state(&setup);
    chipset_setup_dpy(dpy,mode);
}

#include "kgi_chipset.inc"
