/* --------------------------------------------------------------------------
**	Matrox Millennium (mga2064w) chipset driver
** --------------------------------------------------------------------------
**
**	Copyright (C)	1997		Jonas Borgstrom
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
**	This code was derived from the following source of information:
** 
**	Matrox mga2064w reference internet edition
**	XFree86 3.3.1 mga driver
**
**	$Log: mga2x64_main.c,v $
**	Revision 1.9  1998/02/14 13:54:11  jonas_b
**	The mga2x64 driver now works again.
**	
**	Revision 1.8  1997/12/29 18:09:06  jonas_b
**	The driver is now MMIO only (no vga regs/mem).
**	
**	Revision 1.7  1997/12/28 12:24:11  jonas_b
**	Applied Matthias patch and fixed some bugs.
**	
**	Revision 1.6  1997/12/23 20:43:06  jonas_b
**	Added Matrox Millennium II AGP support (untested).
**	
**	Revision 1.5  1997/12/20 17:41:48  grimrath
**	Mystique driver with acceleration for the first time
**	
**	Revision 1.3  1997/12/14 21:55:44  becka
**	uint/sint namespace fixed ... please test ...
**	
**	Revision 1.2  1997/11/23 14:16:04  jonas_b
**	Added 15 and 24 bpp support.
**	
**	Revision 1.1  1997/11/22 11:25:55  jonas_b
**	Added a new driver which uses the "new linking scheme"
**	
**	Revision 1.1  1997/10/13 19:41:18  seeger
**	Initial checkin of Matrox Millenium driver by Jonas Borgstrom
**	
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Jonas_Borgstrom
#define DRIVER_NAME	"Matrox mga2x64 chipset driver"
#define DRIVER_REV	"$Revision: 1.9 $"

#define MGA_CHIPSET_DRIVER /* Use our private struct definitions */

#include <ggi/module.h>

#include <Matrox/mga_common_regs.h>  /* Register definitions */
#include <Matrox/mga_common_mmio.h>
#include <Matrox/mga_common_text16.h>
#include <Matrox/mga_common.h>
#include <Matrox/mga2x64_main.h>

#define	HARDWARE_TEXT_CURSOR 1	/* Hardware cursor */

const char chipset_name_init[] = DRIVER_NAME " rev " DRIVER_REV " loaded";

struct kgi_chipset chipinfo =
{
	"Matrox Graphics",	/* Manufacturer		*/
	"Matrox mga2x64 (I/II)",/* Model		*/
	GM_TEXT16 | GM_1BIT |	/* Available modes	*/
	GM_4BIT | GM_8BIT | 
	GM_15BIT | GM_16BIT |
	GM_24BIT | GM_32BIT,
	0,			/* Flags		*/
	{ 1600, 1200 },		/* Maximum resolution	*/
	0,		        /* to be set!		*/
	{ 0, 220000000 }	/* DCLK limits		*/
};

/*	This table holds the default values for the VGA registers. A copy
**	of it will be filled in as appropriate during mode programming in
**	kgim_chipset_set_mode().
*/

const struct mem_region mga_aperture =
{
	MEM_NULL,
	0,
	0x4000,
	-1,
	"mga2x64 Aperture"
};

const struct mem_region mga_framebuffer =
{
	MEM_NULL,
	0,
	0x800000,
	-1,
	"mga2x64 Frame Buffer"
};

const struct kgi_mmio_region graphic_mmio_linear_fb =
{
	NULL,			/* Next			 */
	"mga2x64 framebuffer",	/* Name		*/
	MMIO_FRAMEBUFFER,	/* Type (Linear)	*/
	MMIO_RDWR | MMIO_SHARED,/* Protection map	*/
     
	0x0,			/* Total size of area	*/
	0x0,			/* Physical size	*/
	0x0,			/* bus address of the windows */
	0,			/* Logical adress	*/
     
	NULL,		        /* Set offset function	*/
	NULL		        /* private		*/
};

/*      This table holds the default values for the VGA registers. A copy
**      of it will be filled in as appropriate during mode programming in
**      kgim_chipset_set_mode().
*/

static struct statetable default_state =
{
	0x00,						/* MISC	*/
	{
		0x03, 0x00, 0x00, 0x00, 0x00			/* SEQ	*/
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* CRT	*/
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x1x	*/
		0x00
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* GRC	*/
		0xFF
	},
	{
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC	*/
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
		0x00, 0x00, 0x0F, 0x00, 0x00
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00 /* ExtVga */
	}
};

static void restore_state(struct kgi_display *dpy, struct statetable *state)
{
	GET_PRIVATE;
	uint32 tmp;

	mem_out16(0x0004, MMIO_CRTXI);
	MISC_out8(dpy, state->MISC);

	/* Enable Interleave mode (64 bit ramdac) if needed */
	pcibios_read_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN, PCI_STATUS_UDF, &tmp);
	pcibios_write_config_dword(MGA_PCI_BUS, MGA_PCI_DEV_FN, PCI_STATUS_UDF, ( tmp & ~0x1000) | (MGA_INTERLEAVE << 12));
	restore_SEQ(dpy, state->SEQ);
	restore_CRT(dpy, state->CRT);
	restore_GRC(dpy, state->GRC);
	restore_ATC(dpy, state->ATC);
	restore_ExtVga(dpy, state->ExtVga);
}

/*
** Int handler.
*/
void mga_interrupt_handler(int irq, void *d, struct pt_regs *regs)
{
	/* Place here what's necessary */
}

/*
**	DAC output functions for the tvp 3026 ramdac
*/

#define	DAC_REGISTERS	0x20

void DAC_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	GET_PRIVATE;
	ASSERT(reg < DAC_REGISTERS);

	mem_out8(val, MGA_RAMDAC + reg);
	DEBUG3("DAC mem %.2x <- %.2x", reg, val);
}

uint8 DAC_in8(struct kgi_display *dpy, uint8 reg)
{
	GET_PRIVATE;
	register uint8 val;

	ASSERT(reg < DAC_REGISTERS);
	val = mem_in8(MGA_RAMDAC + reg);
	DEBUG3("DAC mem %.2x -> %.2x", val, reg);
	return val;
}

/*
**	KGI module interface
*/

/* Mode specific frame buffers */
/* Mode specific frame buffers */
void adjust_timing(struct kgi_mode *mode)
{
        int foo = (mode->request.graphtype == GT_TEXT16) ? 
	  mode->request.dpp.x : 8;

        #define ADJUST(X) mode->x.X -= (mode->x.X % foo)

        ADJUST(width);
        ADJUST(blankstart);
        ADJUST(syncstart);
        ADJUST(syncend);
        ADJUST(blankend);

        if (mode->x.total % foo) {
                mode->x.total += foo - (mode->x.total % foo);
        }

        if (((mode->x.syncend / foo) & 0x3F) ==
	    ((mode->x.syncstart / foo) & 0x3F)) {
                mode->x.syncend -= foo;
        }

        if (((mode->x.blankend / foo) & 0x7F) ==
	    ((mode->x.blankstart / foo) & 0x7F)) {
                mode->x.blankend -= foo;
        }

        if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F)) {
                mode->y.syncend--;
        }

        if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF)) {
                mode->y.blankend--;
        }

        #undef ADJUST
}

/*void adjust_timing(struct kgi_mode *mode)
{
        register int foo = (mode->request.graphtype == GT_TEXT16)
                ? mode->request.text.x : 8;
   
        #define ADJUST(X) mode->x.X -= (mode->x.X % foo)

        ADJUST(width);
        ADJUST(blankstart);
        ADJUST(syncstart);
        ADJUST(syncend);
        ADJUST(blankend);

        if (mode->x.total % foo) {
                mode->x.total += foo - (mode->x.total % foo);
        }
   
        if (((mode->x.syncend / foo) & 0x3F) ==
	    ((mode->x.syncstart / foo) & 0x3F)) {
                mode->x.syncend -= foo;
        }

        if (((mode->x.blankend / foo) & 0x7F) ==
	    ((mode->x.blankstart / foo) & 0x7F)) {
                mode->x.blankend -= foo;
        }
   
        if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F)) {
                mode->y.syncend--;
        }

        if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF)) {
                mode->y.blankend--;
        }

        #undef ADJUST
}*/

int adjust_virtual(struct kgi_mode *mode)
{
	/* !!! Place here what's necessary */
	
	return 0;
}
/*
 * Init Chipset
 */
int kgim_chipset_init(struct kgi_card *card, struct kgi_display *dpy)
{
	int x;
	x = mga_chipset_init(card, dpy, sizeof(struct chipset_private),
			     PCI_DEVICE_ID_MATROX_MIL);
	if(!x) return x;
	else x = mga_chipset_init(card, dpy, sizeof(struct chipset_private),
			     PCI_DEVICE_ID_MATROX_MIL_2);
	if(!x) return x;
	else x = mga_chipset_init(card, dpy, sizeof(struct chipset_private),
			     PCI_DEVICE_ID_MATROX_MIL_2_AGP);
	return x;
}

void kgim_chipset_done(struct kgi_display *dpy)
{
	mga_chipset_done(dpy);

        NOTICE("%s", DRIVER_NAME " rev " DRIVER_REV " removed");
}

static void mga_set_origin(struct kgi_display *dpy, uint x, uint y)
{
	GET_PRIVATE;
	int offset, offset_modulo, ydstorg_modulo;
	int bppshift, base, ydstorg = 0, bpp;

	switch(dpy->mode.request.graphtype) {

	case GT_8BIT:
		bppshift = (MGA_INTERLEAVE) ? 0 : 1;
		bpp = 1;
		break;

	case GT_15BIT:
	case GT_16BIT:
		bppshift = (MGA_INTERLEAVE) ? 1 : 2;
		bpp = 2;
		break;

	case GT_24BIT:
		bppshift = (MGA_INTERLEAVE) ? 0 : 1;
		bpp = 3;
		break;

	case GT_32BIT:
		bppshift = (MGA_INTERLEAVE) ? 2 : 3;
		bpp = 4;
		break;

	default:
		INTERNAL_ERROR;
		return;
	}
   
	if (dpy->mode.request.virt.x * dpy->mode.request.virt.y * bpp
		> 4*1024*1024) {

		offset = (4*1024*1024) % (dpy->mode.request.visible.x * bpp);
		offset_modulo = 4;
		ydstorg_modulo = 64;
		if (dpy->mode.request.graphtype == GT_24BIT)
		    offset_modulo *= 3;
		if (MGA_INTERLEAVE)
		{
		    offset_modulo <<= 1;
		    ydstorg_modulo <<= 1;
		}
		ydstorg = offset / bpp;
		while ((offset % offset_modulo) != 0 ||
		       (ydstorg % ydstorg_modulo) != 0)
		{
		    offset++;
		    ydstorg = offset / bpp;
		}
	}
	base = (y * dpy->mode.request.virt.x + x + ydstorg) >>
			(3 - bppshift);
   
	if (dpy->mode.request.graphtype == GT_24BIT)
		base *= 3;

	CRT_out8(dpy, base, CRT_STARTADDR_L);
	CRT_out8(dpy, base >> 8, CRT_STARTADDR_H);
	ECRT_out8(dpy, (ECRT_in8(dpy, CRTX_EXT0) & ~CRTX_STARTADD_MASK) |
		  ((base >> 16) & CRTX_STARTADD_MASK), CRTX_EXT0);
}
/*
 * Set a Text/Graphic mode
 */
void kgim_chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	IS_CRITICAL;
   
	GET_PRIVATE;
	struct statetable setup = default_state;
	register int foo = 0;
	int DCLKdiv = 8;
	int textx,texty, bppshift;
	int wd, hd, hs, he, ht, vd, vs, ve, vt;

	DEBUG2("Setting mode... (%dx%d %d)\n", mode->request.virt.x,
		mode->request.virt.y, mode->request.graphtype);

/*	textx = mode->request.text.x;
	texty = mode->request.text.y;*/
        textx = mode->request.dpp.x;
        texty = mode->request.dpp.y;
   
	/*	Set up bits per pixel and logical screen width
	*/
	switch (mode->request.graphtype) {

	case GT_TEXT16:
	case GT_8BIT:
	case GT_15BIT:
	case GT_16BIT:
	case GT_24BIT:
	case GT_32BIT:
		break;

	default:
		INTERNAL_ERROR;
		return;
	}
   
	MGA_INTERLEAVE = (dpy->card->chipset->ramsize > (1024 * 2 * 1024)) ? 1: 0;

	if (mode->request.graphtype == GT_TEXT16)
        {

	        MGA_INTERLEAVE = 0;
	
		foo = 2 * mode->request.virt.x;
/*		foo = 2 * mode->request.virt.x / textx;*/
		foo /= 4;	/* Due to the addressing mode used */
		setup.CRT[CRT_LOGICALWIDTH] |= foo & 0xFF;
		setup.CRT[0x17] = CR17_CGA_ADDR_WRAP | CR17_ENABLE_SYNC |
			CR17_CGA_BANKING | CR17_HGC_BANKING;
      
		setup.CRT[0x14] |= CR14_UNDERLINE_MASK;

		setup.SEQ[0x02] = SR02_PLANE0 | SR02_PLANE1;
      
		setup.GRC[0x05] |= GR5_ODD_EVEN;
		setup.GRC[0x06] |= GR6_CHAIN_OE;

		setup.GRC[0x06] |= GR6_MAP_A0_64;

		switch ((DCLKdiv = textx)) {

		case 8:
			setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
			break;

		case 9:
			setup.ATC[0x10] |= AR10_9DOT_CHARS;
			break;

		default:
			INTERNAL_ERROR;
			return;
		}

		setup.CRT[0x09] |= texty - 1;
		setup.ATC[0x13] = (textx == 9) ? 0x08 : 0x00;
	}
	/*
	**	Set up horizontal timing
	*/

	/*	horizontal display end
	*/
	foo = (mode->x.width / DCLKdiv) - 1;
	setup.CRT[0x01] |= foo & 0xFF;

	/*	horizontal blank start
	*/
	foo = (mode->x.blankstart / DCLKdiv);
	setup.CRT[0x02] |= foo & 0xFF;

	/*	horizontal sync start
	*/
	foo = (mode->x.syncstart / DCLKdiv);
	setup.CRT[0x04] |= foo & 0xFF;
	if (mode->request.graphtype == GT_TEXT16)
        {
		/*	horizontal sync end
		*/
		foo = (mode->x.syncend / DCLKdiv);
		setup.CRT[0x05] |= foo & CR05_HSE_MASK;

		/*	horizontal blank end
		*/
		foo = (mode->x.blankend / DCLKdiv);
		setup.CRT[0x03] |= foo & CR03_HBE_MASK;
			setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;

		/*	horizontal total
		*/
		foo = (mode->x.total / DCLKdiv) - 5;
		setup.CRT[0x00] |= foo & 0xFF;
	}
   
	/*
	**	set up vertical timing
	*/

	/*	vertical display end
	*/
	foo = mode->y.width - 1;
	setup.CRT[0x12] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9 : 0;

	/*	vertical blank start
	*/
	foo = mode->y.blankstart - 1;
	setup.CRT[0x15] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8 : 0;
	setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9 : 0;

	/*	vertical sync start
	*/
	foo = mode->y.syncstart;
	setup.CRT[0x10] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9 : 0;

	/*	vertical sync end
	*/
	foo = mode->y.syncend;
	setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;

	/*	vertical blank end
	*/
	foo = mode->y.blankend + 1;
	setup.CRT[0x16] |= foo & 0xFF;

	/*	vertical total
	*/
	foo = mode->y.total - 2;
	setup.CRT[0x06] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8 : 0;
	setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9 : 0;

	setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
	setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;
	setup.MISC |= MISC_CLOCK_1X;

	/*	set splitline to maximum value
	*/
	foo = 0x3ff;
	setup.CRT[0x18] |= foo & 0xFF;
	setup.CRT[0x07] |= (foo & 0x0100) ? CR07_LINECOMPARE_B8 : 0;
	setup.CRT[0x09] |= (foo & 0x0200) ? CR09_LINECOMPARE_B9 : 0;
   
	if(mode->request.graphtype != GT_TEXT16) {

		setup.SEQ[0x00] = 0x00;
                setup.SEQ[0x01] = 0x01; /*0x09*/
                setup.SEQ[0x02] = 0x0F;
                setup.SEQ[0x03] = 0x00;
                setup.SEQ[0x04] = 0x0E;

                setup.GRC[0x00] = 0x00;
                setup.GRC[0x01] = 0x00;
                setup.GRC[0x02] = 0x00;
                setup.GRC[0x03] = 0x00;
                setup.GRC[0x04] = 0x00;
                setup.GRC[0x05] = 0x40;
                setup.GRC[0x06] = 0x05;
                setup.GRC[0x07] = 0x0F;
                setup.GRC[0x08] = 0xFF;

                setup.ATC[0x10] = 0x41;
		setup.ATC[0x11] = 0xFF;
                setup.ATC[0x12] = 0x0F;
                setup.ATC[0x13] = 0x00;
                setup.ATC[0x14] = 0x00;

                setup.CRT[0x17] = 0xC3;
                setup.CRT[0x18] = 0xFF;

		switch(mode->request.graphtype) {

		case GT_8BIT:
			bppshift = (MGA_INTERLEAVE) ? 0 : 1;
			break;

		case GT_15BIT:
		case GT_16BIT:
			bppshift = (MGA_INTERLEAVE) ? 1 : 2;
			break;

		case GT_24BIT:
			bppshift = (MGA_INTERLEAVE) ? 0 : 1;
			break;

		case GT_32BIT:
			bppshift = (MGA_INTERLEAVE) ? 2 : 3;
			break;

		default:
			INTERNAL_ERROR;
			return;
		}

		wd = (mode->request.virt.x *
			((mode->request.graphtype == GT_24BIT) ? 3 : 1)) >>
			(4 - bppshift);

		hd = (mode->x.width >> 3) - 1;
		hs = (mode->x.syncstart >> 3) - 1;
		he = (mode->x.syncend >> 3) - 1;
		ht = (mode->x.total >> 3) - 1;

		vd = mode->y.width - 1;
		vs = (mode->y.syncstart - 1);
		ve = mode->y.syncend - 1;
		vt = mode->y.total - 2;

		setup.ExtVga[5] = 0;
	
		setup.ExtVga[0]	= (wd & 0x300) >> 4;
		setup.ExtVga[1]	= (((ht - 4) & 0x100) >> 8) |
			((hd & 0x100) >> 7) | ((hs & 0x100) >> 6) | (ht & 0x40);

		setup.ExtVga[2]	= ((vt & 0x400) >> 10) | ((vt & 0x800) >> 10) |
				((vd & 0x400) >> 8) | ((vd & 0x400) >> 7) |
				((vd & 0x800) >> 7) | ((vs & 0x400) >> 5) |
				((vs & 0x800) >> 5);

		if (mode->request.graphtype == GT_24BIT)
			setup.ExtVga[3]	= (((1 << bppshift) * 3) - 1) | 0x80;

		else
			setup.ExtVga[3]	= ((1 << bppshift) - 1) | 0x80;


		setup.ExtVga[3] |= ((MGA_CP.chipinfo.ramsize >> 10) == 8192
				    ? 0x10 : (MGA_CP.chipinfo.ramsize >> 10) == 2048 
				    ? 0x08 : 0x00);
		setup.ExtVga[3] |= ((chipinfo.ramsize >> 10) == 8192
			? 0x10 : (chipinfo.ramsize >> 10) == 2048 ? 0x08 : 0x00);

		setup.ExtVga[4]	= 0;

		setup.CRT[0]	= ht - 4;
                setup.CRT[2]	= hd;
		setup.CRT[3]	= (ht & 0x1F) | 0x80;
                setup.CRT[4]	= hs;
		setup.CRT[5]	= ((ht & 0x20) << 2) | (he & 0x1F);
		setup.CRT[7]	= ((vt & 0x100) >> 8 ) | ((vd & 0x100) >> 7 ) |
			((vs & 0x100) >> 6 ) | ((vd & 0x100) >> 5 ) | 0x10 |
			((vt & 0x200) >> 4 ) | ((vd & 0x200) >> 3 ) |
			((vs & 0x200) >> 2 );

		setup.CRT[9]	= ((vd & 0x200) >> 4) | 0x40;
		setup.CRT[16]	= vs & 0xFF;
		setup.CRT[17]	= (ve & 0x0F) | 0x20;
		setup.CRT[18]	= vd & 0xFF;
		setup.CRT[19]	= wd & 0xFF;
		setup.CRT[21]	= vd & 0xFF;
		setup.CRT[22]	= (vt + 1) & 0xFF;

		if (mode->magnify.y == 2) setup.CRT[9] |= 0x80;
	}
   
	/*	Ok.  We'll do the actual hardware programming now.
	*/

	ENTER_CRITICAL;
	kgim_monitor_set_mode(dpy, mode);
	kgim_ramdac_set_mode(dpy, mode);
        kgim_clock_set_mode(dpy, mode);
        kgim_graphics_set_mode(dpy, mode);
	restore_state(dpy, &setup);
	LEAVE_CRITICAL;

	memcpy(&(dpy->mode), mode, sizeof(dpy->mode));

	NOTICE2("setting up display...");

	private->cp.split    = -1;
	private->cp.org_xdot = private->cp.org_ydot = 
		private->cp.org_offs = -1;

	if (mode->request.graphtype == GT_TEXT16) {

		dpy->set_font   = mga_text_set_font;
		dpy->set_split  = mga_set_split;
		dpy->set_frame  = mga_set_frame;
		dpy->set_origin = mga_text_set_origin;

		/*      provide hardware cursor in text mode
		*/
		if (HARDWARE_TEXT_CURSOR) {

			dpy->show_cursor = mga_show_hw_text_cursor;
			dpy->hide_cursor = mga_hide_hw_text_cursor;
			dpy->undo_cursor = NULL;

		} else {

			dpy->show_cursor = mga_show_sw_text_cursor;
			dpy->hide_cursor = mga_hide_sw_text_cursor;
			dpy->undo_cursor = mga_undo_sw_text_cursor;
		}

		/*      provide software cursor too in case DAC didn't already.
		*/
		if (! dpy->show_pointer) {

			dpy->show_pointer = mga_show_sw_text_pointer;
			dpy->hide_pointer = mga_hide_sw_text_pointer;
			dpy->undo_pointer = mga_undo_sw_text_pointer;
		}

		dpy->fb.size.x = mode->request.virt.x;
		dpy->fb.size.y = mode->request.virt.y;
		dpy->fb.bpp    = MAP_16_BPP;
		dpy->fb.pbuf   = MGA_FRAMEBUFFER.baseptr;
		dpy->fb.zbuf   = NULL;
		dpy->fb.ops    = &mgatext_color_io;

		private->cp.fb_size = dpy->fb.size.x * dpy->fb.size.y;

	} else {

		dpy->fb.size.x	= mode->request.virt.x;
		dpy->fb.size.y	= mode->request.virt.y;
		dpy->fb.pbuf	= NULL;
		dpy->fb.zbuf	= NULL;
		dpy->fb.ops	= NULL;

		private->cp.fb_size = dpy->fb.size.x * dpy->fb.size.y;
		
		dpy->show_pointer	= NULL;
		dpy->hide_pointer	= NULL;
		dpy->undo_pointer	= NULL;

		dpy->show_cursor	= NULL;
		dpy->hide_cursor	= NULL;
		dpy->undo_cursor	= NULL;

		dpy->set_font		= NULL;
		dpy->set_split		= mga_set_split;
		dpy->set_frame		= mga_set_frame;
		dpy->set_origin		= mga_set_origin;

	}
	NOTICE1("Mode set complete");
}

int kgim_chipset_command(struct kgi_graphic *state, ggi_uint cmd, ggi_uint iobuf)
{
	ggi_coord  *coord;

	switch (cmd) {
		case CHIP_SETVISFRAME:
			coord = (ggi_coord *)(state->io_buf);

       			if (coord->x < 0 || 
			   coord->x > state->dev.mode.virt.x)
				return -E(CHIPSET,BOGUS);

			if (coord->y < 0 || 
			    coord->y > state->dev.mode.virt.y)
				return -E(CHIPSET,BOGUS);

			kgi_set_origin(&(state->dev), 0, coord->x, coord->y);
			return 0;
		default:
			break;
	}
	return -E(CHIPSET, UNKNOWN);
}

