/* --------------------------------------------------------------------------
 *	Cyrix MediaGX graphics (acceleration) driver
 * --------------------------------------------------------------------------
 *
 *	Copyright (C) 1997-1998		Emmanuel Marty
 *	Copyright (C) 1997-1998		Cyrix Corporation
 *	Copyright (C) 1997-1998		The GGI project
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 * 
 *	You should have received a copy of the GNU General Public License
 *	along with this program; see the file COPYING.  If not, write to
 *	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * --------------------------------------------------------------------------
 *
 *	This code was derived from the following sources of information :
 *
 *	Cyrix MediaGX confidential databooks
 *	Cyrix MediaGX gxvideo.c example sourcecode
 *	Help and code from Sengan Baring-Gould of Cyrix Corporation
 *
 *	Many thanks to the Cyrix Corporation for providing me with
 *	documentation and official support about the GX chipset family.
 *
 * --------------------------------------------------------------------------
 *
 *	$Log: gx.c,v $
 *	Revision 1.2  1998/03/20 00:11:00  emarty
 *	removed external calls for reading/setting accel registers
 *	
 *	Revision 1.1  1998/01/29 09:44:22  emarty
 *	initial checkin
 *	
 */

#include <ggi/maintainers.h>
#define MAINTAINER	Emmanuel_Marty
#define DRIVER_NAME	"Cyrix MediaGX graphics driver"
#define DRIVER_REV	"$Revision: 1.2 $"
#define DEBUG_LEVEL	0

#include <ggi/module.h>
#include <Cyrix/gx.h>
#include "suggest.inc"

/*	Acceleration functions		*/

static int GX_drawbox (struct kgi_graphic *gr, void *iobuf) {
	IS_CRITICAL;
	struct box *boxp = (struct box *)iobuf;
	uint16 gt = gr->dev.mode.graphtype, color;
	mem_vaddr gxbase = ((struct chipset_private *)
		KGIM_PRIVATE (gr->dev.dpy, chipset))->gx_control.baseptr;
	uint32 lockval;

	if ((gt == GT_4BIT) || (gt == GT_TEXT16))
		return -E(GRAPHICS,NOSUP_MODE_MMAP);

	if (boxp->w && boxp->h) {
		color = gr->gc->fg_color;

		if (gt == GT_8BIT)
			color |= (color << 8);

		while (READ_GX32 (gxbase, GP_BLIT_STATUS) & BS_BLIT_PENDING);

		ENTER_CRITICAL;
		lockval = READ_GX32 (gxbase, DC_UNLOCK);
		WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

		WRITE_GX32 (gxbase, GP_DST_XCOOR, boxp->x | (boxp->y << 16));
		WRITE_GX32 (gxbase, GP_WIDTH, boxp->w | (boxp->h << 16));
		WRITE_GX16 (gxbase, GP_PAT_COLOR_0, color);
		WRITE_GX16 (gxbase, GP_RASTER_MODE, 0xF0);
		WRITE_GX16 (gxbase, GP_BLIT_MODE, BM_WRITE_FB);

		WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
		LEAVE_CRITICAL;

		/*	necessary before an LFB access, until GGI can fix it */
		while (READ_GX32 (gxbase, GP_BLIT_STATUS) & BS_BLIT_BUSY);
	}

	return 0;
}

/*	KGI module interface	*/

int kgim_graphics_init (struct kgi_card *card, struct kgi_display *dpy) {
	NOTICE (DRIVER_NAME " rev " DRIVER_REV);

	return EOK;
}

void kgim_graphics_done (struct kgi_display *dpy) {
	NOTICE (DRIVER_NAME " rev " DRIVER_REV " removed");
}

int kgim_graphics_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			      enum kgi_tm_cmd cmd) {
	return EOK;
}

void kgim_graphics_set_mode (struct kgi_display *dpy, struct kgi_mode *mode) {
}

int kgim_graphics_command (struct kgi_graphic *gr, ggi_uint cmd, ggi_uint arg) {
	ggi_suggest *suggest;
	const char *name;

	if (!(gr->dev.state & KGI_MAPPED_ON_DISPLAY)) {
		/*	Don't accelerate unmapped screen	*/
		return -E(GRAPHICS,NOSUP_MODE_MMAP);
	}

	switch (cmd) {

		case GRAPHICS_GETSUGGEST:
			suggest = (ggi_suggest *) (gr->io_buf);
			gr->io_res = gr->io_buf;

			if ((suggest->id < 0) || (suggest->id > 4)) {
				return -E(GRAPHICS,BOGUS);
			}

			if (!(name = sugg_default[gr->dev.mode.graphtype].
			       text[suggest->id])) {

				return -E(GRAPHICS,BOGUS);
			}

			strcpy (suggest->name, name);
			suggest->args[0] = 0;
			suggest->id++;

			if ( (suggest->id >= 4) ||
			     (sugg_default[gr->dev.mode.graphtype].
			      text[suggest->id] == NULL) )

				suggest->id = 0;

			return 0;

		case GRAPHICS_DRAWBOX:
			return (GX_drawbox (gr, gr->io_buf));

		case GRAPHICS_FILLSCREEN:
			return -E(GRAPHICS,NOSUP_ALWAYS_LOWER);

		default:
			return -E(GRAPHICS,NOSUP_ALWAYS_MMAP);

	}
}
