/* ---------------------------------------------------------------------------
 * PVGA1A framebuffer support
 * ---------------------------------------------------------------------------
 *
 * Copyright (C) 1998, Brian S. Julin bri@calyx.com
 * Copyright (C) 1997, The Linux-GGI Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * -----------------------------------------------------------------------
 *
 * ---------------------------------------------------------------------------
 */

#define VGA_WINDOW_SIZE (64 * 1024)

#include <ggi/maintainers.h>
#define	MAINTAINER	Brian_S_Julin
#define DRIVER_NAME	"PVGA1A graphics driver"
#define	DRIVER_REV	"$Revision: 1.1 $"

#define DEBUG_LEVEL	0

/* KGI includes */
#include <asm/io.h>

#include <ggi/module.h>
#include <ggi/ggi_commands.h>

#include "pvga1a_suggest.c"
#include "pvga1a_mmio.c"

#include "WD/pvga1a_io.h"

/* KGI module interface */

int kgim_graphics_init(struct kgi_card *card, struct kgi_display *dpy)
{
	/* FIXME: We should really do something here! */
	NOTICE1("PVGA1A Graphics Driver loaded");
	return EOK;
}

void kgim_graphics_done(struct kgi_display *dpy)
{
	NOTICE1("PVGA1A Graphics Driver removed");
}

int kgim_graphics_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			      enum kgi_tm_cmd cmd)
{
	/* !!! Should we check for accel-friendly dimensions here ??? */
	switch (cmd) {
		case CMD_PROPOSE:
		NOTICE1("Setting up mmio");
		return graphics_mmio_setup(dpy, mode);
		default:
		break;
	}
	
	return EOK;
}

void kgim_graphics_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{
  int lockval;
  /* Reset all bank offsets to 0 */
  outPR0A(0);
  outPR0B(0);
  /* Enable split bank */
  lockval = rwunlockGRC();
  switch(mode->request.graphtype) {
  case GT_8BIT:
    /*    if ((mode->request.virt.x * mode->request.virt.y) > VGA_WINDOW_SIZE) */
    outbitsGRC(GR0B_PR0B_ENAB, GR0B_PR0B_ENAB, GRC_MEMORY); 
      /*    else
	    outbitsGRC(0, GR0B_PR0B_ENAB, GRC_MEMORY); */
    break;
  case GT_TEXT16:
  case GT_TEXT32:
  case GT_1BIT:
  case GT_4BIT:
  default:
    outbitsGRC(0, GR0B_PR0B_ENAB, GRC_MEMORY); 
    break;
  }
  outbGRC(lockval, GRC_UPGRC_LOCK);
}

int kgim_graphics_command (struct kgi_graphic *gr, ggi_uint cmd, ggi_uint arg)
{ 
	ggi_suggest *sugg;
	const char *name;

	switch (cmd) {
		case GRAPHICS_GETSUGGEST:
		sugg = (ggi_suggest *)(gr->io_buf);
		gr->io_res = gr->io_buf;
		
		if (sugg->id < 0 || sugg->id > 1)
			return -E(GRAPHICS, BOGUS);
		if (sugg->id == 1) { /* Generic RAMDAC support */
			strcpy(sugg->name, "generic-ramdac");
			sugg->args[0] = 0;
			sugg->id = 0;
			return 0;
		}
		switch (gr->dev.mode.graphtype) {
			case GT_TEXT16:
			name = sugg_default[SG_TEXT16];
			break;
			case GT_1BIT:
			name = sugg_default[SG_1BIT];
			break;
			case GT_4BIT:
			name = sugg_default[SG_4BIT];
			break;
			case GT_8BIT:
			  if ((gr->dev.mode.virt.x * 
			       gr->dev.mode.virt.y) <= VGA_WINDOW_SIZE)
			    /*	if (gr->dev.mode.visible.x <= 320 &&
			    gr->dev.mode.visible.y <= 200) */
			    name = sugg_default[SG_8BIT];
			  else
			    name = sugg_default[SG_8BIT_MMAP];
			break;
			default:
			return -E(GRAPHICS, BOGUS);
		}
		strcpy(sugg->name, name);
		sugg->args[0] = 0;
		sugg->id = 1;	/* Next is RAMDAC */
		return 0;
		default:
		break;
	}
			
	return -E(GRAPHICS, NOSUP_ALWAYS_MMAP);
}

