/*\
|*| --------------------------------------------------------------------------
|*|	S3 Trio32/Trio64 register definitions
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995, Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#ifndef _S3_Trio_h
#define	_S3_Trio_h

#include "IBM/vga.h"
#include "S3/accel.h"
#include "S3/idac.h"

/*\
|*|	extended registers common to all S3 Vision family members.
\*/

/* additional bits in MISC reg */
	#undef	MISC_VIDEO_OFF
	#define MISC_EXT_CLOCK	0x0C

/* additional bits in InputStatus0 reg */
	#define INS0_MONITOR	0x10

/* Feature Control Register */
	#define	FCTRL_OR_VSYNC	0x08

/* additional sequencer registers (SCLK ... serializer clock) */
	/* SEQ_CLOCK (SR01) */
		#define	SR01_SCLK2	0x04
		#define SR01_SCLK4	0x10
		#define SR01_DCLK2	0x08
	#define	SEQ_UNLOCK	0x08
		#define	SR08_MAGIC	0x06
	#define	SEQ_IO_CTRL	0x09
		#define	SR09_MMIO_ONLY	0x80
	#define	SEQ_BUS_CTRL	0x0A
		#define	SR0A_PVALUEMASK	0x1F
		#define	SR0A__PD_TRIST	0x20
		#define	SR0A_P50_RAS1	0x40
		#define	SR0A_SLOW_WRITE	0x80
	#define	SEQ_DAC_MODE	0x0B
	/* SR0C reserved	*/
	#define	SEQ_SYNC_CTRL	0x0C
	#define	SEQ_EXTENDED	0x0D
		#define	SR0D_HSYNC_NORM	0x00
		#define SR0D_HSYNC_OFF	0x10
		#define	SR0D_HSYNC_ON	0x20
		#define	SR0D_VSYNC_NORM	0x00
		#define	SR0D_VSYNC_OFF	0x40
		#define SR0D_VSYNC_ON	0x80
	/* SR0E reserved	*/
	/* SR0F reserved	*/
	#define	SEQ_MCLK_LOW	0x10
	#define	SEQ_MCLK_HIGH	0x11
	#define	SEQ_DCLK_LOW	0x12
	#define	SEQ_DCLK_HIGH	0x13
	#define	SEQ_CLK_CTRL0	0x14
	#define	SEQ_CLK_CTRL1	0x15
	#define	SEQ_CLK_TEST0	0x16
	#define	SEQ_CLK_TEST1	0x17
	#define	SEQ_CLK_TEST2	0x18

/* additional cathode ray tube controller registers */
	#define	CRT_HWC_FOREGROUND	0x0E
	#define	CRT_HWC_BACKGROUND	0x0F

	#define	CRT_ID_HIGH		0x2D
	#define	CRT_CHIPID_HIGH		0x2D
		#define	CR2D_TRIO	0x88
	#define	CRT_ID_LOW		0x2E
	#define	CRT_CHIPID_LOW		0x2E
		#define	CR2E_TRIO32	0x10
		#define	CR2E_TRIO64	0x11
	#define	CRT_REVISION		0x2F
	#define	CRT_CHIP_REV		0x2F
		#define	CR2F_TRIO64PLUS_MASK	0xF0
		#define	CR2F_TRIO64PLUS_ID	0x40
		#define	CR2F_TRIO64PLUS_REVMASK	0x0F
		
	/* S3 VGA registers */
	#define	CRT_CHIPID	0x30
		#define	CR30_USE_CR2x	0xE1
	#define	CRT_MEMCONFIG0	0x31
		#define	CR31_ENBCPUBASE	0x01
		#define	CR31_2PAGE_SCRN	0x02
		#define	CR31_16BIT_BUS	0x04
		#define	CR31_ENHANCED	0x08
		#define	CR31_HISPEEDTXT	0x40
	#define	CRT_BACKCOMPAT1	0x32
		#define	CR32_ENB_VSYNC_IRQ 0x10
		#define	CR32_VGAMEMWRAP	0x40
	#define	CRT_BACKCOMPAT2	0x33
		#define	CR33_UNLOCK_VDE	0x02
		#define	CR33_VCLK_IS_DCLK 0x08
		#define	CR33_LOCK_DAC	0x10
		#define	CR33_NOBORDER	0x20
		#define	CR33_LOCK_PALETTE 0x40
	#define	CRT_BACKCOMPAT3	0x34
		#define	CR34_PCI_SNOOP	0x01
		#define	CR34_PCI_ABORT	0x02
		#define	CR34_PCI_RETRY	0x04
		#define	CR34_ENB_FIFO	0x10
	#define	CRT_LOCK	0x35
		#define	CR35_CPUBASE_MASK 0x0F
		#define	CR35_LOCK_HTIMING 0x20
		#define	CR35_LOCK_TIMING  0x30
	#define	CRT_CONFIG1	0x36
		#define	CR36_VESALOCAL	0x01
		#define	CR36_PCI	0x02
		#define	CR36_BUSMASK	0x03
		#define	CR36_1CYC_EDO	0x00
		#define	CR36_FASTPAGE	0x0C
		#define	CR36_EDO	0x08
		#define	CR36_RAMMASK	0x0C
		#define	CR36_ENB_BIOS	0x10
		#define	CR36_MEMSIZE_MASK 0xE0
		#define	CR36_4MB	0x00
		#define	CR36_2MB	0x80
		#define	CR36_1MB	0xC0
		#define	CR36_05MB	0xE0
	#define	CRT_CONFIG2	0x37
		#define	CR37_S3_ENABLE	0x01
		#define	CR37_TESTMODE	0x02
		#define	CR37_32K_BIOS	0x04
		#define	CR37_64K_BIOS	0x00
		#define	CR37_INT_CLK	0x08
		#define	CR37_VL_SNOOP	0x10
	#define	CRT_LOCK1	0x38
		#define	CR38_MAGIC	0x48
	#define	CRT_LOCK2	0x39
		#define	CR39_MAGIC	0xA0
		#define	CR39_CR36MAGIC	0xA5
	#define	CRT_MISC1	0x3A
		#define	CR3A_REFRESH0	0x00
		#define	CR3A_REFRESH1	0x01
		#define	CR3A_REFRESH2	0x02
		#define	CR3A_REFRESH3	0x03
		#define	CR3A_ALT_REFRESH 0x04
		#define	CR3A_REFRESHMASK 0x03
		#define	CR3A_TOP_MEM	0x08
		#define	CR3A_NO4BPPMODE	0x10
		#define	CR3A_HISPEEDTXT	0x20
		#define	CR3A_NOPCIBURST	0x80
		#define	CR3A_PCIMASK	0x80
	#define	CRT_FIFOSTART	0x3B
	#define	CRT_INTERLACESTART 0x3C
	/* system configuration resgisters */
	#define	CRT_SYSCONFIG	0x40
		#define	CR40_UNLOCK_ENH	0x01
		#define	CR40_VLRDY_CTRL	0x10
		#define	CR40_VLWR_DELAY	0x20
	#define	CRT_BIOSFLAG	0x41
	#define	CRT_EXTMODE1	0x42
		#define	CR42_INTERLACED	0x20
	#define	CRT_EXTMODE2	0x43
		#define	CR43_DOUBLE_HTIMING 0x80
	/* hardware cursor registers */
	#define	CRT_HWC_MODE	0x45
		#define	CR45_HWC_ENABLE	0x01
		#define	CR45_HWC_1280	0x10
	/* system extension registers */
	#define	CRT_SYSCONTROL1	0x50
		#define	CR50_WIDTHMASK	0xC1
		#define	CR50_1024	0x00
		#define	CR50_2048	0x00
		#define CR50_640	0x40
		#define	CR50_800	0x80
		#define	CR50_1280	0xC0
		#define	CR50_1152	0x01
		#define	CR50_1600	0x81
		#define	CR50_4BPP	0x00
		#define	CR50_8BPP	0x00
		#define	CR50_16BPP	0x10
		#define	CR50_32BPP	0x30
	#define	CRT_SYSCONTROL2	0x51
		#define	CR51_LOGICALWIDTH_B8	0x10
		#define	CR51_LOGICALWIDTH_B9	0x20
	#define	CRT_BIOSFLAG1	0x52
	#define	CRT_MEMCONTROL1	0x53
	   #ifdef TRIO64PLUS
		#define	CR53_NO_BYTE_SWAP	0x00
		#define	CR53_SWAP_BYTE_IN_WORD	0x01
		#define	CR53_SWAP_BYTE_IN_DWORD	0x20
		#define	CR53_MMIO_DISABLE	0x00
		#define	CR53_MMIO_RELOC_MMIO	0x08
		#define	CR53_MMIO_TRIO64_LIKE	0x10
		#define	CR53_MMIO_TRIO_AND_RELOC 0x18
		#define	CR53_MMIO_A0_TO_B0	0x00
		#define	CR53_MMIO_A0_TO_C0	0x20
	   #else
		#define	CR53_ENB_WPB	     0x01
		#define	CR53_ENB_MMIO	     0x10
	   #endif
		#define	CR53_SWAPNIBBLES     0x40
	#define	CRT_MEMCONTROL2	0x54
		#define	CR54_M_MASK	0xF8
	   #ifdef TRIO64PLUS
		#define CR54_NO_BYTE_SWAP	0x00
		#define	CR54_SWAP_BYTE_IN_WORD	0x01
		#define	CR54_SWAP_BYTE_IN_DWORD	0x02
		#define	CR54_SWAP_AS_PCI_OR_VL	0x03
	   #endif
	#define	CRT_DACCONTROL	0x55
		#define	CR55_INPORT_ENB	0x04
		#define	CR55_HWC_X11	0x10
		#define	CR55_VCLK_OFF	0x80
	#define	CRT_EXTSYNC1	0x56
		#define	CR56_HSYNC_TRISTATE	0x02
		#define	CR59_VSYNC_TRISTATE	0x04
	   #ifndef TRIO64PLUS
		#define	CR56_REMOTE	0x01
		#define	CR56_NOHSYNCCLR	0x08
		#define	CR56_START_ODD	0x10
	#define	CRT_EXTSYNC2	0x57
	   #endif			/* no mistake ! */
	#define	CRT_LAWCONTROL	0x58
		#define	CR58_LAW_64K	0x00
		#define	CR58_LAW_1MB	0x01
		#define	CR58_LAW_2MB	0x02
		#define	CR58_LAW_4MB	0x03
		#define	CR58_VL_T1_LATCH 0x08
		#define	CR58_LAW_ENABLE	0x10
		#define	CR58_RAS_PRECHARGE 0x80
	#define	CRT_LAW_POS_H	0x59
	#define	CRT_LAW_POS_L	0x5A
	#define	CRT_BIOSFLAG2	0x5B
	#define	CRT_GOUT_PORT	0x5C
		#define	CR5C_GOUT_MASK	0xFF
	#define	CRT_EH_OVERFLOW	0x5D
		#define	CR5D_HTOTAL_B8	    0x01
		#define	CR5D_HDISPLAYEND_B8 0x02
		#define	CR5D_HBLANKSTART_B8 0x04
		#define	CR5D_HBLANKEND_B6   0x08
		#define	CR5D_HSYNCSTART_B8  0x10
		#define	CR5D_HSYNCEND_B5    0x20
		#define	CR5D_FIFOSTART_B8   0x40
	   #ifndef TRIO64PLUS
		#define	CR5D_BGT_B8         0x08
	   #endif
	#define	CRT_EV_OVERFLOW	0x5E
		#define	CR5E_VTOTAL_B10	     0x01
		#define	CR5E_VDISPLAYEND_B10 0x02
		#define	CR5E_VBLANKSTART_B10 0x04
		#define	CR5E_VSYNCSTART_B10  0x10
		#define	CR5E_LINECOMPARE_B10 0x40
	#define	CRT_MEMCONTROL3	0x60
	#define	CRT_MEMCONTROL4	0x61
	   #ifdef TRIO64PLUS
		#define	CR61_NO_SWAP
		#define	CR61_SWAP_BYTE_IN_WORD	0x20
		#define	CR61_SWAP_BYTE_IN_DWORD	0x40
	   #endif
	#define	CRT_EXTSYNCADJ	0x63
	#define	CRT_EXTMISC0	0x65
		#define	CR65_0x3C3	0x04
		#define	CR65_BLANK_DELAY_0DCLK	0x00
		#define	CR65_BLANK_DELAY_1DCLK	0x08
		#define	CR65_BLANK_DELAY_2DCLK	0x10
		#define	CR65_BLANK_DELAY_3DCLK	0x18
	#define	CRT_EXTMISC1	0x66
		#define	CR66_PIXELBUS_OFF    0x40
		#define	CR66_PCI_DISCONNECT  0x80
	   #ifdef TRIO64PLUS
		#define	CR66_ENB_ENHANCED	0x01
		#define	CR66_RESET_GRAPHIC_ENG	0x02
		#define	CR66_PCI_FIFO_DISCONNECT	0x08
	   #endif
	#define	CRT_EXTMISC2	0x67
		#define	CR67_NO_INVERT_VCLK  0x01
	   #ifdef TRIO64PLUS
		#define	CR67_STREAMS_DISABLE	0x00
		#define	CR67_OVRLAY_ON_VGA	0x04
		#define	CR67_STREAMS_ENABLE	0x0C
	   #endif
		#define	CR67_DAC_MASK	0xF0
		#define	CR67_DAC_MODE0	0x00
		#define	CR67_DAC_MODE1	0x20
		#define	CR67_DAC_MODE2	0x40
		#define	CR67_DAC_MODE3	0x60
		#define	CR67_DAC_MODE8	0x10
		#define	CR67_DAC_MODE9	0x30
		#define	CR67_DAC_MODEA	0x50
		#define	CR67_DAC_MODEB	0x70
	#define	CRT_CONFIG3	0x68
		#define	CR68_CAS_MASK	0x03
		#define	CR68_MAD_MASK	0x0C
		#define	CR68_RASLOW_MASK 0x30
		#define	CR68_RASPCG_MASK 0xA0
		#define	CR68_64B_MEM_BUS	0x80
	#define	CRT_EXTCONTROL3	0x69
		#define	CR69_STARTADDR_MASK 0x0F
	#define	CRT_EXTCONTROL4	0x6A
		#define	CR69_CPUBASE_MASK   0x3F
	#define	CRT_BIOS3	0x6B
	#define	CRT_BIOS4	0x6C
	#define	CRT_MISC3	0x6D
	#define	CRT_CONFIG4	0x6F
		#define	CR6F_TRIO64_MODE	0x01
		#define	CR6F_TRIO64_LPB		0x00
		#define	CR6F_SERIAL_IO_00E8	0x00
		#define	CR6F_SERIAL_IO_00E2	0x02
		#define	CR6F_SERIAL_IO_MEM_ONLY	0x04
		#define	CR6F_WE_DELAY_MASK	0x18

/* setup registers */
#define	S3_SETUP	0x0102
	#define	S3_WAKEUP	0x01
#define	S3_SUBSYSTEM	0x46E8
	#define	S3_ENABLE	0x08
	#define	S3_DISABLE	0x10

/*\
|*|	virtual DAC registers. The SRxx registers have to be programmed
|*|	when outbDAC(val, DAC_reg) is called.
\*/

	#define	DAC_MODE	0x04	/* this maps to SR0B	*/
		#define	DAC_M_VCLKI	0x01
		#define	DAC_M_USE_VCLKI	0x02
		#define	DAC_M_24BPP	0x08
		#define	DAC_M_MODE0	0x00	/* 8bpp, 1:1 DCLK/VCLK	*/	*/
		#define	DAC_M_MODE8	0x10	/* 8bpp, 2:1		*/
		#define	DAC_M_MODE9	0x30	/* 15bpp, 1:1		*/
		#define DAC_M_MODE10	0x50	/* 16bpp, 1:1		*/
		#define	DAC_M_MODE12	0x70	/* 24bpp, 1:3		*/
		#define	DAC_M_MODE13	0xD0	/* 32bpp, 1:1		*/
	/* 0x05 (SR0C) reserved	*/
	#define	DAC_SYNC_CTRL	0x06	/* this maps to SR0D	*/
		#define	DAC_SC_FEAT_ENB	0x01	/* enb feat. connector	*/
		#define	DAC_SC_HS_NRM	0x00
		#define	DAC_SC_HS_OFF	0x10
		#define	DAC_SC_HS_ON	0x20
		#define	DAC_SC_VS_NRM	0x00
		#define	DAC_SC_VS_OFF	0x40
		#define	DAC_SC_VS_ON	0x80
	/* 0x07 (SR0E) reserved	*/
	/* 0x08	(SR0F) reserved	*/
	#define	DAC_MCLK_LOW	0x09	/* this maps to SR10	*/
	#define	DAC_MCLK_HIGH	0x0A	/* this maps to SR11	*/
	#define	DAC_DCLK_LOW	0x0B	/* this maps to SR12	*/
	#define	DAC_DCLK_HIGH	0x0C	/* this maps to SR13	*/
	#define	DAC_CLK_CTRL0	0x0D	/* this maps to SR14	*/
		#define	DAC_C0_DCLK_PWR	0x01	/* DCLK PLL power down	*/
		#define	DAC_C0_MCLK_PWR	0x02	/* MCLK PLL power down	*/
		#define	DAC_C0_ENB_CNT	0x04	/* enable PLL counters	*/
		#define	DAC_C0_MCLKTEST	0x08	/* test MCLK		*/
		#define	DAC_C0_CLR_CNT	0x10	/* clear PLL counters	*/
		#define	DAC_C0_P146_FN	0x20	/* function of pin 146	*/
		#define	DAC_C0_EXT_MCLK	0x40	/* external MCLK	*/
		#define	DAC_C0_EXT_DCLK	0x80	/* external DCLK	*/
	#define	DAC_CLK_CTRL1	0x0E	/* this maps to SR15	*/
		#define	DAC_C1_LD_MCLK	0x01	/* enable MCLK loading	*/
		#define	DAC_C1_LD_DCLK	0x02	/* enable DCLK loading	*/
		#define	DAC_C1_MCLK_OUT	0x04	/* output internal MCLK	*/
		#define	DAC_C1_VCLK_OUT	0x08	/* output internal VCLK	*/
		#define	DAC_C1_DCLK2	0x10	/* divide DCLK by 2	*/
		#define	DAC_C1_CLK_LOAD	0x20	/* load clock values 	*/
		#define	DAC_C1_DCLK_INV	0x40	/* invert DCLK		*/
		#define	DAC_C1_FAST_CYC	0x80	/* fast memory cycle	*/
	/* 0x0F (SR16) reserved	*/
	/* 0x10 (SR17) reserved	*/
	#define	DAC_CLK_TEST2	0x11	/* this maps to SR18	*/
		#define	DAC_CT_TEST_ENB	0x01
		#define	DAC_CT_TEST_RST	0x02
		#define	DAC_CT_TEST_RED	0x04
		#define	DAC_CT_TEST_GRN	0x08
		#define	DAC_CT_TEST_BLU	0x10
		#define	DAC_CT_DAC_PWR	0x20	/* DAC power down	*/
		#define	DAC_CT_FAST_WR	0x40	/* fast LUT write cycle	*/
		#define	DAC_CT_CLK2	0x80	/* clock doubled mode	*/

#endif	/* #ifdef _S3_Trio_h ...	*/
