/*\
|*| --------------------------------------------------------------------------
|*|	monitor driver for multisync monitors
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| --------------------------------------------------------------------------
|*|
|*|  $Log: multisync.c,v $
|*|  Revision 1.7  1998/02/09 15:26:55  grimrath
|*|  New KGI driver Makefile system.  Include faster dependencies, true linking and
|*|  a predefined driver constant for checking with #ifdef ... #endif
|*|
|*|  VGA, the Mystique driver and kernel files have been adapted to the
|*|  text <-> dpp change.
|*|
|*|  Revision 1.6  1998/02/06 19:41:39  vogler
|*|  fixed a typo
|*|
|*|  Revision 1.5  1998/02/02 21:54:35  becka
|*|  Slight fixes.
|*|
|*|  Revision 1.4  1997/12/30 22:09:07  grimrath
|*|  *** empty log message ***
|*|
|*|  Revision 1.3  1997/12/14 21:56:13  becka
|*|  uint/sint namespace fixed ... please test ...
|*|
|*|  Revision 1.2  1997/09/12 07:28:50  taylorj
|*|  Minor: Reformatted source to be consistent with GGI coding style guidelines,
|*|  modified all debugging output to use the DEBUGx macros, and added CVS
|*|  revision logging to headers where needed.
|*|
|*|  Revision 1.1  1997/06/28 12:14:07  martin
|*|  Ported multisync monitor driver from GGI-0.0.8
|*|
|*| ---------------------------------------------------------------------------
\*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Steffen_Seeger
#define DRIVER_NAME	"Multisync monitor driver"
#define DRIVER_REV	"$Revision: 1.7 $"

#ifdef DEBUG
#define NEED_MONITOR_SYMS
#endif

#include <ggi/module.h>
#include "monitors.inc"

/*\
|*|	GGI interface
\*/

int kgim_monitor_init(struct kgi_card *card, struct kgi_display *dpy)
{
	if (!(monitor.synctype & SYNC_MULTISYNC))
	{
		NOTICE0("The monitor configured is not of type MULTISYNC!");
		return -E(MONITOR,UNKNOWN);
	}

	DEBUG1("%s %s Monitor Driver",
	       monitor.manufact, monitor.model);
	DEBUG1("(%dx%d Pixel, %d mm x %d mm, %d-%d kHz, %d-%d Hz)",
	       monitor.maxdots.x, monitor.maxdots.y,
	       monitor.size.x, monitor.size.y,
	       monitor.hfreq.min / 1000 ,monitor.hfreq.max / 1000,
	       monitor.vfreq.min, monitor.vfreq.max);

	card->monitor = &monitor;
	return EOK;
}

void kgim_monitor_done(struct kgi_display *dpy)
{
	DEBUG1("%s %s Monitor driver removed.", 
	       monitor.manufact, monitor.model);
}

int kgim_monitor_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			     enum kgi_tm_cmd *cmd)
{ 
	int hfreq, vfreq, result;
	int width = mode->x.width;

	switch (*cmd)
	{
	case CMD_PROPOSE:
		mode->dclk = monitor.dclk.max;

		for(hfreq = monitor.hfreq.max; hfreq > monitor.hfreq.min;
		    mode->dclk -= 1000)
		{
			/*mode->x.syncstart = width * (width + 41360) / 40000;
			mode->x.syncend = width * (width * 7 / 10 + 8952) / 8000;
			mode->x.total = width * (width + 9360) / 8000;*/
			mode->x.syncstart = width * (1000 + 50) / 1000;
			mode->x.syncend = width * (1000 + 180) / 1000;
			mode->x.total = width * (1000 + 350) / 1000;

			/* !!! A little hack for the Mystique */
			if (!strcmp(dpy->card->model,"Mystique") &&
			    mode->request.graphtype == GT_8BIT) {
				mode->x.total     = (mode->x.total     + 15) & ~15;
				mode->x.syncstart = (mode->x.syncstart + 15) & ~15;
				mode->x.syncend   = (mode->x.syncend   + 15) & ~15;
			}

			mode->x.blankstart = mode->x.syncstart;
			mode->x.blankend = mode->x.syncend;

			hfreq = mode->dclk / mode->x.total;
			if (monitor.hfreq.max < hfreq)
				continue;

			mode->y.syncstart = mode->y.width + 3;
			mode->y.syncend   = mode->y.width + 9;
			mode->y.total     = mode->y.width + hfreq / 1000 + 1; /* 1ms flyback */
			mode->y.blankstart= mode->y.syncstart; 
			mode->y.blankend  = mode->y.syncend;

			mode->x.polarity  = mode->y.polarity = 1; /* Doesn't matter */

			vfreq = mode->dclk / (mode->x.total * mode->y.total);

			if ((monitor.vfreq.min <= vfreq) && 
			    (vfreq <= monitor.vfreq.max))
			{
				
				if (monitor.flags & FL_MON_PROPSIZE)
				{
					mode->request.size.x = 
						mode->x.width * 
						monitor.size.x /
						monitor.maxdots.x;
					mode->request.size.y = 
						mode->y.width * 
						monitor.size.y /
						monitor.maxdots.y;
				}
				else 
				{
					
					mode->request.size = monitor.size;
				}

				DEBUG1("Calculated %i Hz refresh (%i Hz, "
				       "DCLK = %i Hz)",
				       vfreq, hfreq, mode->dclk);

				*cmd = CMD_LOWER;
				return EOK;
			}
		}
		
		NOTICE1("Invalid mode request.");
		return -E(MONITOR,UNKNOWN);
		break;

	case CMD_LOWER:
		if (mode->dclk < monitor.dclk.min)
		{
			
			return -E(MONITOR,UNKNOWN);
		}

		*cmd = CMD_CHECK;
		return EOK;

	case CMD_RAISE:
		if (mode->dclk > monitor.dclk.max)
		{
			
			return -E(MONITOR,UNKNOWN);
		}
		

		*cmd = CMD_CHECK;
		return EOK;

	case CMD_CHECK: 
		hfreq = mode->dclk / mode->x.total;
		vfreq = mode->dclk / (mode->x.total * mode->y.total);
		result =
			(monitor.hfreq.min - (monitor.hfreq.min >> 7) <= hfreq) &&
	/*		(monitor.hfreq.min <= hfreq) && */
			(hfreq <= monitor.hfreq.max) && 
			(monitor.vfreq.min <= vfreq) && 
			(vfreq <= monitor.vfreq.max) && 
			(monitor.dclk.min <= mode->dclk) && 
			(mode->dclk <= monitor.dclk.max);
		if (result) 
		{
			
			DEBUG2("Accepted %i Hz refresh (%i Hz, DCLK=%i Hz)",
			       vfreq, hfreq, mode->dclk);
		}
		else 
		{
			
			DEBUG2("Refresh check failed. (%i Hz, %i Hz, %i Hz)",
			       vfreq, hfreq, mode->dclk);
		}
		
		*cmd = CMD_READY;
		return result ? EOK : -E(MONITOR,UNKNOWN);
		break;

	default:
		INTERNALERROR;
		return -E(MONITOR,UNKNOWN);
	}
}

void kgim_monitor_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
}

int kgim_monitor_command(struct kgi_graphic *gr, ggi_uint cmd, ggi_uint arg)
{
	switch(cmd)
	{
	case MONITOR_GETINFO:
		gr->io_res = &monitor;
		return EOK;
		
	case MONITOR_SETINFO:
		if (suser())
		{
			
			memcpy(&monitor, gr->io_buf, sizeof(monitor));
			return EOK;
		}
		else
		{
			
			return -EPERM;
		}
		
	default:
		return -E(MONITOR,UNKNOWN);
	}
}
