/* --------------------------------------------------------------------------
 *	Cyrix MediaGX internal ramdac driver
 * --------------------------------------------------------------------------
 *
 *	Copyright (C) 1998		Emmanuel Marty
 *	Copyright (C) 1998		Cyrix Corporation
 *	Copyright (C) 1998		The GGI project
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 * 
 *	You should have received a copy of the GNU General Public License
 *	along with this program; see the file COPYING.  If not, write to
 *	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * --------------------------------------------------------------------------
 *
 *	This code was derived from the following sources of information :
 *
 *	Cyrix MediaGX confidential databooks
 *	Cyrix MediaGX gxvideo.c example sourcecode
 *	Help and code from Sengan Baring-Gould of Cyrix Corporation
 *
 *	Many thanks to the Cyrix Corporation for providing me with
 *	documentation and official support about the GX chipset family.
 *
 * --------------------------------------------------------------------------
 *
 *	$Log: gx.c,v $
 *	Revision 1.1  1998/03/19 21:31:39  emarty
 *	initial checkin
 *	
 */

#include <ggi/maintainers.h>
#define MAINTAINER	Emmanuel_Marty
#define DRIVER_NAME	"Cyrix MediaGX DAC driver"
#define DRIVER_REV	"$Revision: 1.1 $"
#define DEBUG_LEVEL	0

#include <ggi/module.h>
#include <Cyrix/gx.h>

/*	Maximum dotclock	*/

#define DAC_MAX_CLK 135000000

/*
**	RAMDAC definition for KGI
*/

static struct kgi_ramdac ramdac = {
	"Cyrix",			/* manufacturer 	*/
	"GX",				/* model		*/
	GM_TEXT16 | GM_8BIT |		/* supported modes	*/
	GM_15BIT | GM_16BIT,
	FL_DAC_GAMMA,			/* flags		*/
	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/
	DAC_16BIT_PB | DAC_8BIT_PB,	/* device type		*/
	{       0,  DAC_MAX_CLK },	/* normal mode		*/
	{       0,  0 },		/* 2xclk mode		*/

	{{	0,  DAC_MAX_CLK },	/* 	GT_TEXT16	*/
	 {	0,  0 },		/* 	GT_TEXT32 	*/
	 {	0,  0 },		/* 	GT_1BIT		*/
	 {	0,  0 },		/* 	GT_4BIT		*/
	 {	0,  DAC_MAX_CLK },	/* 	GT_8BIT		*/
	 {	0,  DAC_MAX_CLK },	/* 	GT_15BIT	*/
	 {	0,  DAC_MAX_CLK },	/* 	GT_16BIT	*/
	 {	0, 0 },			/* 	GT_24BIT	*/
	 {	0, 0 }}			/*	GT_32BIT	*/
};

/*
**	set and get internal CLUT
*/

#define CLUT_MAGIC	1040		/* 2^10 + 2^4 */
#define	CLUT_SHIFT	10

void dac_set_clut (struct kgi_display *dpy, ggi_clut *clut,
		   ggi_uint first, ggi_uint cnt) {
	struct chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr gxbase = private->gx_control.baseptr;
	uint32 lockval, p;
	uint16 *pal = (uint16 *) &clut->data[first];
	
	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	WRITE_GX32 (gxbase, DC_PAL_ADDRESS, first);
	
	ASSERT (cnt);

	while (cnt--) {
		p = (((uint32) *pal++) >> CLUT_SHIFT) << 12;
		p |= (((uint32) *pal++) >> CLUT_SHIFT) << 6;
		p |= (((uint32) *pal++) >> CLUT_SHIFT);
		
		WRITE_GX32 (gxbase, DC_PAL_DATA, p);
	}

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

void dac_get_clut (struct kgi_display *dpy, ggi_clut *clut,
		   ggi_uint first, ggi_uint cnt) {
	struct chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr gxbase = private->gx_control.baseptr;
	uint32 lockval, p;
	uint16 *pal = (uint16 *) &clut->data[first];

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	WRITE_GX32 (gxbase, DC_PAL_ADDRESS, first);
	
	ASSERT (cnt);

	while (cnt--) {
		p = READ_GX32 (gxbase, DC_PAL_DATA);
		*pal++ = ((p >> 12) & 0x3F) * CLUT_MAGIC;
		*pal++ = ((p >> 6)  & 0x3F) * CLUT_MAGIC;
		*pal++ = (p         & 0x3F) * CLUT_MAGIC;
	}

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

/*
**	GGI interface
*/

int kgim_ramdac_init (struct kgi_card *card, struct kgi_display *dpy) {
	NOTICE (DRIVER_NAME " rev " DRIVER_REV);

	card->ramdac = &ramdac;	
	return EOK;
}

void kgim_ramdac_done (struct kgi_display *dpy) {
	NOTICE (DRIVER_NAME " removed ");
}

int kgim_ramdac_check_mode (struct kgi_display *dpy,
			    struct kgi_mode *mode, enum kgi_tm_cmd cmd) {
	ggi_uint gt, newclock = mode->dclk, cmul, cdiv;
	
	switch (gt = mode->request.graphtype) {
		case GGI_AUTO:
			mode->request.graphtype = gt = GT_8BIT;
			break;

		case GT_TEXT16:
		case GT_8BIT:
		case GT_15BIT:
		case GT_16BIT:
			break;
			
		default:
			DEBUG1 ("graphtype %i not supported", gt);
			return -E(RAMDAC, UNKNOWN);
	}

	switch (cmd) {
		case CMD_PROPOSE:
			mode->clk.mul = mode->lclk.mul = 1;
			mode->clk.div = mode->lclk.div = 1;
			mode->flags  &= ~FL_DAC_2XCLOCK;
			
			return EOK;
			
		case CMD_LOWER:
			if (newclock * (cmul = mode->clk.mul) /
				(cdiv = mode->clk.div) < ramdac.normal.min) {
				DEBUG2 ("DCLK too low (%i Hz)", newclock);

				return -E (RAMDAC, CLK_LOW_LIMIT);
			}
			
			if (newclock * cmul / cdiv > ramdac.normal.max)
				mode->dclk = ramdac.normal.max / cmul * cdiv;

			return EOK;

		case CMD_RAISE:
			if (newclock * (cmul = mode->clk.mul) /
				(cdiv = mode->clk.div) > ramdac.normal.max) {
				DEBUG2 ("DCLK too high (%i Hz)", newclock);

				return -E (RAMDAC, CLK_LOW_LIMIT);
			}
			
			if (newclock * cmul / cdiv < ramdac.normal.min)
				mode->dclk = ramdac.normal.min / cmul * cdiv;

			return EOK;
	
		case CMD_CHECK:
			if ((newclock * (cmul = mode->clk.mul) /
				(cdiv = mode->clk.div) < ramdac.normal.min) ||
				(newclock * cmul / cdiv > ramdac.normal.max)) {

				DEBUG2 ("%i Hz DCLK out of limits", newclock);
				return -E(RAMDAC, MODE_CHECK);
			}
		
			return EOK;
		
		default:
			INTERNAL_ERROR;
			
			return -E (RAMDAC, BOGUS);
	}
}

void kgim_ramdac_set_mode (struct kgi_display *dpy, struct kgi_mode *mode) {
	switch (mode->request.graphtype) {
		case GT_TEXT16:
		case GT_8BIT:
			dpy->set_clut = dac_set_clut;
			break;
			
		case GT_15BIT:
		case GT_16BIT:
			dpy->set_clut = NULL;
			break;
			
		default:
			INTERNAL_ERROR;
			break;
	}
}

int kgim_ramdac_command (struct kgi_graphic *graph,ggi_uint cmd,ggi_uint arg) {
	ggi_clut clut;
	
	switch (cmd) {
		case RAMDAC_SETCLUT:
			clut.size = 256;
			clut.data = (ggi_color *) (graph->io_buf);
			dac_set_clut (graph->dev.dpy, &clut, 0, 256);
			return EOK;
		
		case RAMDAC_GETCLUT:
			clut.size = 256;
			clut.data = (ggi_color *) (graph->io_buf);
			graph->io_res = graph->io_buf;
			dac_get_clut (graph->dev.dpy, &clut, 0, 256);
			return EOK;
		
		default:
			return -E (RAMDAC, NOSUP_ALWAYS_CANT);	/* can't */
	}
}
			