/* ---------------------------------------------------------------------------
 *	VGA ramdac driver
 * ---------------------------------------------------------------------------
 *
 * Copyright (C) 1995-1997 The Linux-GGI Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ---------------------------------------------------------------------------
 *
 * $Log: vga.c,v $
 * Revision 1.18  1997/12/18 00:19:36  becka
 * Killed a few bugs.
 *
 * Revision 1.17  1997/12/14 21:56:20  becka
 * uint/sint namespace fixed ... please test ...
 *
 * Revision 1.16  1997/12/09 10:11:48  seeger
 * Global cleanup for new include directory locations.
 *
 * Revision 1.15  1997/09/13 00:13:32  taylorj
 * *** empty log message ***
 *
 * Revision 1.14  1997/09/12 07:28:52  taylorj
 * Minor: Reformatted source to be consistent with GGI coding style guidelines,
 * modified all debugging output to use the DEBUGx macros, and added CVS
 * revision logging to headers where needed.
 *	
 * Revision 1.13  1997/09/06 06:17:21  taylorj
 * *** empty log message ***
 *	
 * Revision 1.12  1997/09/06 04:12:02  taylorj
 * *** empty log message ***
 *	
 * Revision 1.11  1997/06/16 20:53:07  seeger
 * Changed to GGI coding style and new error/message reporting.
 *	
 * Revision 1.9  1997/04/18 21:40:45  seeger
 * Changed head to GGI coding style.
 * Changed to new MAINTAINER scheme
 * DAC_PRECISION must be defined local to the ramdac driver.
 *
 * Revision 1.8  1997/04/17 06:17:12  ezrec
 * Minor: Fixes to work with the new system.h and module.h files...
 *
 * Revision 1.7  1997/04/17 05:08:05  ezrec
 * Major: New ggiOpen() semantics ie ggiOpen("display-KGI:/dev/graph3",NULL)
 * First attempt at linear_1 frame buffer
 * Moved thread support `to the side'
 * Added DPRINT() support
 * Added display-KGI driver
 *
 * Revision 1.6  1997/04/12 04:51:30  ezrec
 * Minor: Fixes to the VGA subsystem to work w/Steffen's new MMAP system
 *
 * Revision 1.5  1997/04/10 06:36:40  ezrec
 * Major: Fixes for Steffen's new MMAP...
 *
 * Revision 1.4  1997/04/10 03:34:08  ezrec
 * Major: More fixes, and I'm attempting (ha!) to use Steffen's new MMAP 
 * scheme..
 *
 * Revision 1.3  1997/04/09 04:18:01  ezrec
 * Major: Added palettes in include/, fixed 8bit for VGA, and
 * got libggi-dynamic operational... Now to make it work
 * for the _new_ MMAP scheme!
 *
 * Revision 1.2  1997/03/21 01:12:26  ezrec
 * Minor: Name change for kgim_ramdac_ioctl -> kgim_ramdac_command
 *
 * Revision 1.1  1997/02/03 15:10:22  ezrec
 * Major: Added IBM/VGA ramdac driver
 *
 * ---------------------------------------------------------------------------
 */

#include <ggi/maintainers.h>
#define	MAINTAINER	Jon_Taylor
#define	DRIVER_NAME	"VGA DAC driver"
#define	DRIVER_REV	"$Revision: 1.18 $"
#define	DEBUG_LEVEL	0

#include <ggi/module.h>
#include <ggi/ggi_commands.h>

#define	DAC_PRECISION	6

#include "IBM/vga_dac.h"
#include "IBM/clut.inc"

static struct kgi_ramdac ramdac = 
{
	"IBM",				/* Manufacturer		*/
	"VGA",				/* Model		*/
	GM_TEXT16 | GM_4BIT | GM_8BIT,	/* Supported modes	*/
	0,			 	/* Flags		*/
	{ 0xFFFF, 0xFFFF },		/* Maximum resolution	*/
	DAC_8BIT_PB,  			/* Device type		*/
	{       0,  32000000 },		/* Normal mode		*/
	{       0,         0 },		/* 2xclk mode		*/
	{
		{	  0,  32000000 },	/* GT_TEXT16	*/
		{	  0,         0 },	/* GT_TEXT32	*/
		{         0,  32000000 },	/* GT_1BIT	*/
		{	  0,  32000000 },	/* GT_4BIT	*/
		{         0,  32000000 },	/* GT_8BIT	*/	
		{	  0,         0 },	/* GT_15BIT	*/
		{ 	  0,         0 },	/* GT_16BIT	*/
		{         0,         0 },	/* GT_24BIT	*/
		{	  0,         0 }	/* GT_32BIT	*/
	}
};

/* GGI interface */

int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev. " DRIVER_REV);
	return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " removed.");
}

int kgim_ramdac_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
	enum kgi_tm_cmd cmd)
{
	ggi_uint newclock = mode->dclk;
  
	switch (mode->request.graphtype) {

	case GT_TEXT16:
		break;

	case GT_1BIT:
		break;

	case GT_4BIT:
		break;   

	case GT_8BIT:
		break;
  
	default:
  		DEBUG2("couldn't handle graphtype %i", 
		       mode->request.graphtype);
		return -E(RAMDAC,MODE_CHECK);
	}

	switch (cmd) {

	case CMD_PROPOSE:
		mode->lclk.mul = mode->clk.mul = 1;
		mode->lclk.div = mode->clk.div = 1;
		return EOK;

	case CMD_LOWER:

		if ((newclock < ramdac.clock[mode->request.graphtype].min) ||
			(newclock < ramdac.normal.min)) {

			DEBUG2("DCLK too low (%i Hz).", newclock);
			return -E(RAMDAC,CLK_LOW_LIMIT);
		}

		if (ramdac.clock[mode->request.graphtype].max < newclock) {

			newclock = ramdac.clock[mode->request.graphtype].max;
		}
		mode->dclk = newclock;
		return EOK;

	case CMD_RAISE:
		if ((newclock > ramdac.clock[mode->request.graphtype].max) ||
			(newclock > ramdac.normal.max)) {

			DEBUG2("DCLK too high (%i Hz).", newclock);
			return -E(RAMDAC,CLK_HIGH_LIMIT);
		}

		if (newclock < ramdac.clock[mode->request.graphtype].min) {

			newclock = ramdac.clock[mode->request.graphtype].min;
		}

		if (newclock < ramdac.normal.min) {

			newclock = ramdac.normal.min;
		}

		mode->dclk = newclock;
		return EOK;

     case CMD_CHECK:
		if ((mode->lclk.mul != 1) || (mode->lclk.div != 1) ||
		    (mode->clk.mul != 1) || (mode->clk.div != 1)) {

			return -E(RAMDAC,MODE_CHECK);
		}

		if ((mode->dclk < ramdac.clock[mode->request.graphtype].min) ||
		    (mode->dclk < ramdac.normal.min)) {

			return -E(RAMDAC,CLK_LOW_LIMIT);
		}

		if ((mode->dclk > ramdac.clock[mode->request.graphtype].max) ||
		    (mode->dclk > ramdac.normal.max)) {
			
			return -E(RAMDAC,CLK_HIGH_LIMIT);
		}

		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(RAMDAC,BOGUS);
	}
}

void kgim_ramdac_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	DAC_out8(dpy, 0xFF, DAC_PIXEL_MASK);
	dpy->set_clut = dac_set_clut;
}

int kgim_ramdac_command(struct kgi_graphic *graph, ggi_uint cmd, ggi_uint arg)
{
	/* This may set 
	 * -  Hardware cursor
	 * -  CLUT (or gamma correction table)
	 */
	ggi_clut clut;

	switch (cmd) { 

	case RAMDAC_SETCLUT:
		clut.size = 256;
		clut.data = (ggi_color *)(graph->io_buf);
		dac_set_clut(graph->dev.dpy, &clut, 0, 256);
		return 0;

	case RAMDAC_GETCLUT:
		clut.size = 256;
		clut.data = (ggi_color *)(graph->io_buf);
		graph->io_res = graph->io_buf;
		dac_get_clut(graph->dev.dpy, &clut, 0, 256);
		return 0;
	}
  	return -E(RAMDAC, NOSUP_ALWAYS_CANT);  /* The ramdac can't do this */
}
