/* ---------------------------------------------------------------------------
**	ICS 5342 ramdac driver
** ---------------------------------------------------------------------------
**
**	Copyright (C) 1997 by Andreas Vogler, [duff@allgaeu.org]
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
**	This code is derived from the following source of information
**	[1]	ICS 5342 data sheet
**
**	$Log: 5342.c,v $
**	Revision 1.9  1998/03/19 21:22:36  emarty
**	GGI_AUTO support
**	
**	Revision 1.8  1998/01/29 08:32:13  emarty
**	set 135 mhz as max freq rather than 80
**	
**	Revision 1.7  1997/12/18 22:38:35  becka
**	Fixed up RAMDAC drivers. Small spelling error corrected.
**	
**	Revision 1.6  1997/12/18 09:51:33  emarty
**	removed clut setting for good
**	
**	Revision 1.5  1997/12/16 13:33:40  emarty
**	commented clut setting out
**	
**	Revision 1.4  1997/12/14 21:56:21  becka
**	uint/sint namespace fixed ... please test ...
**	
**	Revision 1.3  1997/09/02 17:23:28  vogler
**	minor bug fixes
**	
**	Revision 1.2  1997/08/15 16:12:37  vogler
**	added 5300 and 5342 GENDACS
**	
**	Revision 1.1  1997/08/03 20:13:26  seeger
**	ICS 5342 DAC driver by Andreas Vogler, initial checkin
**	
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Andreas_Vogler
#ifndef DRIVER_MODEL
#define DRIVER_MODEL	"5342"
#endif
#define DRIVER_NAME	"ICS " DRIVER_MODEL " DAC driver"
#define	DRIVER_REV	"$Revision: 1.9 $"

#define	DEBUG_LEVEL	0

#include <ggi/module.h>

#define DAC_PRECISION 6

#include "IBM/vga_dac.h"
#include "IBM/clut.inc"
#include "ICS/5342.h"

/*	define this if your chipset supports 
**	3 dac register select lines 
*/
/*	NOTE: you should always assume only two RS lines. (As it will work
**	with all chipsets then. -- seeger
*/
#undef DAC_RS2

/*	ICS5342-1	80MHz version
**	ICS5342-2	110MHz version
**	ICS5342-3	135MHz version
*/

/*	#define DAC_MAX_CLK 80000000	*/ /* ICS5342-1 */
/* #define DAC_MAX_CLK 110000000 */
#define DAC_MAX_CLK 135000000

static struct kgi_ramdac ramdac =
{
	"Integrated Circuit Systems",	/* manufacturer		*/
 	DRIVER_MODEL,				/* model		*/
	GM_TEXT16 | GM_1BIT |
	GM_4BIT | GM_8BIT |		/* supported modes	*/
	GM_15BIT | GM_16BIT |
	GM_24BIT
#ifdef _ICS_5300_
	,
#else
        | GM_32BIT,
#endif
	FL_DAC_GAMMA, 			/* flags		*/
 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/

#ifdef _ICS_5300_
	DAC_8BIT_PB,
#else
	DAC_16BIT_PB | DAC_8BIT_PB,	/* device type		*/
#endif
	{       0,  DAC_MAX_CLK },	/* normal mode		*/
	{       0,  0 },		/* 2xclk mode		*/

	{{	  0,  DAC_MAX_CLK },	/* 	GT_TEXT16	*/
	 {	  0,  0 },		/* 	GT_TEXT32 	*/
	 {       0,  DAC_MAX_CLK },	/* 	GT_1BIT		*/
	 {	  0,  DAC_MAX_CLK },	/* 	GT_4BIT		*/
	 {       0,  DAC_MAX_CLK },	/* 	GT_8BIT		*/
	 {	  0,  DAC_MAX_CLK },	/* 	GT_15BIT	*/
	 { 	  0,  DAC_MAX_CLK },	/* 	GT_16BIT	*/
	 {       0,  DAC_MAX_CLK },	/* 	GT_24BIT	*/
#ifdef _ICS_5300_
	 {        0, 0 }}
#else
	 {	  0,  DAC_MAX_CLK }}	/* 	GT_32BIT	*/
#endif
};

#ifndef DAC_RS2

/*
**	support functions
*/

void dac_to_comm(struct kgi_display *dpy)
{
	register int i;
    
	DAC_in8(dpy, DAC_PR_INDEX);

	for (i = 0; i < 4; i++) {

		DAC_in8(dpy, DAC_PIXEL_MASK);
	}
}

void dac_reset(struct kgi_display * dpy)
{
	DAC_in8(dpy, DAC_PR_INDEX);
}

#endif /* DAC_RS2 */

/*
**	GGI interface
*/

int kgim_ramdac_init(struct kgi_card *card,struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev " DRIVER_REV "\n");
	card->ramdac = &ramdac;
	return EOK;
}

void kgim_ramdac_done(struct kgi_display * dpy)
{
	DEBUG0(DRIVER_NAME " removed");
}

#define CLKDIV (mode->clk.div)
#define CLKMUL (mode->clk.mul)

int kgim_ramdac_check_mode (struct kgi_display *dpy,
			    struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
	int mul = 1, div = 1;
	int newclock = mode->dclk;
	
	switch(mode->request.graphtype) {
		
	case GGI_AUTO:		/* default to 8bpp */
		mode->request.graphtype = GT_8BIT;
		
	case GT_TEXT16:		/* 8 bit pixelbus */
	case GT_1BIT:
	case GT_4BIT:
	case GT_8BIT:
		break;

#ifdef _ICS_5300_
	case GT_15BIT:
        case GT_16BIT:
                mul=2;		/* 8 bit pixelbus */
                break;
#else  
	case GT_15BIT:		/* 16 bit pixelbus */
	case GT_16BIT:
		break;
#endif

	case GT_24BIT:		/* 8 bit pb */
		mul = 3; 
		break;

#ifndef _ICS_5300_	
	case GT_32BIT:		/* 16 bit pb */
		mul = 2; 
		break;
#endif

	default:
		DEBUG0("unknown graphtype requested");
		return -E(RAMDAC, UNKNOWN);
	}

	switch (cmd) {

	case CMD_PROPOSE:
		mode->clk.mul = mode->lclk.mul = mul;
		mode->clk.div = mode->lclk.div = div;
		mode->flags &= ~FL_DAC_2XCLOCK;

		if (mode->request.graphtype == GT_15BIT ||
			mode->request.graphtype == GT_16BIT ||
			mode->request.graphtype == GT_24BIT ||
			mode->request.graphtype == GT_32BIT) {

			mode->flags |= FL_DAC_GAMMA;
		}
		return EOK;
	
	case CMD_LOWER:
		if (newclock * CLKMUL / CLKDIV < ramdac.normal.min) {

			DEBUG0("DCLK too low (%i MHz)", newclock);
			return -E(RAMDAC, CLK_LOW_LIMIT);
		}

		if (newclock * CLKMUL/CLKDIV > ramdac.normal.max) {

			mode->dclk = ramdac.normal.max / CLKMUL * CLKDIV;
		}
		return EOK;
	
	case CMD_RAISE:
		if (newclock * CLKMUL / CLKDIV > ramdac.normal.max) {

			DEBUG0("DCLK too high (%i MHz)", newclock);
			return -E(RAMDAC, CLK_HIGH_LIMIT);
		}

		if (newclock * CLKMUL / CLKDIV < ramdac.normal.min) {

			mode->dclk = ramdac.normal.min / CLKMUL * CLKDIV;
		}
		return EOK;
	
	case CMD_CHECK:
		if(mode->lclk.mul != mul || mode->lclk.div != div) {
			DEBUG0("invalid clk or lclk ratio: %i:%i", \
				mode->lclk.mul, mode->lclk.div);
#if 0			/* not sure if this is really a problem ??? */
			return -E(RAMDAC, MODE_CHECK);
#endif
		}

		if (newclock * CLKMUL / CLKDIV < ramdac.normal.min ||
			newclock * CLKMUL / CLKDIV < ramdac.normal.min) {

			DEBUG0(" %i MHz DCLK out of limits", newclock);
			return -E(RAMDAC, MODE_CHECK);
		}
		return EOK;
	
	default:
		INTERNAL_ERROR;
		return -E(RAMDAC,BOGUS);
	}
}

#undef CLKMUL
#undef CLKDIV

void kgim_ramdac_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	int cmd = DAC_MODE_0;	/* default power up mode */

	switch(mode->request.graphtype) {	

	case GT_1BIT:
	case GT_4BIT:
	case GT_TEXT16:	
	case GT_8BIT:
		break;

#ifdef _ICS_5300_
	case GT_15BIT:
		cmd = DAC_MODE_1;
		break;

	case GT_16BIT:
		cmd = DAC_MODE_2;
		break;
#else
	case GT_15BIT:
		cmd = DAC_MODE_5;
		break;
	
	case GT_16BIT:
		cmd = DAC_MODE_6;
		break;
#endif	
	case GT_24BIT:
		cmd = DAC_MODE_3;
		break;
#ifndef _ICS_5300_	
	case GT_32BIT:
		cmd = DAC_MODE_7;
		break;
#endif	
	default:
		INTERNAL_ERROR;
		break;
	}

	DAC_out8(dpy, 0xFF, DAC_PIXEL_MASK);

	#ifdef DAC_RS2
		/*	the chipset uses 3 register select lines
		*/
		DAC_out8(dpy, cmd, DAC_COMMAND);
	#else
		/*	the chipset uses only 2 register select lines
		*/
		dac_to_comm(dpy);
		DAC_out8(dpy, cmd, DAC_PIXEL_MASK);
		dac_reset(dpy);
	#endif /* DAC_RS2 */

	dpy->set_clut = dac_set_clut;
}

int kgim_ramdac_command (struct kgi_graphic *graph, ggi_uint cmd, ggi_uint arg)
{
	ggi_clut clut;
    
	switch(cmd) { 
	
	case RAMDAC_SETCLUT:
		clut.size = 256;
		clut.data = (ggi_color *)(graph->io_buf);
		dac_set_clut(graph->dev.dpy, &clut, 0, 256);
		return EOK;
	
	case RAMDAC_GETCLUT:
		clut.size = 256;
		clut.data = (ggi_color *)(graph->io_buf);
		graph->io_res = graph->io_buf;
		dac_get_clut(graph->dev.dpy, &clut, 0, 256);
		return EOK;
	}
	return -E(RAMDAC,NOSUP_ALWAYS_CANT);  /* This driver can't do this. */
}
