/* --------------------------------------------------------------------------
**	SC 15025 / SC15026 ramdac driver
** --------------------------------------------------------------------------
**
**	Copyright (C) 1996-1997 by Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
**	This code was derived from
**	[1]	Sierra Semiconductor Standard Products Databook, Sept 1992
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Steffen_Seeger
#define	DRIVER_NAME	"SC 15025/SC 15026 DAC driver"
#define	DRIVER_REV	"$Revision: 1.7 $"

#define	SC1502x_125	0		/* 125 MHz version	*/
#define	SC1502x_110	1		/* 110 MHz version	*/
#define	SC1502x_80	2		/*  80 MHz version	*/
#define	SC1502x_66	3		/*  66 MHz verison	*/

#ifndef SC_VERISON
	#define	SC_VERSION	SC1502x_110	/* default version to assume  */
#endif

/* ---------------- no user serviceable code below this line ---------------- */

#ifdef SC15026

	#define	SC1502x	"SC 15026"
#else
	#define	SC1502x	"SC 15025"
#endif

#define	DAC_PRECISION	8	/* we can do 8 bit per RGB value	*/

#include <ggi/module.h>
#include "SC/1502x.h"
#include "IBM/clut.inc"

#define	SUPPORTED (GM_TEXT16 | GM_4BIT | GM_8BIT | \
	GM_15BIT | GM_16BIT | GM_24BIT | GM_32BIT)

static struct kgi_ramdac dacinfo[4] =
{
	{
		"Sierra Semiconductor",		/* manufacturer		*/
	 	SC1502x " - 125",		/* model		*/
 		SUPPORTED,			/* supported modes	*/
	 	FL_DAC_8BIT | FL_DAC_GAMMA | 	/* flags		*/
		FL_DAC_2XCLOCK,
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/

		 DAC_8BIT_PB,  			/* device type		*/
		  {       0, 125000000 },	/* normal mode		*/
		  {       0,  80000000 },	/* 2xclk mode		*/

		 {{	  0, 125000000 },	/* 	GT_TEXT16	*/
		  {       0,         0 },	/*	GT_TEXT32	*/
		  {       0, 125000000 },	/* 	GT_1BIT		*/
		  {	  0, 125000000 },	/* 	GT_4BIT		*/
		  {       0, 125000000 },	/* 	GT_8BIT		*/
		  {	  0,  80000000 },	/* 	GT_15BIT	*/
		  { 	  0,  80000000 },	/* 	GT_16BIT	*/
		  {       0,  41666666 },	/* 	GT_24BIT	*/
		  {	  0,  40000000 }}	/* 	GT_32BIT	*/
        },
	{
		"Sierra Semiconductor",		/* manufacturer		*/
	 	SC1502x " - 110",		/* model		*/
 		SUPPORTED,			/* supported modes	*/
	 	FL_DAC_8BIT | FL_DAC_GAMMA |  	/* flags		*/
		FL_DAC_2XCLOCK,
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/

		 DAC_8BIT_PB,  			/* device type		*/
		  {       0, 110000000 },	/* normal mode		*/
		  {       0,  65000000 },	/* 2xclk mode		*/

		 {{	  0, 110000000 },	/* 	GT_TEXT16	*/
		  {	  0,         0 },	/*	GT_TEXT32	*/
		  {       0, 110000000 },	/* 	GT_1BIT		*/
		  {	  0, 110000000 },	/* 	GT_4BIT		*/
		  {       0, 110000000 },	/* 	GT_8BIT		*/
		  {	  0,  65000000 },	/* 	GT_15BIT	*/
		  { 	  0,  65000000 },	/* 	GT_16BIT	*/
		  {       0,  36666666 },	/* 	GT_24BIT	*/
		  {	  0,  32500000 }}	/* 	GT_32BIT	*/
        },
	{
		"Sierra Semiconductor",		/* manufacturer		*/
	 	SC1502x " - 80",		/* model		*/
 		SUPPORTED,			/* supported modes	*/
	 	FL_DAC_8BIT | FL_DAC_GAMMA | 	/* flags		*/
		FL_DAC_2XCLOCK,
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/

		 DAC_8BIT_PB,  			/* device type		*/
		  {       0,  80000000 },	/* normal mode		*/
		  {       0,  50000000 },	/* 2xclk mode		*/

		 {{	  0,  80000000 },	/* 	GT_TEXT16	*/
		  {       0,         0 },	/*	GT_TEXT32	*/
		  {       0,  80000000 },	/* 	GT_1BIT		*/
		  {	  0,  80000000 },	/* 	GT_4BIT		*/
		  {       0,  80000000 },	/* 	GT_8BIT		*/
		  {	  0,  50000000 },	/* 	GT_15BIT	*/
		  { 	  0,  50000000 },	/* 	GT_16BIT	*/
		  {       0,  26666666 },	/* 	GT_24BIT	*/
		  {	  0,  25500000 }}	/* 	GT_32BIT	*/
        },
	{
		"Sierra Semiconductor",		/* manufacturer		*/
	 	SC1502x " - 66",		/* model		*/
 		SUPPORTED,			/* supported modes	*/
	 	FL_DAC_8BIT | FL_DAC_GAMMA | 	/* flags		*/
		FL_DAC_2XCLOCK,
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/

		 DAC_8BIT_PB,  			/* device type		*/
		  {       0,  66000000 },	/* normal mode		*/
		  {       0,  66000000 },	/* 2xclk mode		*/

		 {{	  0,  66000000 },	/* 	GT_TEXT16	*/
		  {       0,         0 },	/*	GT_TEXT32	*/
		  {       0,  66000000 },	/* 	GT_1BIT		*/
		  {	  0,  66000000 },	/* 	GT_4BIT		*/
		  {       0,  66000000 },	/* 	GT_8BIT		*/
		  {	  0,  50000000 },	/* 	GT_15BIT	*/
		  { 	  0,  50000000 },	/* 	GT_16BIT	*/
		  {       0,  22000000 },	/* 	GT_24BIT	*/
		  {	  0,  25000000 }}	/* 	GT_32BIT	*/
         }
};

static struct kgi_ramdac *ramdac = dacinfo + SC_VERSION;  /* !!! obsolete */

/*	I/O routines to access indexed extended registers.
**	NOTE: ERPF must be set to one before calling inbEXT, outbEXT !
*/

static inline uint8 EXT_in8(struct kgi_display *dpy, uint8 ggiindex)
{
	DAC_out8(dpy, ggiindex, DAC_E_INDEX_R);
	return DAC_in8(dpy, DAC_E_DATA);
}

static inline void EXT_out8(struct kgi_display *dpy, uint8 value, uint8 ggiindex)
{
	DAC_out8(dpy, ggiindex, DAC_E_INDEX_W);
	DAC_out8(dpy, value, DAC_E_DATA); 
}

static void CMD_out8(struct kgi_display *dpy, uint8 value)
{
	ggi_uint i;

	DAC_in8(dpy, DAC_PW_INDEX);	/* reset the IPF	*/

	for (i = 5; --i; ) {		/* set the IPF		*/

		DAC_in8(dpy, DAC_PIXEL_MASK);
	}

	DAC_out8(dpy, value, DAC_PIXEL_MASK);
}

static uint8 CMD_in8(struct kgi_display *dpy)
{
	ggi_uint i;

	DAC_in8(dpy, DAC_PW_INDEX);	/* reset the IPF	*/

	for (i = 5; --i; ) {		/* set the IPF		*/

		DAC_in8(dpy, DAC_PIXEL_MASK);
	}

	return DAC_in8(dpy, DAC_PIXEL_MASK);
}

/*
**	true color CLUT operations
*/

static void dac_set_hclut(struct kgi_display *dpy, ggi_clut *clut,
	ggi_uint first, ggi_uint cnt)
{
	register ggi_sint len = cnt;
	ggi_color *data = clut->data + first;

	DAC_out8(dpy, first >> 2, DAC_PW_INDEX);

	while (len > 0) {

		DAC_out8(dpy, data->r >> SHIFT, DAC_P_DATA);
		DAC_out8(dpy, data->g >> SHIFT, DAC_P_DATA);
		DAC_out8(dpy, data->b >> SHIFT, DAC_P_DATA);

		data += 4;
		len  -= 4;
	}
}

#if 0
static int dac_get_hclut(struct kgi_display *dpy, ggi_clut *clut,
	ggi_uint first, ggi_uint cnt)
{
	ggi_uint len = cnt;
	ggi_color *data = clut->data + first;

	first &= 0xFC;
	len -= (4 - (first & 3)) & 3;

	DAC_out8(dpy, first >> 2, DAC_PR_INDEX);

	while (len > 0) {

		ggi_uint i;
		register uint16 col;

		col = DAC_in8(dpy, DAC_P_DATA);	data->r = col*MAGIC;
		col = DAC_in8(dpy, DAC_P_DATA);	data->g = col*MAGIC;
		col = DAC_in8(dpy, DAC_P_DATA);	data->b = col*MAGIC;

		for (i = 5; --i; ) {

			data[i] = *data;
		}

		data += 4;
		len  -= 4;
	}

	return len;
}
#endif

/*
**	KGI module interface
*/

int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	uint8 pix = DAC_in8(dpy, DAC_PIXEL_MASK);
	uint8 cmd;

	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	DAC_out8(dpy, 0xFF, DAC_PIXEL_MASK);

	cmd = CMD_in8(dpy);
	CMD_out8(dpy, cmd | DAC_CMD_ERPF);

	if (EXT_in8(dpy, DAC_EXT_ID0) == 'S') {

		if (((EXT_in8(dpy, DAC_EXT_ID1) << 8) |
			EXT_in8(dpy, DAC_EXT_ID2)) == 15025) {

			DEBUG1("%s %s rev %c\n", ramdac->manufact, \
				ramdac->model, EXT_in8(dpy, DAC_EXT_ID3));
			EXT_out8(dpy, DAC_AUX_8BIT, DAC_EXT_AUXCTRL);
			DAC_out8(dpy, cmd, DAC_E_COMMAND);
			return EOK;
		}
	}

	DAC_out8(dpy, pix, DAC_PIXEL_MASK);	/* restore pixelmask	*/
	DAC_out8(dpy, DAC_in8(dpy, DAC_PW_INDEX), DAC_PW_INDEX);
						/* reset adress counter	*/

	NOTICE("no supported DAC detected");
	return -E(RAMDAC, UNKNOWN);
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " removed.");
}

int kgim_ramdac_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
	enum kgi_tm_cmd cmd)
{
	ggi_uint maxpclk, pclk;
  
	switch(mode->request.graphtype) {

	case GT_TEXT16:
	case GT_1BIT:
	case GT_4BIT:
	case GT_8BIT:

		switch(cmd) {

		case CMD_PROPOSE:
			mode->lclk.mul = mode->clk.mul = 1;
			mode->lclk.div = mode->clk.div = 1;
			break;

		case CMD_CHECK:
			if ((mode->lclk.mul != 1) || (mode->lclk.div != 1) ||
				(mode->clk.mul != 1) || (mode->clk.div != 1)) {

				return -E(RAMDAC, UNKNOWN);
			}
			break;

			default:
		}
		maxpclk = ramdac->normal.max;
		break;

	case GT_15BIT:
	case GT_16BIT:
		switch(cmd) {

		case CMD_PROPOSE:
			mode->lclk.mul = mode->clk.mul = 1;
			mode->lclk.div = mode->clk.div = 1;
			break;

		case CMD_CHECK:
			if ((mode->lclk.div != 1) ||
				(mode->lclk.mul != mode->clk.mul) ||
				(mode->lclk.div != mode->clk.div)) {
 
				return -E(RAMDAC, UNKNOWN);
			}
			break;

		default:
		}

		switch(mode->lclk.mul) {

		case 1:
			maxpclk = ramdac->dblclk.max;
			break;

		case 2:
			maxpclk = ramdac->normal.max;
			break;

		default:
			return -E(RAMDAC, UNKNOWN);
		}

		break;

	case GT_24BIT:
		switch(cmd) {

		case CMD_PROPOSE:
			mode->lclk.div = mode->clk.div = 1;
			mode->lclk.mul = mode->clk.mul = 3;
			break;

		case CMD_CHECK:
			if ((mode->lclk.div != 1) || (mode->clk.div != 1) ||
				(mode->lclk.mul != 3) || (mode->clk.mul != 3)) {

				return -E(RAMDAC, UNKNOWN);
			}
			break;

		default:
		}
		maxpclk = ramdac->normal.max;
		break;

	case GT_32BIT:
		switch(cmd) {

		case CMD_PROPOSE:
			mode->lclk.mul = mode->clk.mul = 2;
			mode->lclk.div = mode->clk.div = 1;
			break;

		case CMD_CHECK:
			if ((mode->lclk.div != 1) ||
				(mode->lclk.mul != mode->clk.mul) ||
				(mode->lclk.div != mode->clk.div)) {

				return -E(RAMDAC, UNKNOWN);
			}
			break;

		default:	
		}

		switch(mode->lclk.div) {
		case 2:
			maxpclk = ramdac->dblclk.max;
			break;

		case 4:
			maxpclk = ramdac->normal.max;
			break;

		default:
			return -E(RAMDAC, UNKNOWN);
		}
		break;

	default:
		INTERNAL_ERROR;
		return -E(RAMDAC, UNKNOWN);
	}

	pclk = mode->dclk * mode->lclk.mul / mode->lclk.div;

	DEBUG2("maxpclk = %i Hz, pclk = %i Hz\n", maxpclk, pclk);

	switch(cmd) {

	case CMD_LOWER:
		if (maxpclk < pclk) {

			mode->dclk = maxpclk * mode->lclk.div / mode->lclk.mul;
		}
		return EOK;

	case CMD_RAISE:
		return EOK;

	case CMD_CHECK:
		return ((pclk < 0) || (maxpclk < pclk))
			? -E(RAMDAC, UNKNOWN) : EOK;

	default:
		return EOK;
	}
}

void kgim_ramdac_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	uint8 cmd = DAC_CMD_USECLUT;	/* command register value	*/
        uint8 rep = DAC_REP_MODE0;	/* repack register value	*/

	switch(mode->request.graphtype) {

	case GT_TEXT16:
	case GT_1BIT:
	case GT_4BIT: 
	case GT_8BIT:
		cmd = DAC_CMD_MODE0 /* ??? | DAC_C_NORMALCLK */;
		break;

	case GT_15BIT:
		cmd |= DAC_CMD_MODE1;
		cmd |= (mode->lclk.mul==1) ? DAC_CMD_DBL_CLK : DAC_CMD_NRM_CLK;
		break;

	case GT_16BIT:
		cmd |= DAC_CMD_MODE3;
		cmd |= (mode->lclk.mul==1) ? DAC_CMD_DBL_CLK : DAC_CMD_NRM_CLK;
		break;

	case GT_24BIT:
		cmd |= DAC_CMD_MODE4;
		break;

	case GT_32BIT:
		rep = DAC_REP_MODE1;
		cmd |= DAC_CMD_MODE4;
		cmd |= (mode->lclk.mul==2) ? DAC_CMD_DBL_CLK : DAC_CMD_NRM_CLK;
		break;

	default:
		INTERNALERROR;
		return;
	}

	DEBUG2("graphmode %i, cmd = %.2X, rep = %.2X\n", \
		mode->request.graphtype, cmd, rep);

	CMD_out8(dpy, cmd | DAC_CMD_ERPF);	/* map extended registers */

	EXT_out8(dpy, DAC_AUX_8BIT, DAC_EXT_AUXCTRL);

	EXT_out8(dpy, 0xFF, DAC_EXT_PIX_MASK0);
	EXT_out8(dpy, 0xFF, DAC_EXT_PIX_MASK1);
	EXT_out8(dpy, 0xFF, DAC_EXT_PIX_MASK2);

	EXT_out8(dpy, rep, DAC_EXT_REPACK);
	DAC_out8(dpy, cmd, DAC_E_COMMAND);	/* unmap extended registers */
	DAC_out8(dpy, 0xFF, DAC_PIXEL_MASK);

	dpy->set_clut = ((mode->request.graphtype == GT_15BIT) ||
		(mode->request.graphtype == GT_16BIT))
		? dac_set_hclut : dac_set_clut;
}

int kgim_ramdac_command (struct kgi_graphic *graph, ggi_uint cmd, ggi_uint arg)
{
	/*	this may set 
	**	-  dac_params (take care if 504 or 505 was detected).
	**	-  hardware cursor
	**	-  CLUT (or gamma correction table).
	*/

	ggi_clut clut;

	switch(cmd) {

	case RAMDAC_SETCLUT:
		clut.size = 256;
		clut.data=(ggi_color *)(graph->io_buf);
		dac_set_clut(graph->dev.dpy,&clut,0,256);
		return 0;
	case RAMDAC_GETCLUT:
		clut.size=256;
		clut.data=(ggi_color *)(graph->io_buf);
		graph->io_res=graph->io_buf;
		dac_get_clut(graph->dev.dpy,&clut,0,256);
		return 0;
	}

	/*	This driver can't do this. Maybe a simpler command will do.
	*/
	return -E(RAMDAC,NOSUP_ALWAYS_LOWER);	
}
