/* 
   Graphics library for GGI. Horizontal lines.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>

#include "ggi-dl.h"

/**********************************/
/* draw/get/put a horizontal line */
/**********************************/

int GGIdrawhline(ggi_visual *vis,int x,int y,int w)
{
	/* Clipping */
	if (y<(VIS_GC(vis)->cliptl.y) || y>=(VIS_GC(vis)->clipbr.y)) return 0;
	if (x< (VIS_GC(vis)->cliptl.x)) {
		int diff=(VIS_GC(vis)->cliptl.x)-x;
		x+=diff;
		w-=diff;
	}
	if (x+w>(VIS_GC(vis)->clipbr.x)) {
		w=(VIS_GC(vis)->clipbr.x)-x;
	}
	if (w>0) {
		int i;
		uint32 *fb=(uint32 *)VIS_FB_LINEAR(vis);

		fb+=y*VIS_FB_WIDTH(vis)+x;		
		for (i=0;i<w;i++) 
			*(fb++)=VIS_GC_FGCOLOR(vis);
	}
	return 0;
}

int _GGIdrawhline(ggi_visual *vis,int x,int y,int w)
{
	int i;
	uint32 *fb=(uint32 *)VIS_FB_LINEAR(vis);

	fb+=y*VIS_FB_WIDTH(vis)+x;		
	for (i=0;i<w;i++) 
		*(fb++)=VIS_GC_FGCOLOR(vis);

	return 0;
}

int GGIputhline(ggi_visual *vis,int x,int y,int w,void *buffer)
{ 
	uint32 *mem;

	/* Clipping */
	if (y<(VIS_GC(vis)->cliptl.y) || y>=(VIS_GC(vis)->clipbr.y)) return 0;
	if (x< (VIS_GC(vis)->cliptl.x)) {
		int diff=(VIS_GC(vis)->cliptl.x)-x;
		x+=diff;
		w-=diff;
		buffer=((uint32 *)buffer)+diff;
	}
	if (x+w>(VIS_GC(vis)->clipbr.x)) {
		w=(VIS_GC(vis)->clipbr.x)-x;
	}
	if (w>0) {
		mem =(uint32 *)VIS_FB_LINEAR(vis);
		mem+=y*VIS_FB_WIDTH(vis)+x;
		memcpy(mem,buffer,w*sizeof(uint32));
	}
	return 0;
}

int GGIgethline(ggi_visual *vis,int x,int y,int w,void *buffer)
{ 
	uint32 *mem;

	if (x<0||y<0||x+w>VIS_FB_WIDTH(vis)||y>=VIS_FB_HEIGHT(vis)) return -1;

	mem =(uint32 *)VIS_FB_LINEAR(vis);
	mem+=y*VIS_FB_WIDTH(vis)+x;

	memcpy(buffer,mem,w*sizeof(uint32));

	return 0;
}

