/* 
   Graphics library for GGI. Horizontal lines.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>

#include "ggi-dl.h"

/**********************************/
/* draw/get/put a horizontal line */
/**********************************/

int GGIdrawhline(ggi_visual *vis,int x,int y,int w)
{
	uint8 *fb;
	uint8 fg;
	
	CHECKXYW(vis,x,y,w);

	fb = (uint8 *) VIS_FB_LINEAR(vis)+y*((VIS_FB_WIDTH(vis)+x)>>1);
	fg = (uint8) VIS_GC_FGCOLOR(vis) & 0x0F;
	
	/* x is odd.  Read-modify-write left pixel. */
	if (x & 0x01) {
		*(fb++) = (*fb & 0xF0) | fg;
		w--;
	}
	
	memset(fb, fg, w>>1);

	/* Dangling right pixel. */
	if (w & 0x01) {
		*(fb+(w>>1)) = ( *(fb+(w>>1)) & 0x0F) | (fg << 4);
	}
		
	return EOK;
}

int _GGIdrawhline(ggi_visual *vis,int x,int y,int w)
{
	uint8 *fb;
	uint8 fg;

	fb = (uint8 *) VIS_FB_LINEAR(vis)+y*((VIS_FB_WIDTH(vis)+x)>>1);
	fg = (uint8) VIS_GC_FGCOLOR(vis) & 0x0F;
	
	/* x is odd.  Read-modify-write left pixel. */
	if (x & 0x01) {
		*(fb++) = (*fb & 0xF0) | fg;
		w--;
	}
	
	memset(fb, fg, w>>1);

	/* Dangling right pixel. */
	if (w & 0x01) {
		*(fb+(w>>1)) = ( *(fb+(w>>1)) & 0x0F) | (fg << 4);
	}
		
	return EOK;
}

int GGIputhline(ggi_visual *vis,int x,int y,int w,void *buffer)
{ 
	uint8 *fb;
	uint16 color;
	uint8 *buf = (uint8 *)buffer;

	CHECKXYW(vis,x,y,w);
	
	fb=(uint8 *)VIS_FB_LINEAR(vis)+y*((VIS_FB_WIDTH(vis)+x)>>1);
	
	/* Might as well take the memcpy speed up (multiple byte copy). */
	if (!(x & 0x01)) {
		memcpy(fb,buf,w>>1);
		if (w & 0x01)
			*(fb+(w>>1)) = ( *(fb+(w>>1)) & 0x0F) | (*((uint8 *)buf+(w>>1)) << 4);
		return EOK;
	}
	
	/* x is odd. */
	/* Color is a 12-bit quantity.  Upper 8 bits hold the memory to copy.
	 * Lower 4 bits hold the lower 4 bits of the buffer byte read; we'll
	 * need it next loop. */
	/* Could be optimized for multiple-byte copy. */

	color = *fb >> 4;
	
	while (w-=2 > 0)
	{
		color <<= 8;
		color |= *(buf++);
		*(fb++) = color >> 4;
	}
	
	if (!w) {
		*fb = (color << 4) | (*fb & 0x0F);
	}
	
	return EOK;
}


int GGIgethline(ggi_visual *vis,int x,int y,int w,void *buffer)
{ 
	uint8 *fb;
	uint16 color;
	uint8 *buf = (uint8 *)buffer;

	CHECKXYW(vis,x,y,w);
	
	fb=(uint8 *)VIS_FB_LINEAR(vis)+y*((VIS_FB_WIDTH(vis)+x)>>1);
	
	/* If width is odd, the last 4 bits of the buffer's last
	 * byte will hold an extraneous pixel.  No big deal. */
	
	if (!(x & 0x01)) {
		memcpy(buf, fb, (w>>1) + (w & 0x01));
		return EOK;
	}
	
	/* x is odd. */
	color = *fb & 0x0F;
	
	while (w-=2 > 0)
	{
		color <<= 8;
		color |= *(++fb);
		*(buf++) = color >> 4;
	}
	
	return EOK;
}

