/* 
   Graphics library for GGI. Pixels.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"


/* These magic numbers are calculated like so:
 *   2^14+2^12+..2^4+2^2+1
 *
 *  Where the differece between the powers is equal to
 * the number of bits of color. Pretty spiff stuff...
 */

#if GGI_COLOR_PRECISION == 16
#  define MAGIC_2 21845
#  define MAGIC_3 9362
#endif

#ifndef MAGIC_2
#  error GGI_COLOR_PRECISION has changed - we need new magic numbers!
#endif

/*********************************************
 * Let's pretend we have a 3:3:2 color cube! *
 *********************************************/

/* Return `3:3:2' mapping
 */
ggi_pixel GGImapcolor(ggi_visual *vis,ggi_color *col)
{
	return	((col->r>>(GGI_COLOR_PRECISION-3))<<5)|
		((col->g>>(GGI_COLOR_PRECISION-3))<<2)|
		(col->b>>(GGI_COLOR_PRECISION-2));
}

/* Map `3:3:2' pixel to color
 */
int GGIunmappixel(ggi_visual *vis,ggi_uint pixel,ggi_color *col)
{
	if (pixel>255 || col==NULL) 
		return -1;

	col->b=(pixel & 0x3) * MAGIC_2; 
	col->g=((pixel>>2)&0x7) * MAGIC_3;
	col->r=((pixel>>5)&0x7) * MAGIC_3;

	return EOK;	
}

/* Pack the colors into an array
 */
int GGIpackcolors(ggi_visual *vis,void *outbuf,ggi_color *cols,int len)
{
	uint8 *obuf=(uint8 *)outbuf;
	int i;

	for (i=0;i<len;i++) 
		*(obuf++)=(uint8)(ggiMapColor(vis,(cols++)));

	return 0;
}	
	
/* Unpack into the ggi_color array the values of the pixels
 */
int GGIunpackpixels(ggi_visual *vis,void *outbuf,ggi_color *cols,int len)
{
	uint8 *obuf=(uint8 *)outbuf;
	int i;

	for (i=0;i<len;i++) 
		ggiUnmapPixel(vis,*(obuf++),(cols++));

	return 0;
}	
	
  
