/* 
   Graphics library for GGI.

   Copyright (C) 1998 Brian S. Julin - bri@calyx.com
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"
#include "linmm_banked.h"

/********************************/
/* draw/get/put a vertical line */
/********************************/
int GGIdrawvline(ggi_visual *vis,int x,int y,int h)
{
  unsigned char *pixpt;
  unsigned int w=VIS_FB_WIDTH(vis);
  unsigned int color = VIS_GC_FGCOLOR(vis);

	/* Clipping */
	if (x< (VIS_GC(vis)->cliptl.x) ||
	    x>=(VIS_GC(vis)->clipbr.x)) return EOK;
	if (y< (VIS_GC(vis)->cliptl.y)) {
		int diff=(VIS_GC(vis)->cliptl.y)-y;
		y     +=diff;
		h-=diff;
	}
	if (y+h>(VIS_GC(vis)->clipbr.y)) {
		h=(VIS_GC(vis)->clipbr.y)-y;
	}  

  pixpt = BANKFB + ((y * VIS_FB_WIDTH(vis) + x) << LOGBYTPP);
  
  /* Supports only 8, 16, 32 bit modes -- 
     impossible to pixel nonaligned */ 
  
  for(;h--;pixpt+=w << LOGBYTPP) 
    memcpy(pixpt, &color, 1 << LOGBYTPP);
  
  return EOK;
}

int _GGIdrawvline(ggi_visual *vis,int x,int y,int h)
{
  unsigned char *pixpt;
  unsigned int w=VIS_FB_WIDTH(vis);
  unsigned int color = VIS_GC_FGCOLOR(vis);

  pixpt = BANKFB + ((y * VIS_FB_WIDTH(vis) + x) << LOGBYTPP);
  
  /* Supports only 8, 16, 32 bit modes -- 
     impossible to pixel nonaligned */ 
  
  for(;h--;pixpt+=w << LOGBYTPP) 
    memcpy(pixpt, &color, 1 << LOGBYTPP);
  
  return EOK;
}



int GGIputvline(ggi_visual *vis,int x,int y,int h,void *buff)
{
  unsigned char *pixpt, *buffer=(unsigned char *)buff;
  unsigned int w=VIS_FB_WIDTH(vis);


	/* Clipping */
	if (x< (VIS_GC(vis)->cliptl.x) ||
	    x>=(VIS_GC(vis)->clipbr.x)) return EOK;
	if (y< (VIS_GC(vis)->cliptl.y)) {
		int diff=(VIS_GC(vis)->cliptl.y)-y;
		y     +=diff;
		h-=diff;
	}
	if (y+h>(VIS_GC(vis)->clipbr.y)) {
		h=(VIS_GC(vis)->clipbr.y)-y;
	}  

  pixpt = BANKFB + ((y * VIS_FB_WIDTH(vis) + x) << LOGBYTPP);
  
  /* Supports only 8, 16, 32 bit modes -- 
     impossible to pixel nonaligned */   
  for(;h--;pixpt+=w << LOGBYTPP) { 
    memcpy(pixpt, buffer, 1 << LOGBYTPP);
    buffer += (1 << LOGBYTPP);
  }
  
  return EOK;
}


int _GGIputvline(ggi_visual *vis,int x,int y,int h,void *buff)
{
  unsigned char *pixpt, *buffer=(unsigned char *)buff;
  unsigned int w=VIS_FB_WIDTH(vis);

  pixpt = BANKFB + ((y * VIS_FB_WIDTH(vis) + x) << LOGBYTPP);
  
  /* Supports only 8, 16, 32 bit modes -- 
     impossible to pixel nonaligned */   
  for(;h--;pixpt+=w << LOGBYTPP) { 
    memcpy(pixpt, buffer, 1 << LOGBYTPP);
    buffer += (1 << LOGBYTPP);
  }
  
  return EOK;
}

int GGIgetvline(ggi_visual *vis,int x,int y,int h,void *buff)
{
  unsigned char *pixpt,*buffer=(unsigned char *)buff;
  unsigned int w=VIS_FB_WIDTH(vis);
  
  CHECKXYH(vis,x,y,h);
  
  pixpt = RBANKFB + ((y * VIS_FB_WIDTH(vis) + x) << LOGBYTPP);
  
  /* Supports only 8, 16, 32 bit modes -- 
     impossible to pixel nonaligned */ 
  
  for(;h--;pixpt+=w << LOGBYTPP) { 
    memcpy(buffer, pixpt, 1 << LOGBYTPP);
    buffer += (1 << LOGBYTPP);
  }
  return EOK;
}

