#ifndef __KERNEL__
#warning This is a device. MUST be compiled in __KERNEL__ context.
#endif

#define CLASSNAME "devevent"
#define MAXMINOR 32
#define READBUFSIZE 8192

static int major=61;	/* 61 is reserved for experimental use. */

#undef DEBUG
#ifdef DEBUG
#define DEBUG_LEVEL 2
#endif

#undef DEBUG
#include "evstack.h"

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
#include <asm/uaccess.h>
#else
#include <asm/segment.h>
#endif

#include <linux/sched.h>

#define SIGNAL_OCCURED  (current->signal & ~current->blocked)

typedef struct {
	int refcount;
	int minor;
	int wrbufpos;
	unsigned char wrbuffer[256];
	ggi_event *sendpage;
	int rdmask,crdmask;
	unsigned char *rdhead,*rdtail;
	unsigned char *rcvpage;
	int overflow;
	struct wait_queue *waiting_for_event;
} devevent_state;

static devevent_state defaultstate=
{	1,
	-1,
	0,
	"",
	NULL,
	0xffffffff,0,	/* Default to reading all unconsumed events */
	NULL,NULL,
	NULL,
	0,
	NULL
};

static evstack *minortable[MAXMINOR]={NULL,};

static int bind_minor(evstack *stk)
{
	int x;
	for(x=0;x<MAXMINOR;x++)
		if (minortable[x]==NULL)
		{
			minortable[x]=stk;return x;
		}
	return -1;
}

static ev_id devevent_id;
static int  devevent_use=0;

static int  devevent_open(evstack *blank);
static int  devevent_close(evstack *slf);
static void devevent_handler(evstack *slf,ggi_event *ev,int *left);
static int  devevent_uses(evstack *slf);

static evstack_operations devevent_ops={
	devevent_open,
	devevent_close,
	devevent_handler,
	devevent_uses,
	NULL,
	NULL
};

static int devevent_open(evstack *blank)
{
	devevent_state *des;

	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	if ((des=blank->private))
	{
		des->refcount++;
		DEBUG2("Stack cloned (minor %d).",des->minor);
	}
	else
	{	
		int minor;
	
	        des=blank->private=ev_malloc(sizeof(devevent_state));
                if ((minor=bind_minor(blank))<0)
                {
                	ev_free(des); return -ENODEV;
                }
                memcpy(des,&defaultstate,sizeof(devevent_state));
                des->sendpage=ev_evpage_alloc();
                des->rdhead=des->rdtail=des->rcvpage=ev_malloc(READBUFSIZE);
                des->minor=minor;
		DEBUG2("Stack created (minor %d).",des->minor);
	}
	sprintf(blank->name,"/dev/event v1.0 (inst. %.4x min.%2d)",
		blank->id&EV_ID_MASK,des->minor);
	
	MOD_INC_USE_COUNT;
	devevent_use++;

	return 0;
}



static int devevent_close(evstack *slf)
{
	devevent_state *des=slf->private;
        
	if (--des->refcount==0)		/* No more instances left */
	{
		minortable[des->minor]=NULL;	/* Hang up the device */
	        ev_evpage_free(des->sendpage);
	        ev_free(des->rcvpage);
		ev_free(des);
		DEBUG2("Stack destroyed (minor %d).",des->minor);
	}
	MOD_DEC_USE_COUNT;
	devevent_use--;

	return 0;
}

	
/* /dev/event module. Relays all commands to userspace.
 */
static void devevent_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	int size;
	unsigned char *head,*evptr;
	devevent_state *des=slf->private;

	DEBUG2("Stack handler (minor %d).",des->minor);
	for(;ev->size;ev=ev_evpage_next(ev))       /* Loop through all events in page */
	{
		/* We are interested in sending cooked data to the user. */

		if ( ( des->rdmask & (1<<ev->any.type) ) ||
		     ( ev_is_consumed(ev) && 
		       ( des->crdmask & (1<<ev_unconsumed_type(ev)) ) ) )
		{
			size=ev->any.size;head=des->rdhead;evptr=(unsigned char *)ev;
			while( size-- && ! des->overflow )
			{
				*head++=*evptr++;
				if ( head == des->rcvpage+READBUFSIZE )
					head=des->rcvpage;

				if ( head == des->rdtail ) 
				{
					DEBUG2("Overflow /dev/event%d.\n",des->minor);
					des->overflow=1;
					break;
				}
			}
			if (!des->overflow) {
			
				des->rdhead=head;
				wake_up_interruptible(&(des->waiting_for_event));
			}
		}
	}
}

static int devevent_uses(evstack *stack)
{
	return devevent_use;
}

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
static long event_write(struct inode *inode, struct file *file,
			const char *buf, unsigned long count)
#else
static int event_write(struct inode * inode, struct file * file, const char * buf, int count)
#endif
{
	unsigned int minor = MINOR(inode->i_rdev);
	evstack *slf=NULL;
	devevent_state *des;
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
	long long cntup=0;
#else
	int cntup=0;
#endif
	
	if ( minor>=0 && minor <MAXMINOR ) slf=minortable[minor];
	if (!slf) return -ENODEV;	/* Minor not yet allocated */
	des=slf->private;

	/* WARNING - we rely on the fact, that event-size in an unsigned char located */
	/* at the first byte of an event. */
	while(count)
	{
		if ( des->wrbufpos >= des->wrbuffer[0] ) 
			des->wrbufpos=0;	/* This protects against overflow, too.
						 * Actually - "This can't happen ...",
						 * but we use it for resetting, too. */

		if ( count && des->wrbufpos==0 )
		{
			get_user(des->wrbuffer[des->wrbufpos++],buf++);count--;cntup++;
		}
		while( count && des->wrbufpos < des->wrbuffer[0] )
		{
			get_user(des->wrbuffer[des->wrbufpos++],buf++);count--;cntup++;
		}

		if ( des->wrbufpos >= des->wrbuffer[0] )	/* We have a complete event. Execute it. */
		{
			/* !!! FIXME - force minimum size ... */
			if (slf->report_to)
				ev_send(des->sendpage,NULL,(ggi_event *)des->wrbuffer,slf->report_to);
		}
		
		if (need_resched)
			schedule();
	}

	DEBUG2("Event write (minor %d,%d).",minor,des->minor);

	return cntup;
}

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
static long event_read(struct inode *inode, struct file *file,
			char *buf, unsigned long cnt)
#else
static int event_read(struct inode *inode, struct file *file,
			char *buf, int cnt)
#endif
{
	unsigned int minor = MINOR(inode->i_rdev);
	evstack *slf=NULL;
	devevent_state *des;
	int cntup=0;
	
	if ( minor>=0 && minor <MAXMINOR ) slf=minortable[minor];
	if (!slf) return -ENODEV;	/* Minor not yet allocated */
	des=slf->private;

	while(cnt)
	{
		current->state=TASK_INTERRUPTIBLE;
	     	while( cnt && des->rdhead != des->rdtail )
		{
			put_user(*des->rdtail++,buf++);
			cnt--;cntup++;
			if ( des->rdtail == des->rcvpage + READBUFSIZE )
				des->rdtail = des->rcvpage;
		}
		if ( SIGNAL_OCCURED || (file->f_flags & O_NONBLOCK) ) {
			if (cntup) des->overflow=0;
			return cntup;
		} else if (cnt) interruptible_sleep_on(&(des->waiting_for_event));
	}
	DEBUG2("Event read (minor %d,%d).",minor,des->minor);
	
	des->overflow=0;
	return cntup;
}

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
static long long event_lseek(struct inode *inode, struct file *file,
			long long offset, int origin)
#else
static int event_lseek(struct inode * inode, struct file * file,
		    off_t offset, int origin)
#endif
{
	unsigned int minor = MINOR(inode->i_rdev);
	evstack *slf=NULL;
	devevent_state *des;
	
	if ( minor>=0 && minor <MAXMINOR ) slf=minortable[minor];
	if (!slf) return -ENODEV;	/* Minor not yet allocated */
	des=slf->private;

	DEBUG2("Event lseek (minor %d,%d).",minor,des->minor);
	
	/* Seeking is not allowed. However we will resync, if someone tries. */
	des->wrbufpos=0;des->rdtail=des->rdhead;
	return -ESPIPE;
}

static int event_open(struct inode * inode, struct file * file)
{
	unsigned int minor = MINOR(inode->i_rdev);
	evstack *slf=NULL;
	devevent_state *des;
	
	if ( minor>=0 && minor <MAXMINOR ) slf=minortable[minor];
	if (!slf) return -ENODEV;	/* Minor not yet allocated */
	des=slf->private;
	
	des->wrbufpos=0;des->rdtail=des->rdhead;	/* resync */
	MOD_INC_USE_COUNT;

	DEBUG2("Event open (minor %d,%d).",minor,des->minor);

	return 0;
}

static void event_release(struct inode * inode, struct file * file)
{
	unsigned int minor = MINOR(inode->i_rdev);
	evstack *slf=NULL;
	
	MOD_DEC_USE_COUNT;

	if ( minor>=0 && minor <MAXMINOR ) slf=minortable[minor];
	if (!slf) return;	/* -ENODEV, Minor not yet allocated */

	DEBUG2("Event close (minor %d,%d).",minor,des->minor);
}


static int event_ioctl(struct inode *inode, struct file *file,
		    unsigned int cmd, unsigned long arg)
{
	unsigned int minor = MINOR(inode->i_rdev);
	evstack *slf=NULL;
	int retval = 0;
	
	if ( minor>=0 && minor <MAXMINOR ) slf=minortable[minor];
	if (!slf) return -ENODEV;	/* Minor not yet allocated */

	DEBUG2("Event close (minor %d,%d).",minor,des->minor);

	switch ( cmd ) {
		default:
			retval = -EINVAL;
	}
	return retval;
}

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,26)
static int event_select(struct inode *inode, struct file *file, int type,
	select_table *wait)
{
	unsigned int minor = MINOR(inode->i_rdev);
	evstack *slf=NULL;
	
	if ( minor>=0 && minor <MAXMINOR ) slf=minortable[minor];
	if (!slf) return -ENODEV;	/* Minor not yet allocated */

	switch (type) {

	case SEL_IN:
		/* !!!	Implement input event filtering per file! */
		if (des->rdtail != des->rdhead) {

			return 1;
		}
		select_wait(&(des->waiting_for_event), wait);
		return 0;

	case SEL_OUT:
	default:
		return 0;
	}
}
#else
static unsigned int event_select(struct file *file, poll_table *wait)
{
#if 0
	unsigned int minor = MINOR(inode->i_rdev);
	evstack *slf=NULL;
	
	if ( minor>=0 && minor <MAXMINOR ) slf=minortable[minor];
	if (!slf) return -ENODEV;	/* Minor not yet allocated */

	/* !!!	Implement input event filtering per file! */
	poll_wait(&(des->waiting_for_event), wait);
	if (des->rdtail != des->rdhead) {
			return POLLIN | POLLRDNORM;
	}
#endif
	return 0;
}
#endif /* v2.1.26 */

static struct file_operations event_fops = {
	event_lseek,
	event_read,
	event_write,
	NULL,		/* event_readdir */
	event_select,	/* event_select */
	event_ioctl,
	NULL,		/* event_mmap */
	event_open,
	event_release
};

#ifdef MODULE
#define init_devevent init_module
#endif

int init_devevent(void)
{ 
	if (register_chrdev(major,CLASSNAME,&event_fops)) {
		printk(__FILE__": unable to get major %d\n", major);
		return -EIO;
	}

	devevent_id=ev_register_class(CLASSNAME,&devevent_ops);
	printk("Registered %s.\n",CLASSNAME);

	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	unregister_chrdev(major,CLASSNAME);
	ev_unregister_class(devevent_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif

