#include <ggi/libggi.h>
#include "evstack.h"

typedef struct scroller_state
{	int curx,cury;
	int width,height;
	ggi_visual_t visual;
	ggi_color bg_cache,fg_cache;
	ggi_uint bg_ccol,fg_ccol;
	int in_bg;
	int mousex,mousey;
} scroller_state;

static scroller_state defaultstate = { 
	0,0,
	40,25,
	NULL,
	{0,0,0},{0,0,0},
	0,0,
	1,
	0,0
};

/* A sample graphmode scroller routine. This one uses GGI.
 */

static int graphscroll_use;
static int graphscroll_open(evstack *blank);
static int graphscroll_close(evstack *slf);
static void graphscroll_handler(evstack *slf,ggi_event *ev,int *left);
static int graphscroll_uses(evstack *slf);
static int graphscroll_write(evstack *slf, char *str);
static int graphscroll_read(evstack *slf, int lineno, int maxlen, char *str);

static evstack_operations graphscroll_ops = {
	graphscroll_open,
	graphscroll_close,
	graphscroll_handler,
	graphscroll_uses,
	graphscroll_write,
	graphscroll_read
};

/* FIXME ! Why doesn't this work directly ? Move it to GGI */
static void ggiMYCopyBox(ggi_visual_t visual,int x,int y,int width,int height,
		int targetx,int targety)
{
	int yy;
	char buffer[1024];
	
	width<<=3;height<<=3;x<<=3;y<<=3;targetx<<=3;targety<<=3;

	/**/ if (targety<=y)
	for(yy=0;yy<height;yy++)
	  	{
	  		ggiGetHLine(visual,x      ,y      +yy,width,&buffer);
	  		ggiPutHLine(visual,targetx,targety+yy,width,&buffer);
 	  	}
	else 
	for(yy=height-1;yy>=0;yy--)
	  	{
	  		ggiGetHLine(visual,x      ,y      +yy,width,&buffer);
	  		ggiPutHLine(visual,targetx,targety+yy,width,&buffer);
	  	}
}

static int graphscroll_open(evstack *blank)
{
	scroller_state *dat;

	sprintf(blank->name,"GraphScroll v1.0 (instance %.4x)",
		blank->id & EV_ID_MASK);

	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	dat=(scroller_state *)ev_malloc(sizeof(scroller_state));
	memcpy(dat,&defaultstate,sizeof(scroller_state));
	dat->visual=MASTER_VIS;

	blank->private=dat;

	graphscroll_use++;

	return 0;
}

static int graphscroll_close(evstack *slf)
{
	graphscroll_use--;
	
	return 0;
}


/* Event handler code.
 */
static void graphscroll_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	unsigned int	c;
	ggi_cmd_event *evdat;
	struct scroller_state *ss=slf->private;

	/* remove old HW-cursor */
	if (!ss->in_bg)
	{	ggiGetPixel(ss->visual,(ss->curx<<3)+4,(ss->cury<<3)+7,&c);
		c ^= 0xFF;
		ggiPutPixel(ss->visual,(ss->curx<<3)+4,(ss->cury<<3)+7,c);
		ggiGetPixel(ss->visual,ss->mousex,ss->mousey,&c);
		c ^= 0x55;
		ggiPutPixel(ss->visual,ss->mousex,ss->mousey,c);
	}
	for(;ev->size;ev=ev_evpage_next(ev))       /* Loop through all events in page */
	{
		/* We are interested in translating CmdScrollWriteChar to 
		   scroller-actions */

		switch(ev->any.type)
		{ 
		  case evCommand:
		  	evdat=(ggi_cmd_event *)ev;
		  	switch(evdat->code)
		  	{
		  	  case CmdScrollWriteChar:
		  	  	ev_consume(ev);
				if (ss->in_bg) break;
				DEBUG1("got event %x.",evdat->code);
				{ DataScrollWriteChar *dt=(DataScrollWriteChar *)evdat->data;
				  int x;
				  if (memcmp(&dt->fg_col,&ss->fg_cache,sizeof(ggi_color))) 
				  	ss->fg_ccol=ggiMapColor(ss->visual,ss->fg_cache=dt->fg_col);
				  if (memcmp(&dt->bg_col,&ss->bg_cache,sizeof(ggi_color))) 
				  	ss->bg_ccol=ggiMapColor(ss->visual,ss->bg_cache=dt->bg_col);
				  ggiSetGCForeground(ss->visual,ss->fg_ccol);
				  ggiSetGCBackground(ss->visual,ss->bg_ccol);
				  for(x=0;x<dt->numchr;x++)
				  {
				  	if ( ss->curx < 0 ||
					     ss->curx >= ss->width ||
					     ss->cury < 0 ||
				       	     ss->cury >= ss->height ) break;

				  	DEBUG1("Drawing 0x%x at (%d,%d)",dt->chr[x],ss->curx,ss->cury);
				  	ggiPutc(ss->visual,ss->curx<<3,ss->cury<<3,dt->chr[x]&0xff);
				  	if (dt->attrib&AT_UNDERLINE) ggiDrawHLine(ss->visual,ss->curx<<3,(ss->cury<<3)+7,8);
				  	ss->curx+=dt->dx;ss->cury+=dt->dy;
				  }
				} break;
		  	  case CmdScrollGotoXY:
		  	  	ev_consume(ev);
				DEBUG1("got event %x.",evdat->code);
				{ DataScrollGotoXY *dt=(DataScrollGotoXY *)evdat->data;
				  switch(dt->curnum)
				  { case CUR_NORM:
				  	if ( dt->x < 0) dt->x=0;
					if ( dt->x >= ss->width ) dt->x=ss->width-1;
					if ( dt->y < 0 ) dt->y=0;
					if ( dt->y >= ss->height ) dt->y=ss->height-1;
					ss->curx=dt->x>>dt->fixedpointshift;
					ss->cury=dt->y>>dt->fixedpointshift;
					break;
				    case CUR_MOUSE:
					ss->mousex=(dt->x<<3)>>dt->fixedpointshift;
					ss->mousey=(dt->y<<3)>>dt->fixedpointshift;
				  	if ( ss->mousex < 0) ss->mousex=0;
					if ( ss->mousex >= ss->width*8 ) ss->mousex=ss->width*8-1;
					if ( ss->mousey < 0 ) ss->mousey=0;
					if ( ss->mousey >= ss->height*8 ) ss->mousey=ss->height*8-1;
				  }
				} break;
		  	  case CmdScrollScroll:
		  	  	ev_consume(ev);
				if (ss->in_bg) break;
				DEBUG1("got event %x.",evdat->code);
				{ DataScrollScroll *dt=(DataScrollScroll *)evdat->data;

				  /* Check if box is on-screen. */
				  if ( dt->x < 0 ||
				       dt->x >= ss->width ||
				       dt->y < 0 ||
				       dt->y >= ss->height ) break;

				  /* adjust box. */
				  if ( dt->x + dt->width > ss->width )
				      dt->width=ss->width-dt->x;
				  if ( dt->y + dt->height > ss->height )
				      dt->height=ss->height-dt->x;

				  if ( dt->dx==0 && dt->dy==0 )
				  { 
				    ggiSetGCForeground(ss->visual,0);
					ggiDrawBox(ss->visual,dt->x<<3,dt->y<<3,dt->width<<3,dt->height<<3);
				  }
				  else
				  { ggiMYCopyBox(ss->visual,dt->x,dt->y,
						dt->width,dt->height,
						dt->x+dt->dx,dt->y+dt->dy);
				    ggiSetGCForeground(ss->visual,0);
				    if (dt->dy<0)
					ggiDrawBox(ss->visual,dt->x<<3,(dt->y+dt->height+dt->dy)<<3,dt->width<<3,(-dt->dy)<<3);
				    if (dt->dy>0)
					ggiDrawBox(ss->visual,dt->x<<3,dt->y<<3,dt->width<<3,dt->dy<<3);
				    if (dt->dx<0)
					ggiDrawBox(ss->visual,(dt->x+dt->width+dt->dx)<<3,dt->y<<3,(-dt->dx)<<3,dt->height<<3);
				    if (dt->dx>0)
					ggiDrawBox(ss->visual,dt->x<<3,dt->y<<3,dt->dx<<3,dt->height<<3);
				  }
				  DEBUG1("Scroll (%d,%d,%d,%d -> %d,%d)",
				  	dt->x,dt->y,dt->width,dt->height,dt->dx,dt->dy);

				} break;
		  	  case CmdVtActivate:
		  	  	{ DataVtActivate *dva=(DataVtActivate *)evdat->data;
		  	  	  if (dva->onoff)
				  {
					ggiSetGraphMode(MASTER_VIS,320,200,320,200,GT_8BIT);
				  	/*fromback(ss);*/ss->in_bg=0;
				  } else
				  {
				  	/*toback(ss);*/ss->in_bg=1;
				  	ggiSetGCForeground(ss->visual,0);
				  	ggiFillscreen(ss->visual);
				  }
		  	  	} break;
		  	  default:
		  	  	break;
			}
			break;
		  default: 
		  	break;
		}
	}
	/* Write new HW-cursor */
	if (!ss->in_bg)
	{	ggiGetPixel(ss->visual,(ss->curx<<3)+4,(ss->cury<<3)+7,&c);
		c ^= 0xFF;
		ggiPutPixel(ss->visual,(ss->curx<<3)+4,(ss->cury<<3)+7,c);
		ggiGetPixel(ss->visual,ss->mousex,ss->mousey,&c);
		c ^= 0x55;
		ggiPutPixel(ss->visual,ss->mousex,ss->mousey,c);
	}
}

static int graphscroll_uses(evstack *slf)
{
	return graphscroll_use;
}
                                                        
static int graphscroll_write(evstack *slf, char *str)
{
	return -EINVAL;
}

static int graphscroll_read(evstack *slf,int lineno,int maxlen,char *str)
{
	if (lineno==0) 
	{
		strcpy(str,"No info available."); return 0;
	}
	return -EINVAL;
}
                                                        
void init_graphscroll(void)
{       
	ggiInit();
	if (!MASTER_VIS)
	{	MASTER_VIS=ggiOpen(NULL);
		ggiSetGraphMode(MASTER_VIS,320,200,320,200,GT_8BIT);
	}

	ev_register_class("graphscroll",&graphscroll_ops);
	printk("Registered graphscroll.\n");
}
