#undef DEBUG
#ifdef DEBUG
#define DEBUG_LEVEL 2
#endif

#define __KERNEL__
#ifdef KBD_AMERICAN
#include "kbd-american.c"
#endif
#ifdef KBD_GERMAN
#include "kbd-german.c"
#endif
#undef __KERNEL__

#include "evstack.h"

/* A sample keytable module
 */
#define CLASSNAME "loadkeys"
static ev_id loadkeys_id;
static int  loadkeys_use;

static int  loadkeys_open(evstack *blank);
static int  loadkeys_close(evstack *slf);
static void loadkeys_handler(evstack *slf,ggi_event *ev,int *len);
static int  loadkeys_uses(evstack *slf);

static evstack_operations loadkeys_ops = {
	loadkeys_open,
	loadkeys_close,
	loadkeys_handler,
	loadkeys_uses,
	NULL,
	NULL
};

/* This struct will be the default for all newly allocated keytables.
 */

typedef struct loadkeys_state
{
	MODIFIER_DATA
	struct kgi_keymap *map;
	int refcount;
	int down_mod[20];	/* Would like to kill this one ... */
	int npadch;		/* helper for building ALR-Keypad ASCII*/
} loadkeys_state;

loadkeys_state defaultstate= 
{ 
	0,0,0,0 , 
	&default_keymap, 
	1,
	{0,} 
	,-1 
};

static int loadkeys_open(evstack *blank)
{
	struct loadkeys_state *dat;
	
	sprintf(blank->name,"LoadKeys v1.0 (instance %.4x)",
		blank->id & EV_ID_MASK);

	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";


	if ((dat=blank->private))
	{
		dat->refcount++;
		DEBUG2("Stack cloned (rc %d).",dat->refcount);
	}
	else
	{	
		/* Set up the initial state */
		blank->private=dat=(loadkeys_state *)ev_malloc(sizeof(loadkeys_state));
		memcpy(dat,&defaultstate,sizeof(loadkeys_state));
		dat->map=(struct kgi_keymap *)ev_malloc(sizeof(struct kgi_keymap));
		memcpy(dat->map,defaultstate.map,sizeof(struct kgi_keymap));

		DEBUG2("Stack created (rc %d).",des->refcount);
	}

        MOD_INC_USE_COUNT;
	loadkeys_use++;

	return 0;
}

static int loadkeys_close(evstack *slf)
{
	loadkeys_state *des=slf->private;
        
	if (--des->refcount==0)		/* No more instances left */
	{
		ev_free(des);
		DEBUG2("Stack destroyed.",0);
	}

        MOD_DEC_USE_COUNT;
	loadkeys_use--;
	
	return 0;
}

/* Much of this code was copied from kgi.c (c) Steffen Seeger
 */
 
/*
 *	input handling
 */

static inline void do_special(evstack *slf,ggi_event *event,int *space_left)
{
	if (event->any.type != evKeyPress) {

		return;
	}

	DEBUG2("doing special key %.4x", event->key.sym);

	switch(U(event->key.sym)) {

	case K_ENTER:
		return;

	case K_SH_REGS:
		printk("SHOWREGS ! FIXME !\n");
/*	!!!	show_regs(pt_regs);
*/		return;

	case K_SH_MEM:
		printk("SHOWMEM ! FIXME !\n");
/*		show_mem();
		event->any.type = evNothing;
*/		return;

	case K_SH_STAT:
		printk("SHOWSTATE ! FIXME !\n");
/*		show_state();
		event->any.type = evNothing;
*/		return;

	case K_BREAK:
		return;

	case K_CONS:
/* FIXME ! THIS IS INCONSISTENT. */
/*		f->want_focus = f->last_focus;
*/		return;

	case K_CAPS:
	case K_NUM:
	case K_HOLD:
	case K_SCROLLFORW:
	case K_SCROLLBACK:
		return;

	case K_BOOT:
/*		ctrl_alt_del();
		event->any.type = evNothing;
*/		return;

	case K_CAPSON:
		return;

	case K_COMPOSE:
		return;

	case K_SAK:
/*		if (VALID_DEV_ID(f->focus) && kgidev[f->focus]) {

			do_SAK(kgidev[f->focus]->tty);
		}
*/		return;

	case K_DECRCONSOLE:
/*		{
			register int start, search;

			if (f->focus < 0) {

				start = MAX_NR_DEVICES - 1;
				search = start;

			} else {

				ASSERT(f->focus < MAX_NR_DEVICES);

				search = ((start = f->focus) > 0)
					? f->focus - 1 : MAX_NR_DEVICES - 1;
			}

			while ((search != start) && !kgidev[search] &&
				(focus_map[search] != f->id)) {

				if (--search <  0) {

					search = MAX_NR_DEVICES - 1;
				}
			}

			if (search != start) {

				f->want_focus = search;
			}
		}
*/		return;

	case K_INCRCONSOLE:
/*		{
			register int start, search;

			if (f->focus < 0) {

				start = search = 0;

			} else {

				ASSERT(f->focus < MAX_NR_DEVICES);

				search = ((start = f->focus) < MAX_NR_DEVICES-1)
					? f->focus + 1 : 0;
			}

			while ((search != start) && !kgidev[search] &&
				(focus_map[search] != f->id)) {

				if (++search >= MAX_NR_DEVICES) {

					search = 0;
				}
			}

			if (search != start) {

				f->want_focus = search;
			}
		}
*/		return;
			
	case K_SPAWNCONSOLE:
	case K_BARENUMLOCK:
		return;

	case K_TEXTSCREEN:
/*		if (VALID_DEV_ID(f->focus)
			&& (f->focus >= MAX_NR_CONSOLES)) {

			f->want_focus = f->focus - MAX_NR_CONSOLES;
			mark_bh(KEYBOARD_BH);
		}
*/		return;

	case K_GRAPHSCREEN:
/*		if (VALID_DEV_ID(f->focus)
			&& (f->focus < MAX_NR_CONSOLES)) {

			f->want_focus = f->focus + MAX_NR_CONSOLES;
			mark_bh(KEYBOARD_BH);
		}
*/		return;

	case K_TOGGLESCREEN:
/*		if (VALID_DEV_ID(f->focus)) {

			f->want_focus = (f->focus < MAX_NR_CONSOLES) 
					? f->focus + MAX_NR_CONSOLES
					: f->focus - MAX_NR_CONSOLES;
			mark_bh(KEYBOARD_BH);
		}
*/		return;

	default:
		DEBUG0("unknown special key %.4x", event->key.sym);
	}
}

static inline void do_modifier(evstack *slf,ggi_event *event, int *space_left)
{
	struct loadkeys_state *state=slf->private;
	register ksym = U(event->key.sym);

	DEBUG2("old modifiers: %.2x %.2x %.2x %.2x, modifier key %.4x", \
		state->effect, state->normal, state->locked, state->sticky, ksym);

	/*	The effect of STICKY and LOCKED modifiers is toggled whenever
	**	they are pressed.
	*/
	if ((K_FIRST_STICKY <= ksym) && (ksym <= K_LAST_STICKY)) {

		NOTICE2("sticky modifer key.");

		if (event->key.type == evKeyPress) {

			state->sticky ^= (1 << KVAL(ksym - K_FIRST_STICKY));	
		}
	}

	if ((K_FIRST_LOCKED <= ksym) && (ksym <= K_LAST_LOCKED)) {

		NOTICE2("locked modifier key.");

		if (event->key.type == evKeyPress) {

			state->locked ^= (1 << KVAL(ksym - K_FIRST_LOCKED));
		}
	}

	/*	If a NORMAL modifier is pressed at least once, it goes into
	**	effect and clears the effect of the corresponding LOCKED
	**	modifier. Thus there is no effect in the EFFECTIVE modifers.
	**	The effect of the NORMAL modifier is not cleared until the
	**	last is released (if pressed multiple times), which cleares
	**	the effect of the corresponding LOCKED modifier too.
	*/
	if ((K_FIRST_NORMAL <= ksym) && (ksym <= K_LAST_NORMAL)) {

		register ggi_uint mask = 1 << KVAL(ksym - K_FIRST_NORMAL);

		NOTICE2("normal modifier key.");

		if (event->key.type == evKeyPress) {

			state->down_mod[KVAL(ksym - K_FIRST_NORMAL)]++;
			state->normal |=  mask;
			state->locked &= ~mask;

		} else {

			ASSERT(event->key.type == evKeyRelease);
			ASSERT(state->down_mod[KVAL(ksym - K_FIRST_NORMAL)] > 0);
			if (state->down_mod[KVAL(ksym - K_FIRST_NORMAL)] <= 0)
				state->down_mod[KVAL(ksym - K_FIRST_NORMAL)]=1;

			if (! --(state->down_mod[KVAL(ksym - K_FIRST_NORMAL)])) {

				state->normal &= ~mask;
				state->locked &= ~mask;
			}
		}
	}

	state->effect = state->normal ^ state->locked ^ state->sticky;

	DEBUG2("new modifers: %.2x %.2x %.2x %.2x", state->effect, state->normal,\
		state->locked, state->sticky);
}

static inline void do_ascii(evstack *slf,ggi_event *event, int *space_left)
{
	int base, value;
	struct loadkeys_state *state=(loadkeys_state *)slf->private;

	DEBUG2("doing ascii key %.4x", event->key.sym);

	if (event->key.type == evKeyRelease) {

		return;
	}

	if (KVAL(event->key.sym) < 10) {

		value = KVAL(event->key.sym);
		base = 10;

	} else {

		value = KVAL(event->key.sym) - 10;
		base = 16;
	}

	state->npadch = (state->npadch == -1) ? value : (state->npadch*base + value);

	if (base == 16) {

		state->npadch &= 0xFFFF;

	} else {

		if (state->npadch > 100000) {

			state->npadch %= 100000;
		}
	} 
}

static inline void do_action(evstack *slf,ggi_event *event,int *space_left)
{
	ggi_uint val = KVAL(event->key.sym);
	ggi_uint typ = KTYP(event->key.sym);

	if ((1 << event->any.type) & ~(emKeyPress | emKeyRelease)) {

		return;
	}

	DEBUG2("key %s, code %.2x, sym %.2x", event->key.type == evKeyPress\
		? "down" : "up", event->key.code, event->key.sym);

	if (typ >= 0xF0) {

		typ -= 0xF0;

		switch (typ) {

		case KT_SPEC:
			if (val < K_LAST_SPEC) {

				do_special(slf, event,space_left); 
			}
			return;

		case KT_CONS:
			if (event->key.type == evKeyPress) {
				DataVtSwitch dss;
				char outbuf[256];

				dss.newvt=val;
				build_cmd_event(outbuf,CmdVtSwitch,&dss,sizeof(dss));
				DEBUG2("VtSwitch %d",val);
				ev_evpage_add(event,space_left,(ggi_event *)&outbuf);
			}
			return;

		case KT_SHIFT:
			if (val < K_LAST_MODIFIER) {

				do_modifier(slf, event, space_left);
			}
			return;
	
		case KT_ASCII:
			if (val < K_LAST_ASCII) {

				do_ascii(slf, event, space_left);
			}
			return;
		}
	}
}



static void loadkeys_handler(evstack *slf,ggi_event *ev,int *space_left)
{ 
  struct ggi_key_event *evdat;

  for(;ev->size;ev=ev_evpage_next(ev))	/* Loop through all events in page */
  {
    /* We are only interested in translating raw keycodes. */
    switch(ev->any.type)
    { case evKeyPress:
      case evKeyRelease:
      case evKeyRepeat:
      case evKeyState:

	DEBUG1("got a key: %d\n",ev->key.code);

        evdat=(ggi_key_event *)ev;

        /* modify the event */

	if (evdat->sym==0 && ev->key.code < 0x8000) {

		struct loadkeys_state *st = slf->private;
		struct kgi_keymap *k = st->map;
		unicode *map = (st->effect < k->keymap_size) 
				? k->keymap[st->effect] : NULL;

		if (map) {

			ev->key.sym = ((k->keymin <= ev->key.code) &&
				       (ev->key.code <= k->keymax))
				? map[ev->key.code - k->keymin] : K_VOID;

		} else {

			map = k->keymap[0];

			if ((k->keymin <= ev->key.code) &&
				(ev->key.code <= k->keymax) &&
				(KTYP(U(map[ev->key.code - k->keymin]))
				== KT_SHIFT)) {

				ev->key.sym = map[ev->key.code-k->keymin];

			} else {

				ev->key.sym = K_VOID;
			}
		}

	        DEBUG1("Got rawkey %x. Modified %x.",evdat->code,evdat->sym);
	}
	do_action(slf,ev,space_left);
	
	/* FIXME - we should copy the local modifier information to the event */
        break;
      default: 
        break;
    }
  }
}

static int loadkeys_uses(evstack *stack)
{
	return loadkeys_use;
}

static int loadkeys_write(evstack *stack, char *str)
{
	return -EINVAL;
}

static int loadkeys_read(evstack *slf,int lineno,int maxlen,char *str)
{
	if (lineno==0) 
	{
		strcpy(str,"No info available."); return 0;
	}
	return -EINVAL;
}

#ifdef MODULE
#define init_loadkeys init_module
#endif

int init_loadkeys(void)
{ 
	loadkeys_id=ev_register_class(CLASSNAME,&loadkeys_ops);
	printk("Registered %s.\n",CLASSNAME);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	ev_unregister_class(loadkeys_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif
