#include <ggi/libggi.h>
#include "evstack.h"

typedef struct scroller_state
{	int curx,cury;
	int mousex,mousey;
	int selendx,selendy;
	int selstx,selsty;
	int width,height;
	ggi_visual_t visual;
	short int *backbuffer;
	short int *ggibuffer;
	int in_bg;
} scroller_state;

static scroller_state defaultstate = { 
	0,0,
	-1,-1,
	-1,-1,
	-1,-1,
	80,25,
	NULL,
	NULL,
	NULL,
	1
};

/* A sample usermode scroller routine. This one uses GGI.
 */

static int userscroll_use;
static int userscroll_open(evstack *blank);
static int userscroll_close(evstack *slf);
static void userscroll_handler(evstack *slf,ggi_event *ev,int *left);
static int userscroll_uses(evstack *slf);
static int userscroll_write(evstack *slf, char *str);
static int userscroll_read(evstack *slf, int lineno, int maxlen, char *str);

static evstack_operations userscroll_ops = {
	userscroll_open,
	userscroll_close,
	userscroll_handler,
	userscroll_uses,
	userscroll_write,
	userscroll_read
};

static void toback(struct scroller_state *ss)
{
	memcpy(ss->backbuffer,ss->ggibuffer,2*ss->width*ss->height);
}

static void fromback(struct scroller_state *ss)
{
	memcpy(ss->ggibuffer,ss->backbuffer,2*ss->width*ss->height);
}

static void MYCopyBox(struct scroller_state *ss,short int *ptr,int x1,int y1,int w,int h,int tx,int ty)
{
	int xx,yy;
	if (y1<ty)
	{
		for(yy=h-1;yy>=0;yy--)
			memmove(ptr+tx+(ty+yy)*ss->width,ptr+x1+(y1+yy)*ss->width,2*w);
	}
	else
	{
		for(yy=0;yy<h;yy++)
			memmove(ptr+tx+(ty+yy)*ss->width,ptr+x1+(y1+yy)*ss->width,2*w);
	}
}

static void MYDrawBox(struct scroller_state *ss,short int *ptr,int x1,int y1,int w,int h)
{
	int yy;
	for(yy=0;yy<h;yy++)
		memset(ptr+x1+(y1+yy)*ss->width,0,w*2);
}

static int userscroll_open(evstack *blank)
{
	scroller_state *dat;

	sprintf(blank->name,"UserScroll v1.0 (instance %.4x)",
		blank->id & EV_ID_MASK);

	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	dat=(scroller_state *)ev_malloc(sizeof(scroller_state));
	memcpy(dat,&defaultstate,sizeof(scroller_state));
	dat->visual=MASTER_VIS;
	dat->backbuffer=ev_malloc(dat->width*dat->height*sizeof(short int));
	dat->ggibuffer=ggiGetFB(dat->visual);

	blank->private=dat;

	userscroll_use++;

	return 0;
}

static int userscroll_close(evstack *slf)
{
	scroller_state *dat=slf->private;
	ev_free(dat->backbuffer);

	userscroll_use--;
	
	return 0;
}


/* Event handler code.
 */
static void userscroll_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	ggi_cmd_event *evdat;
	struct scroller_state *ss=slf->private;

	short int *curbuf= (ss->in_bg) ? ss->backbuffer : ss->ggibuffer ;

	/* remove old HW-cursor */
	if (!ss->in_bg)
	{	*(ss->ggibuffer+ss->curx+ss->cury*ss->width) ^= 0xFF00;
		if (ss->mousex>=0)
			*(ss->ggibuffer+ss->mousex+ss->mousey*ss->width) ^= 0x8000;
		if (ss->selendx>=0&&ss->selstx>=0)
		{	short int *p1,*p2,*p3;
			p1=ss->ggibuffer+ss->selendx+ss->selendy*ss->width;
			p2=ss->ggibuffer+ss->selstx +ss->selsty *ss->width;
			for(p3=p1;p3<=p2;p3++) *p3 ^= 0x2000;
			for(p3=p2;p3<=p1;p3++) *p3 ^= 0x2000;
		}
	}
	
	for(;ev->size;ev=ev_evpage_next(ev))       /* Loop through all events in page */
	{
		/* We are interested in translating CmdScrollWriteChar to 
		   scroller-actions */

		switch(ev->any.type)
		{ 
		  case evCommand:
		  	evdat=(ggi_cmd_event *)ev;
		  	switch(evdat->code)
		  	{
		  	  case CmdScrollWriteChar:
		  	  	ev_consume(ev);
				DEBUG1("got event %x.",evdat->code);
				{ DataScrollWriteChar *dt=(DataScrollWriteChar *)evdat->data;
				  int x;
				  for(x=0;x<dt->numchr;x++)
				  {
				  	short int color;
				  	if ( ss->curx < 0 ||
					     ss->curx >= ss->width ||
					     ss->cury < 0 ||
				       	     ss->cury >= ss->height ) break;

				  	DEBUG1("Drawing 0x%x at (%d,%d)",dt->chr[x],ss->curx,ss->cury);
					color=  ((dt->bg_col.r&0x8000)>>1)|
						((dt->bg_col.g&0x8000)>>2)|
						((dt->bg_col.b&0x8000)>>3)|
						((dt->fg_col.r&0x8000)>>5)|
						((dt->fg_col.g&0x8000)>>6)|
						((dt->fg_col.b&0x8000)>>7);
					*(curbuf+ss->curx+ss->cury*ss->width) = dt->chr[x]|color;
				  	ss->curx+=dt->dx;ss->cury+=dt->dy;
				  }
				} break;
		  	  case CmdScrollGotoXY:
		  	  	ev_consume(ev);
				DEBUG1("got event %x.",evdat->code);
				{ DataScrollGotoXY *dt=(DataScrollGotoXY *)evdat->data;
				  switch(dt->curnum)
				  { case CUR_NORM:
					  ss->curx=dt->x>>dt->fixedpointshift;
					  ss->cury=dt->y>>dt->fixedpointshift;
					  if ( ss->curx < 0 ) ss->curx=0;
					  if ( ss->curx >= ss->width ) ss->curx=ss->width-1;
					  if ( ss->cury < 0 ) ss->cury=0;
					  if ( ss->cury >= ss->height ) ss->cury=ss->height-1;
					  break;
				    case CUR_MOUSE:
					  ss->mousex=dt->x>>dt->fixedpointshift;
					  ss->mousey=dt->y>>dt->fixedpointshift;
					  if ( ss->mousex < 0 ||
					       ss->mousex >= ss->width ||
					       ss->mousey < 0 ||
					       ss->mousey >= ss->height )
						  ss->mousex=-1;
					  break;
				    case CUR_SELECTEND:
					  ss->selendx=dt->x>>dt->fixedpointshift;
					  ss->selendy=dt->y>>dt->fixedpointshift;
					  if ( ss->selendx < 0 ||
					       ss->selendx >= ss->width ||
					       ss->selendy < 0 ||
					       ss->selendy >= ss->height )
						  ss->selendx=-1;
					  break;
				    case CUR_SELECTSTART:
					  ss->selstx=dt->x>>dt->fixedpointshift;
					  ss->selsty=dt->y>>dt->fixedpointshift;
					  if ( ss->selstx < 0 ||
					       ss->selstx >= ss->width ||
					       ss->selsty < 0 ||
					       ss->selsty >= ss->height )
						  ss->selstx=-1;
					  break;
				  }
				} break;
		  	  case CmdScrollScroll:
		  	  	ev_consume(ev);
				DEBUG1("got event %x.",evdat->code);
				{ DataScrollScroll *dt=(DataScrollScroll *)evdat->data;

				  /* Check if box is on-screen. */
				  if ( dt->x < 0 ||
				       dt->x >= ss->width ||
				       dt->y < 0 ||
				       dt->y >= ss->height ) break;

				  /* adjust box. */
				  if ( dt->x + dt->width > ss->width )
				      dt->width=ss->width-dt->x;
				  if ( dt->y + dt->height > ss->height )
				      dt->height=ss->height-dt->x;

				  if ( dt->dx==0 && dt->dy==0 )
				  { 
					MYDrawBox(ss,curbuf,dt->x,dt->y,dt->width,dt->height);
				  }
				  else
				  { 
#ifndef SPEED_TEST
				    MYCopyBox(ss,curbuf,dt->x,dt->y,
						dt->width,dt->height,
						dt->x+dt->dx,dt->y+dt->dy);
				    if (dt->dy<0)
					MYDrawBox(ss,curbuf,dt->x,dt->y+dt->height+dt->dy,dt->width,-dt->dy);
				    if (dt->dy>0)
					MYDrawBox(ss,curbuf,dt->x,dt->y,dt->width,dt->dy);
				    if (dt->dx<0)
					MYDrawBox(ss,curbuf,dt->x+dt->width+dt->dx,dt->y,-dt->dx,dt->height);
				    if (dt->dx>0)
					MYDrawBox(ss,curbuf,dt->x,dt->y,dt->dx,dt->height);
#endif
				  }
				  DEBUG1("Scroll (%d,%d,%d,%d -> %d,%d)",
				  	dt->x,dt->y,dt->width,dt->height,dt->dx,dt->dy);

				} break;

		  	  case CmdVtActivate:
		  	  	{ DataVtActivate *dva=(DataVtActivate *)evdat->data;
		  	  	  if (dva->onoff)
				  {
					ggiSetTextMode(MASTER_VIS,80,25,80,25,8,16,GT_TEXT16);
					ss->ggibuffer=ggiGetFB(ss->visual);
				  	fromback(ss);ss->in_bg=0;
				  } else
				  {
				  	toback(ss);ss->in_bg=1;
				  	memset(ss->ggibuffer,0,ss->width*ss->height*2);
				  }
				  curbuf= (ss->in_bg) ? ss->backbuffer : ss->ggibuffer ;
		  	  	} break;

		  	  default:
		  	  	break;
			}
		  default: 
		  	break;
		}
	}
	/* Write new HW-cursor */
	if (!ss->in_bg)
	{	*(ss->ggibuffer+ss->curx+ss->cury*ss->width) ^= 0xFF00;
		if (ss->mousex>=0)
			*(ss->ggibuffer+ss->mousex+ss->mousey*ss->width) ^= 0x8000;
		if (ss->selendx>=0&&ss->selstx>=0)
		{	short int *p1,*p2,*p3;
			p1=ss->ggibuffer+ss->selendx+ss->selendy*ss->width;
			p2=ss->ggibuffer+ss->selstx +ss->selsty *ss->width;
			for(p3=p1;p3<=p2;p3++) *p3 ^= 0x2000;
			for(p3=p2;p3<=p1;p3++) *p3 ^= 0x2000;
		}
	}
}

static int userscroll_uses(evstack *slf)
{
	return userscroll_use;
}
                                                        
static int userscroll_write(evstack *slf, char *str)
{
	return -EINVAL;
}

static int userscroll_read(evstack *slf,int lineno,int maxlen,char *str)
{
	if (lineno==0) 
	{
		strcpy(str,"No info available."); return 0;
	}
	return -EINVAL;
}
                                                        
void init_userscroll(void)
{       
	ggiInit();
	if (!MASTER_VIS) 
	{	MASTER_VIS=ggiOpen(NULL);
		ggiSetTextMode(MASTER_VIS,80,25,80,25,8,16,GT_TEXT16);
	}

	ev_register_class("userscroll",&userscroll_ops);
	printk("Registered userscroll.\n");
}
