/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <signal.h>

#include "ggi-dl.h"
#include "internal.h"
#include "debug.h"
#include "aavisual.h"

#include <ggi/ggi_commands.h>

int GGIsetorigin(ggi_visual *vis,int x,int y)
{
	return E(DRIVER,NOSUP_ALWAYS_CANT);
}

int GGIsetsplitline(ggi_visual *vis,int y)
{
	return E(DRIVER,NOSUP_ALWAYS_CANT);
}

static int _GGIdomode(ggi_visual *vis)
{
	int err;
	ggi_suggest sug;

	char sugs[1][80];
	
	switch(VIS_MODE(vis)->graphtype)
	{
	  case GT_8BIT : err=8;break;
	  default: return -EINVAL;
	}

	sprintf(sugs[0],"generic-linear-%d",err);

	_ggiZapMode(vis,~GGI_DL_OPDISPLAY);

	_ggiSetupMode(vis);

	err=(_ggiOpenDL(vis,"generic-stubs","")==NULL);
	if (err) {
		fprintf(stderr,"display-AA: Can't open the \"generic-stubs\" "
				"library.\n");
	}

	sug.id=0;
	do {
		/* FIXME ! */
		strcpy(sug.name,sugs[sug.id]);
		strcpy(sug.args,"");

		err=(_ggiOpenDL(vis,sug.name,sug.args)==NULL);
		if (err) {
			fprintf(stderr,"display-X: Can't find an appropriate "
				       "library for %s (%s)\n",
					sug.name,sug.args);
			return err;
		} else {
			DPRINT("Success in loading %s (%s)\n",sug.name,sug.args);
		}
	} while (sug.id!=0);

	/* Override the color operations for X if we are 8 bit palette. */
	if (AA_PRIV(vis)->colormap) { 
		free(AA_PRIV(vis)->colormap);AA_PRIV(vis)->colormap=NULL;
	}

	if (VIS_MODE(vis)->graphtype==GT_8BIT) {

		int x;
		AA_PRIV(vis)->nocols=256;
		AA_PRIV(vis)->colormap=malloc(256*sizeof(ggi_color));
		for(x=0;x<AA_PRIV(vis)->nocols;x++)
		{
			AA_PRIV(vis)->colormap[x].r=((x&0xe0)>>5)*0x2492;
			AA_PRIV(vis)->colormap[x].g=((x&0x1c)>>2)*0x2492;
			AA_PRIV(vis)->colormap[x].b=((x&0x03)   )*0x5555;
		}
		vis->opcolor->mapcolor=GGImapcolor;
		vis->opcolor->setpalvec=GGIsetpalvec;
		vis->opcolor->getpalvec=GGIgetpalvec;
		/* This is just to trigger the initial setup. */
		GGIsetpalvec(vis,0,256,AA_PRIV(vis)->colormap);
	}
	/* color */
	vis->opcolor->unmappixel=GGIunmappixel;

	return EOK;
}

int _GGIdoflush(struct AAhooks *aahook)
{
	static int lock=0;
	
	if (lock++!=0) {lock--;return -EAGAIN;}	/* FXIME use spinlocks here !*/

	aa_renderpalette(aahook->context,aahook->pal, &aa_defrenderparams, 0, 0, 
			aa_scrwidth(aahook->context), aa_scrheight(aahook->context));
	aa_flush(aahook->context);

	lock--;	/* FIXME Spinlocks ... */

	return EOK;
}

static struct AAhooks *ahk;
#include "../common/mansync.inc"

int GGIsetmode(ggi_visual *vis,ggi_mode *tm)
{ 
	struct AAhooks *aahook; 
	
	if(GGIcheckmode(vis, tm))
		return -1;

	aahook=VIS_PRIVATE(vis);

	if (!(aahook->isasync))
		_mansync_ignore(vis);

	if (VIS_FB_LINEAR(vis)!=NULL) {
		VIS_FB_LINEAR(vis)=NULL;
		VIS_FB_LINEAR_SIZE(vis)=0;
	}

    	aahook->context=aa_autoinit(&aa_defparams);
	tm->virt.x=aa_imgwidth (aahook->context);
	tm->virt.y=aa_imgheight(aahook->context);
	VIS_FB_LINEAR(vis)=aa_image(aahook->context);
	VIS_FB_LINEAR_SIZE(vis)=tm->virt.x*tm->virt.y;
	aa_autoinitkbd(aahook->context,AA_SENDRELEASE);
	aa_autoinitmouse(aahook->context,AA_MOUSEALLMASK);

	_GGIdoflush(ahk=aahook);

	if (!(aahook->isasync))
		_mansync_cont(vis);
	
	memcpy(VIS_MODE(vis),tm,sizeof(ggi_mode));

	return _GGIdomode(vis);
}

/**********************************/
/* check any mode (text/graphics) */
/**********************************/
int GGIcheckmode(ggi_visual *vis,ggi_mode *tm)
{
	int err = 0;

	if (vis==NULL)
		return -1;
	
	if(tm->graphtype==GGI_AUTO) tm->graphtype=GT_8BIT;
	if(tm->graphtype!=GT_8BIT) err = -1;

	if(	(tm->dpp.x!=1 && tm->dpp.x!=GGI_AUTO) ||
		(tm->dpp.y!=1 && tm->dpp.y!=GGI_AUTO))
		{ err = -1; }
	tm->dpp.x = tm->dpp.y = 1;

	/* FIXME: Still doesn't check visible and virt size */

	return err;
}

/************************/
/* get the current mode */
/************************/
int GGIgetmode(ggi_visual *vis,ggi_mode *tm)
{
	DPRINT("In GGIgetmode(%p,%p)\n",vis,tm);
	if (vis==NULL)
		return -1;
	
	/* We assume the mode in the visual to be o.k. */
	memcpy(tm,VIS_MODE(vis),sizeof(ggi_mode));

	return 0;
}
