/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <signal.h>

#include "aavisual.h"
#include "ggi-dl.h"

/* 
 * We are guaranteed to have been checked already...
 */

int GGIdlinit(ggi_visual *vis,const char *args)
{
	struct AAhooks *aainfo;

	DPRINT("AA-lib starting.\n");

	VIS_PRIVATE(vis)=aainfo=malloc(sizeof(struct AAhooks));
	if (VIS_PRIVATE(vis)==NULL)
		return GGI_DL_ERROR;

	aainfo->context=0;
	aainfo->colormap=NULL;

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 10) {
		/* Has mode management */
		vis->opdisplay->flush=GGIflush;
		vis->opdisplay->getmode=GGIgetmode;
		vis->opdisplay->setmode=GGIsetmode;
		vis->opdisplay->checkmode=GGIcheckmode;
	}

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 15) {
		/* Has Event management */
		vis->opdisplay->eventpoll=GGIeventpoll;
		vis->opdisplay->eventread=GGIeventread;
	}

	return GGI_DL_OPDISPLAY;
}

int GGIdlcleanup(ggi_visual *vis)
{
	struct AAhooks *aainfo;

	/* prevents a soon-to-be unloaded function from being called */
	_mansync_stop(vis);

	/* FIXME !!! */
	if ((aainfo=VIS_PRIVATE(vis)) != NULL)
	{
		if (aainfo->context)
		{
			aa_uninitmouse(aainfo->context);
			aa_uninitkbd(aainfo->context);
			aa_close(aainfo->context);
		}
		if(aainfo->colormap)
			free(aainfo->colormap);

		free(VIS_PRIVATE(vis));
	}

	return EOK;
}

#include "ggidlinit.h"
