/* --------------------------------------------------------------------------- 
** LibGGI GLIDE target
** ---------------------------------------------------------------------------
** Copyright (C) 1997, Jon Taylor <taylorj@ecs.csus.edu>
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
** ---------------------------------------------------------------------------
*/

#include <unistd.h>
#include <sys/fcntl.h>
#include <stdlib.h>
#include <signal.h>

#include "glidevisual.h"
#include "ggi-dl.h"

int        _GGIbpc = 0;
int        _GGInocols = 0;
ggi_color *_GGIcolormap = NULL;

/* 
** We are guaranteed to have been checked already...
*/

int GGIdlinit(ggi_visual *vis, const char *args)
{
	struct glide_hooks *glideinfo;
	DPRINT("GLIDE-lib starting");
	
	grGlideInit();
	grSstSelect(0);
	
	VIS_PRIVATE(vis) = glideinfo = malloc(sizeof(struct glide_hooks));
	if (VIS_PRIVATE(vis) == NULL)
	{
		
		return GGI_DL_ERROR;
	}

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 10) 
	{
		/* Has mode management */
		vis->opdisplay->flush = GGIflush;
		vis->opdisplay->getmode = GGIgetmode;
		vis->opdisplay->setmode = GGIsetmode;
		vis->opdisplay->checkmode = GGIcheckmode;
	}

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 15) 
	{
		/* Has Event management */
		vis->opdisplay->eventpoll = GGIeventpoll;
		vis->opdisplay->eventread = GGIeventread;
	}
	
	vis->op2d->setsplitline = GGIsetsplitline;
	vis->op2d->setorigin = GGIsetorigin;

	return GGI_DL_OPDISPLAY | GGI_DL_OP2D;
}

int GGIdlcleanup(ggi_visual *vis)
{
	signal(SIGALRM, SIG_IGN);
	
	if (VIS_FD(vis) > -1)
		close(VIS_FD(vis));
	
	grGlideShutdown();
	
	return EOK;
}

#include "ggidlinit.h"
