/* $Id: color.c,v 1.3 1998/02/17 21:22:22 vogler Exp $
 *
 * Copyright 1998 MenTaLboY - mentalboy@geocities.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <errno.h>

#include "TIvisual.h"
#include "debug.h"
#include "ggi-dl.h"

#include <ggi/ggi.h>
#include <ggi/ggi_commands.h>

extern int _GGIbpc;
extern int _GGInocols;
extern ggi_color *_GGIcolormap;

#define GGI_COLOR_COMPONENT_TYPE uint16

#define THREE_THIRDS (GGI_COLOR_COMPONENT_TYPE)((1<<(GGI_COLOR_PRECISION+1))-1)
#define TWO_THIRDS (GGI_COLOR_COMPONENT_TYPE)((1<<(GGI_COLOR_PRECISION+2))/3)
#define ONE_THIRD (GGI_COLOR_COMPONENT_TYPE)((1<<(GGI_COLOR_PRECISION+1))/3) 
static const ggi_color vga16_palette[16] = {
	{ 0, 0, 0 },
	{ 0, 0, TWO_THIRDS },
	{ 0, TWO_THIRDS, 0 },
	{ 0, TWO_THIRDS, TWO_THIRDS },
	{ TWO_THIRDS, 0, 0 },
	{ TWO_THIRDS, 0, TWO_THIRDS },
	{ TWO_THIRDS, ONE_THIRD, 0 }, 
	{ TWO_THIRDS, TWO_THIRDS, TWO_THIRDS },
	{ ONE_THIRD, ONE_THIRD, ONE_THIRD },
	{ ONE_THIRD, ONE_THIRD, THREE_THIRDS },
	{ ONE_THIRD, THREE_THIRDS, THREE_THIRDS },
	{ THREE_THIRDS, ONE_THIRD, ONE_THIRD },
	{ THREE_THIRDS, ONE_THIRD, THREE_THIRDS },
	{ THREE_THIRDS, THREE_THIRDS, ONE_THIRD },
	{ THREE_THIRDS, THREE_THIRDS, THREE_THIRDS }
};
#undef ONE_THIRD
#undef TWO_THIRDS
#undef THREE_THIRDS

#undef GGI_COLOR_COMPONENT_TYPE

static inline chtype map_char_to_chtype(ggi_visual *vis, char c) {
	struct TIhooks *tiinfo;
	tiinfo = VIS_PRIVATE(vis);
	switch (c) {
		case '\000': return ' ';
		default:     return tiinfo->charmap[(int)c];
	}
}

static inline chtype get_color_pair(ggi_visual *vis, int fg, int bg) {
	if ( COLOR_PAIRS == 0 ) return 0;
	return COLOR_PAIR(((COLORS-fg%COLORS-1)+(bg%COLORS*COLORS))%COLOR_PAIRS);
}

static inline chtype map_text16_to_ncurses(ggi_visual *vis, uint16 pixel)
{
	int fg, bg;
	struct TIhooks *tiinfo;
	tiinfo = VIS_PRIVATE(vis);

	fg = pixel >> 8 & 0x0F;
	bg = pixel >> 12 & 0x0F;

	return map_char_to_chtype(vis, pixel & 0xFF)
		| tiinfo->color16_table[fg+(bg<<4)];
}

static inline chtype map_text32_to_ncurses(ggi_visual *vis, uint32 pixel)
{
	chtype attributes;

	attributes = ( pixel & ATTR_HALF ) ? A_DIM : A_NORMAL;
	attributes |= ( pixel & ATTR_BRIGHT ) ? A_STANDOUT : A_NORMAL;
	attributes |= ( pixel & ATTR_UNDERLINE ) ? A_UNDERLINE : A_NORMAL;
	attributes |= ( pixel & ATTR_BOLD ) ? A_BOLD : A_NORMAL;
	attributes |= ( pixel & ATTR_ITALIC ) ? A_STANDOUT : A_NORMAL;
	attributes |= ( pixel & ATTR_REVERSE ) ? A_REVERSE : A_NORMAL;
	attributes |= ( pixel & ATTR_BLINK ) ? A_BLINK : A_NORMAL;
	attributes |= ( ( 1 << 23 ) & pixel ) ? A_ALTCHARSET : A_NORMAL;

	return ( ( pixel >> 24 ) & 0xFF) | attributes
		| get_color_pair(vis, (ATTR_FGCOLOR & pixel)>>8,
			ATTR_BGCOLOR & pixel);
		
}

static int paint_ncurses_window16(ggi_visual *vis, WINDOW *win, int width,
                                  int height)
{
	unsigned char *fb_walk;
	int splitline;
	int x_limit, y_limit;
	int fb_pad;
	int x, y;
	chtype *linebuffer;
	size_t buffer_size;

	{
		struct TIhooks *tiinfo;
		ggi_mode *mode;
		int fb_width, fb_height, vis_width, vis_height;

		fb_walk = VIS_FB_LINEAR(vis);
		fb_width = VIS_FB_WIDTH(vis);
		fb_height = VIS_FB_HEIGHT(vis);

		tiinfo = VIS_PRIVATE(vis);

		mode = VIS_MODE(vis);
		vis_width = mode->visible.x;
		vis_height = mode->visible.y;

		x_limit = ( width < vis_width ) ? width : vis_width;
		y_limit = ( height < vis_height ) ? height : vis_height;

		fb_walk += ( tiinfo->origin.x + tiinfo->origin.y * fb_width ) << 1;
		fb_pad = ( fb_width - x_limit ) << 1;

		splitline = tiinfo->splitline;

		buffer_size = sizeof(chtype) * width;
		linebuffer = alloca(buffer_size);	
		memset(linebuffer, 0, buffer_size);
	}

	for ( y = 0 ; y < y_limit ; y++ ) {
		if ( y == splitline ) {
			fb_walk = VIS_FB_LINEAR(vis);
		}
		for ( x = 0 ; x < x_limit ; x++, fb_walk += 2 ) {
			linebuffer[x] = map_text16_to_ncurses(vis, *((uint16 *)fb_walk));
		}
		fb_walk += fb_pad;
		mvwaddchnstr(win, y, 0, linebuffer, width);
	}
	if ( y < width ) {
		memset(linebuffer, 0, buffer_size);
		for ( ; y < height ; y++ ) {
			mvwaddchnstr(win, y, 0, linebuffer, width);
		}
	} 
	return EOK;
}

static int paint_ncurses_window32(ggi_visual *vis, WINDOW *win, int width,
                                  int height)
{
	unsigned char *fb_walk;
	int splitline;
	int x_limit, y_limit;
	int fb_pad;
	int x, y;
	chtype *linebuffer;
	size_t buffer_size;

	{
		struct TIhooks *tiinfo;
		ggi_mode *mode;
		int fb_width, fb_height, vis_width, vis_height;

		fb_walk = VIS_FB_LINEAR(vis);
		fb_width = VIS_FB_WIDTH(vis);
		fb_height = VIS_FB_HEIGHT(vis);

		tiinfo = VIS_PRIVATE(vis);

		mode = VIS_MODE(vis);
		vis_width = mode->visible.x;
		vis_height = mode->visible.y;

		x_limit = ( width < vis_width ) ? width : vis_width;
		y_limit = ( height < vis_height ) ? height : vis_height;

		fb_walk += ( tiinfo->origin.x + tiinfo->origin.y * fb_width ) << 1;
		fb_pad = ( fb_width - x_limit ) << 2;

		splitline = tiinfo->splitline;

		buffer_size = sizeof(chtype) * width;
		linebuffer = alloca(buffer_size);	
		memset(linebuffer, 0, buffer_size);
	}

	for ( y = 0 ; y < y_limit ; y++ ) {
		if ( y == splitline ) {
			fb_walk = VIS_FB_LINEAR(vis);
		}
		for ( x = 0 ; x < x_limit ; x++, fb_walk += 4 ) {
			linebuffer[x] = map_text32_to_ncurses(vis, *((uint32 *)fb_walk));
		}
		fb_walk += fb_pad;
		mvwaddchnstr(win, y, 0, linebuffer, width);
	}
	if ( y < width ) {
		memset(linebuffer, 0, buffer_size);
		for ( ; y < height ; y++ ) {
			mvwaddchnstr(win, y, 0, linebuffer, width);
		}
	} 
	return EOK;
}

int paint_ncurses_window(ggi_visual *vis, WINDOW *win, int width, int height)
{
	switch (VIS_MODE(vis)->graphtype) {
		case GT_TEXT16: return paint_ncurses_window16(vis, win, width, height);
		case GT_TEXT32: return paint_ncurses_window32(vis, win, width, height);
		default: return -EINVAL;
	}
}

ggi_pixel GGImapcolor(ggi_visual *vis, ggi_color *col)
{
	switch (VIS_MODE(vis)->graphtype) {
		case GT_TEXT16: {
			long distance;
			int current;
			long closest_distance=(1<<(GGI_COLOR_PRECISION+1))*3;
			int closest=0;
			for ( current = 0 ; current < 16 ; current++ ) {
				distance = abs(col->r - vga16_palette[current].r);
				distance += abs(col->g - vga16_palette[current].g);
				distance += abs(col->b - vga16_palette[current].b);
				if ( distance <= closest_distance ) {
					closest = current;
				}
			}
			return (closest<<8)|219;
		}
		/* FIXME text32 support needed, but text32 pixel bitmapping may change soon */
		default: return -ENOSYS;
	}
}

int GGIunmappixel(ggi_visual *vis, ggi_pixel pix, ggi_color *col)
{
	switch (VIS_MODE(vis)->graphtype) {
		case GT_TEXT16: {
			pix = ( pix >> 8 ) & 0x0F;
			memcpy(col, &vga16_palette[pix%16], sizeof(ggi_color));
			return 0;
		}
		/* FIXME text32 support needed, but text32 pixel bitmapping may change soon */
		default: return -ENOSYS;
	}
}
