/* $Id: mode.c,v 1.4 1998/02/17 21:22:24 vogler Exp $
 *
 * Copyright 1998 MenTaLboY - mentalboy@geocities.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <errno.h>

#define NEED_GRAPHTYPE_SYMS

#include "TIvisual.h"
#include "debug.h"
#include "ggi-dl.h"

#include <ggi/ggi_commands.h>

extern int _GGIbpc;
extern int _GGInocols;
extern ggi_color *_GGIcolormap;

int GGIsetorigin(ggi_visual *vis, int x, int y)
{
	struct TIhooks *tiinfo;
	ggi_mode *mode;

	tiinfo = VIS_PRIVATE(vis);
	mode = VIS_MODE(vis);

	x /= mode->dpp.x; y /= mode->dpp.y; /* terminfo can only set origin
                                               with pixel granularity, so
                                               internally, terminfo setorigin
                                               works in pixels */

	if ( ( x < 0 ) || ( x > ( mode->virt.x - mode->visible.x ) ) )
		return -EINVAL;
	if ( ( y < 0 ) || ( y > ( mode->virt.y - mode->visible.y ) ) )
		return -EINVAL;

	tiinfo->origin.x = x;
	tiinfo->origin.y = y;

	ggiFlush(vis);
	
	return EOK;
}

int GGIsetsplitline(ggi_visual *vis, int line)
{
	struct TIhooks *tiinfo;
	ggi_mode *mode;

	tiinfo = VIS_PRIVATE(vis);
	mode = VIS_MODE(vis);
	
	line /= mode->dpp.y; /* terminfo can only set splitline with pixel
                                granularity, so internally, terminfo splitline
                                works in pixels */ 

	if ( ( line < 0 ) || ( line > mode->visible.y ) )
		return -EINVAL;

	tiinfo->splitline = line;

	ggiFlush(vis);

	return EOK;
}

int GGIflush(ggi_visual *vis)
{
	struct TIhooks *tiinfo;
	ggi_mode *mode;

	tiinfo = VIS_PRIVATE(vis);
	mode = VIS_MODE(vis);

	lock_ncurses();
	set_term(tiinfo->scr);
	paint_ncurses_window(vis, stdscr, COLS, LINES);
	refresh();
	unlock_ncurses();
	
	return EOK;
}

static int _GGIloadstub(ggi_visual *vis, ggi_suggest *sug)
{
	int status;
	status = ( _ggiOpenDL(vis, sug->name, sug->args) == NULL );
	if ( status ) {
		fprintf(stderr, "display-terminfo: Unable to load an "
				"appropriate library for %s (%s)\n", sug->name,
				sug->args);
	} else {
		DPRINT("display-terminfo: Loaded %s (%s)\n", sug->name, sug->args);
	}
	return status;
}

static int _GGIdomode(ggi_visual *vis)
{
	ggi_suggest sug;
	struct TIhooks *tiinfo;

	tiinfo = VIS_PRIVATE(vis);

	_ggiZapMode(vis, ~GGI_DL_OPDISPLAY);
	_ggiSetupMode(vis);

	strcpy(sug.name, "generic-stubs");
	strcpy(sug.args, "");

	_GGIloadstub(vis, &sug);

	switch (VIS_MODE(vis)->graphtype) {
		case GT_TEXT16: strcpy(sug.name, "generic-text-16"); break;
		case GT_TEXT32: strcpy(sug.name, "generic-text-32"); break;
		default: return -EINVAL;
	}
	strcpy(sug.args, "");

	_GGIloadstub(vis, &sug);

	vis->op2d->setorigin = GGIsetorigin;
	vis->op2d->setsplitline = GGIsetsplitline;
	vis->opcolor->mapcolor = GGImapcolor;
	vis->opcolor->unmappixel = GGIunmappixel;

	tiinfo->virgin = 0;
	tiinfo->origin.x = tiinfo->origin.y = 0;
	tiinfo->splitline = VIS_MODE(vis)->visible.y;

	lock_ncurses();
	set_term(tiinfo->scr);
	wclear(stdscr);
	refresh();
	unlock_ncurses();
	return EOK;
} 

int GGIsetmode(ggi_visual *vis, ggi_mode *tm)
{
	struct TIhooks *tiinfo;
	int status;
	void *fb;
	size_t fb_size;

	tiinfo = VIS_PRIVATE(vis);

	DPRINT("display-terminfo: setmode mode %s %dx%d (%dx%d dots, %dx%d font)\n",
		GRAPHTYPE(tm->graphtype),
		tm->visible.x, tm->visible.y,
		tm->visible.x * tm->dpp.x, tm->visible.y * tm->dpp.y,
		tm->dpp.x, tm->dpp.y);

	status = GGIcheckmode(vis, tm);
	if ( status ) return status;

	DPRINT("display-terminfo: approved mode %s %dx%d (%dx%d dots, %dx%d font)\n",
		GRAPHTYPE(tm->graphtype),
		tm->visible.x, tm->visible.y,
		tm->visible.x * tm->dpp.x, tm->visible.y * tm->dpp.y,
		tm->dpp.x, tm->dpp.y);

	if (VIS_FB_LINEAR(vis) != NULL) {
		free(VIS_FB_LINEAR(vis));
		VIS_FB_LINEAR(vis) = NULL;
		VIS_FB_LINEAR_SIZE(vis) = 0;
	}

	fb_size = ( ( tm->graphtype == GT_TEXT16 ) ? 2 : 4 ) * tm->virt.x * tm->virt.y;
	fb = malloc(fb_size);
	VIS_FB_LINEAR(vis) = fb;
	VIS_FB_LINEAR_SIZE(vis) = fb_size / ( ( tm->graphtype == GT_TEXT16 ) ? 2 : 4 );

	memcpy(VIS_MODE(vis), tm, sizeof(ggi_mode));

	return _GGIdomode(vis);
}

int GGIcheckmode(ggi_visual *vis, ggi_mode *tm)
{
	int status=0;
	struct TIhooks *tiinfo;

	tiinfo = VIS_PRIVATE(vis);

	lock_ncurses();
	set_term(tiinfo->scr);
	if ( tm->dpp.x == GGI_AUTO ) tm->dpp.x = 8;
	if ( tm->dpp.y == GGI_AUTO ) tm->dpp.y = 8;
	if ( ( tm->visible.x > COLS  )
		|| ( tm->visible.y > LINES ) )
			status = GGI_DL_ERROR;
	/* FIXME we should also take into account the number of pages desired */
	tm->visible.x = COLS;
	tm->visible.y = LINES;
	unlock_ncurses();
	if ( tm->virt.x < tm->visible.x ) tm->virt.x = tm->visible.x;
	if ( tm->virt.y < tm->visible.y ) tm->virt.y = tm->visible.y;
	tm->frames = ( tm->virt.x / tm->visible.x ) * ( tm->virt.y / tm->visible.y );
	if ( ( tm->graphtype != GT_TEXT16 ) && ( tm->graphtype != GT_TEXT32 ) ) {
		tm->graphtype = GT_TEXT32;
		status = GGI_DL_ERROR;
	}
	return status;
}

int GGIgetmode(ggi_visual *vis, ggi_mode *tm)
{
	if ( vis == NULL ) return -1;
	memcpy(tm, VIS_MODE(vis), sizeof(ggi_mode));
	DPRINT("display-terminfo: getmode mode %s %dx%d (%dx%d dots, %dx%d font)\n",
		GRAPHTYPE(tm->graphtype),
		tm->visible.x, tm->visible.y,
		tm->visible.x * tm->dpp.x, tm->visible.y * tm->dpp.y,
		tm->dpp.x, tm->dpp.y);
	return 0;
}
