/* 
 *  Graphics library for GGI.
 * 
 *  Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
 * 
 *  If you do any modifications, I would like you to send diffs to me
 *  to allow for collecting a more and more complete set of drivers and
 *  to improve existing ones.
 * 
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * ----------------------------------------------------------------------
 * 
 *	$Log: init.c,v $
 *	Revision 1.27  1998/03/12 20:41:14  msundberg
 *	For now, don't really exit in ggiExit() when _ggiLibIsUp > 1
 *	
 *	Revision 1.26  1998/03/12 11:53:21  msundberg
 *	libggi: fixed bug: open visuals were not closed in ggiExit.
 *	Removed last traces of the "focus" notion.
 *	
 *	Revision 1.25  1998/03/09 02:10:31  msundberg
 *	Include dll initcode.
 *	
 *	Revision 1.24  1998/03/08 04:12:55  ajapted
 *	Minor: added the target parsing routine _GGIparsetarget.
 *	
 *	Revision 1.23  1998/03/04 08:40:39  hniemann
 *	changed _ggiParseMode to ggiParseMode, changed _ggi?PrintMode to ggiPrintMode
 *
 *	Revision 1.22  1998/03/03 16:37:23  hniemann
 *	LIBGGI_DEFMODE now parsed with _ggiParseMode
 *	
 *	Revision 1.21  1998/02/28 13:15:46  msundberg
 *	_ggiLibIsUp was not decremented in ggiExit when it was > 1.
 *	
 *	Revision 1.20  1998/02/15 21:52:26  msundberg
 *	Fixed the bugs when falling back on the svgalib-target.
 *	
 *	Revision 1.19  1998/02/14 03:07:11  msundberg
 *	Bugfix: ggiOpen() should clean up and return NULL when failing.
 *	Addition: Try svgalib-target if KGI-target fails.
 *	
 *	Revision 1.18  1998/02/08 11:20:58  vogler
 *	applied patches for text->dpp change by mentalboy@geocities.com
 *	
 *	Revision 1.17  1998/02/02 14:43:42  becka
 *	several patches, terminfo target.
 *	
 *	Revision 1.16  1998/01/25 15:21:31  becka
 *	GGI_AUTO fixes. Pleaase test !
 *	
 *	Revision 1.15  1998/01/23 13:44:57  becka
 *	New Lib minor. Final break-all sanity fixes.
 *	
 *	Revision 1.14  1997/12/18 00:19:37  becka
 *	Killed a few bugs.
 *	
 *	Revision 1.13  1997/12/15 17:28:42  becka
 *	Various slight fixes.
 *	
 *	Revision 1.12  1997/12/14 21:56:34  becka
 *	uint/sint namespace fixed ... please test ...
 *	
 *	Revision 1.11  1997/12/13 01:28:42  becka
 *	Alpha fixes. LibGGI mode negotiation changes ...
 *	
 *	Revision 1.10  1997/12/08 17:17:46  becka
 *	Keymaps right now ...
 *	
 *	Revision 1.9  1997/12/07 16:27:55  becka
 *	LibGGI extension fixes
 *	
 *	Revision 1.8  1997/12/04 14:50:51  becka
 *	Slight fixes.
 *	
 *	Revision 1.7  1997/12/03 17:08:22  becka
 *	Fixed X ressource-keeping bug.
 *	
 *	Revision 1.6  1997/11/30 13:55:56  becka
 *	Threads fixes, Extension handling.
 *	
 *	Revision 1.5  1997/11/25 19:39:43  becka
 *	default mode made available to outside.
 *	
 *	Revision 1.4  1997/11/22 20:47:01  becka
 *	LIBGGI_DEFMODE support.
 *	
 *	Revision 1.3  1997/11/15 01:13:19  willie
 *	Changed GPL to LGPL License
 *	
 *	Revision 1.2  1997/11/04 23:11:22  becka
 *	OSF1 support. Some fixups.
 *	
 *	Revision 1.1  1997/10/12 04:53:41  tfries
 *	new subdir
 *	
 *	Revision 1.14  1997/06/14 23:48:48  ezrec
 *	Minor: Fixed some of Todd's Makefile bugs..
 *	
 *	Revision 1.13  1997/06/04 05:41:28  ezrec
 *	Minor: First pass at gamma-map support (_completely_ untested), and
 *	       libggi auto-detects the X `DISPLAY' environment variable and
 *	       uses the display-X driver. (Which doesn't work yet...)
 *
 *	Revision 1.12  1997/06/02 01:03:53  ezrec
 *	Minor: Changed the ggiOpen(NULL) default to correspond with the
 *	       the top level `make devices'
 *
 *	Revision 1.11  1997/05/27 02:23:47  ezrec
 *	Minor: Changed `LIBGGI_DEFFAULT' envvar to `LIBGGI_DISPLAY'
 *
 *	Revision 1.10  1997/05/23 23:29:56  ezrec
 *	Major: Moved event handling to the display driver, added 2.1.40 support.
 *		libggi-dynamic debugging is now environment-variable controlled.
 *
 *	Revision 1.9  1997/05/16 05:12:12  ezrec
 *	Major: Event support, and updated ggilib shim. Xggi 0.0.8 binaries should
 *		work now, with a minor caveat on mouse buttons..
 *
 *	Revision 1.8  1997/04/17 05:08:11  ezrec
 *	Major: 	New ggiOpen() semantics ie ggiOpen("display-KGI:/dev/graph3",NULL)
 *		First attempt at linear_1 frame buffer
 *		Moved thread support `to the side'
 *		Added DPRINT() support
 *		Added display-KGI driver
 *
 *	Revision 1.7  1997/03/25 04:54:11  ezrec
 *	Major: libggi now handles libggi.conf files, IBM vga driver suggests
 *	      default ioctl RAMDAC DLL, default ioctl RAMDAC DLL support,
 *	      -Wall added to the libggi-dynamic Makefiles.
 *
 *	Revision 1.6  1997/03/23 08:27:40  ezrec
 *	Major: libggi-dynamic now uses `stubs' instead of #defines so that
 *	       new versions of the library won't break old ones...
 *
 * 
 * ----------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>

#include "internal.h"
#include "display/common/parsing.inc"

/*
 * Initalize the strutures for the library
 */
static struct {
	ggi_mutex	mutex;	/* Lock when changing.. */
	int visuals;		/* # of visuals active */
	ggi_visual *visual;	/* Linked list of all visuals */
} _ggiVisuals;	/* This is for remembering visuals. */
static int 		_ggiLibIsUp=0;

static char  ggiconfstub[512]=GGICONFFILE;
static char *ggiconffile=ggiconfstub+4;

int ggiDebugState=0;

int _ggiExtensionSize=0;
ggi_extension *_ggiExtension;

int ggiInit(void)
{
	int err;
	const char *str;

	_ggiLibIsUp++;
	if (_ggiLibIsUp>1) return EOK;	/* Initialize only at first call. */

	_ggiLockInit(&_ggiVisuals);
	_ggiVisuals.visuals=0;
	_ggiVisuals.visual=NULL;

	str=getenv("LIBGGI_DEBUG");
	if (str!=NULL) {
		ggiDebugState=atoi(str);
		DPRINT("Debugging=%d\n",ggiDebugState);
	}

	str=getenv("LIBGGI_DEFMODE");
	if (str!=NULL) {
#if 1
		ggiParseMode(str,&ggiDefaultMode);
#else
		int visx,visy,virx,viry,dppx,dppy,gmod;

		visx=ggiDefaultMode.visible.x;
		visy=ggiDefaultMode.visible.y;
		virx=ggiDefaultMode.virt.x;
		viry=ggiDefaultMode.virt.y;
		dppx=ggiDefaultMode.dpp.x;
		dppy=ggiDefaultMode.dpp.y;
		
		switch(ggiDefaultMode.graphtype)
		{
		  case GGI_AUTO: gmod=-1;break;
		  case GT_8BIT : gmod=8;break;
		  case GT_15BIT: gmod=15;break;
		  case GT_16BIT: gmod=16;break;
		  case GT_24BIT: gmod=24;break;
		  case GT_32BIT: gmod=32;break;
		  case GT_TEXT16: gmod=-16;break;
		  case GT_TEXT32: gmod=-32;break;
		  default: fprintf(stderr,"init.c: Invalid default depth!\n");
		}

		sscanf(str,"%dx%dx%d %dx%d %dx%d",
			&visx,&visy,&gmod, &virx,&viry, &dppx,&dppy);

		ggiDefaultMode.visible.x=visx;
		ggiDefaultMode.visible.y=visy;
		ggiDefaultMode.virt.x   =virx;
		ggiDefaultMode.virt.y   =viry;
		ggiDefaultMode.dpp.x   =dppx;
		ggiDefaultMode.dpp.y   =dppy;

		switch (gmod) {
			case -1: ggiDefaultMode.graphtype=GGI_AUTO;break;
			case 1:	ggiDefaultMode.graphtype=GT_1BIT;break;
			case 4:	ggiDefaultMode.graphtype=GT_4BIT;break;
			case 8:	ggiDefaultMode.graphtype=GT_8BIT;break;
			case 15: ggiDefaultMode.graphtype=GT_15BIT;break;
			case 16: ggiDefaultMode.graphtype=GT_16BIT;break;
			case 24: ggiDefaultMode.graphtype=GT_24BIT;break;
			case 32: ggiDefaultMode.graphtype=GT_32BIT;break;
			case -16: ggiDefaultMode.graphtype=GT_TEXT16;break;
			case -32: ggiDefaultMode.graphtype=GT_TEXT32;break;
			default: fprintf(stderr,"init.c: Invalid default depth!\n");
			 break;
		}
#endif
	}

	err=_ggiLoadConfig(ggiconffile);
	if (err==0)
		return EOK;

	fprintf(stderr,"libggi: couldn't open %s", ggiconffile );

	_ggiLockDestroy(&_ggiVisuals);
	_ggiLibIsUp--;
	return err;
}

void ggiExit(void)
{
	ggi_extension *tmp,*next;

	if (!_ggiLibIsUp)
		return;

	if (_ggiLibIsUp > 1) {
		_ggiLibIsUp--;
		return;
	}

	DPRINT("ggiExit: really destroying.\n");
	while (_ggiVisuals.visual!=NULL) 
		ggiClose(_ggiVisuals.visual);

	_ggiLockDestroy(&_ggiVisuals);
	
	for (tmp=_ggiExtension;tmp;tmp=next)
	{
		next=tmp->next;
		free(tmp);
	}
	
	_ggiLibIsUp=0;

	return;
}

void ggiPanic(const char *format,...)
{
	fprintf(stderr,"%s",format);
	ggiExit();
	exit(1);
}


/* Opens a visual.
 */
ggi_visual *ggiOpen(const char *driver,...)
{
	va_list drivers;
	ggi_visual *vis;
	char *cp;
	char str[1024];
	char target[1024];
	int  success=0;

	if (!_ggiLibIsUp)
		return NULL;

	if (driver==NULL) {
		void *ret;
		/* LIBGGI_DISPLAY _always_ overrides */
		cp=getenv("LIBGGI_DISPLAY");
		if (cp!=NULL)
			return ggiOpen(cp,NULL);

		/* Try the X display.. */
		cp=getenv("DISPLAY");
		if (cp!=NULL) {
			strcpy(str,"display-X:");
			strcat(str,cp);
			return ggiOpen(str,NULL);
		}

		/* Try the console.. */
		ret=ggiOpen("display-KGI:/dev/graphic",NULL);
		if (ret != NULL)
			return ret;
		
		/* Try svgalib target.. */
		return ggiOpen("display-svga",NULL);
	}

	_ggiLock(&_ggiVisuals);

	vis=_ggiNewVisual();

	va_start(drivers,driver);

	for (;driver!=NULL;driver=va_arg(drivers,const char *)) {

		DPRINT("Loading driver %s\n",driver);

		if (_GGIparsetarget((char *)driver,target) == NULL) {
			continue;
		}
		cp=strchr(target, ':');

		if (cp != NULL) {
			*cp++ = 0;
		}
		if (_ggiOpenDL(vis,target,cp) != NULL) {
			success=1;
		}
	}
	
	if (success) {
		vis->next=_ggiVisuals.visual;
		_ggiVisuals.visual=vis;
		_ggiVisuals.visuals++;
	} else {
		_ggiDestroyVisual(vis);
		vis=NULL;
	}
	
	_ggiUnlock(&_ggiVisuals);

	va_end(drivers);

	return vis;
}
	
/* ggiClose
 *	Attempt to close the visual associated with a vt
 *      returns 0 on success, nonzero on error
 */
int ggiClose(ggi_visual *visual)
{
	ggi_visual *vis,*pvis=NULL;

	if (!_ggiLibIsUp)
		return -1;

	_ggiLock(&_ggiVisuals);
	for (vis=_ggiVisuals.visual;vis!=NULL;pvis=vis,vis=vis->next)
		if (vis==visual)
			break;

	if (vis==NULL) {
		_ggiUnlock(&_ggiVisuals);
		return -1;
	}

	if (pvis==NULL)
		_ggiVisuals.visual=vis->next;
	else
		pvis->next=vis->next;

	_ggiVisuals.visuals--;

	vis->next=NULL;
	_ggiDestroyVisual(vis);
	
	_ggiUnlock(&_ggiVisuals);

	return 0;
}

int ggiRegisterExtension(int size,
			int (*load)(ggi_visual_t,const char *,const char *),
			int (*release)(ggi_visual_t,const char *))
{
	int off;
	ggi_extension *tmp,*ext=malloc(sizeof(ggi_extension));

	off=sizeof(ggi_visual)+_ggiExtensionSize;

	_ggiExtensionSize+=size;
	
	ext->size=size;
	ext->load=load;
	ext->release=release;
	ext->next=NULL;
	
	if (_ggiExtension) {
		for (tmp=_ggiExtension;tmp->next;tmp=tmp->next) ;
		tmp->next=ext;
		ext->prev=tmp;			
	} else {
		_ggiExtension=ext;
		ext->prev=NULL;
	}

	return off;
}

#include "ggilibinit.h"
