/* 
   Graphics library for GGI. Vertical lines.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>

#include "ggi-dl.h"
#include "vga4.h"

/**********************************/
/* draw/get/put a vertical line */
/**********************************/

/* Worst case - 1 page fault
 */
int GGIdrawvline(ggi_visual *vis,int x,int y,int height)
{
	volatile unsigned char *adr;
	int sw,bm;
	unsigned char foo;

	/* Clipping */
	if (x< (VIS_GC(vis)->cliptl.x) ||
	    x>=(VIS_GC(vis)->clipbr.x)) return EOK;
	if (y< (VIS_GC(vis)->cliptl.y)) {
		int diff=(VIS_GC(vis)->cliptl.y)-y;
		y     +=diff;
		height-=diff;
	}
	if (y+height>(VIS_GC(vis)->clipbr.y)) {
		height=(VIS_GC(vis)->clipbr.y)-y;
	}

	bm=(0x80>>(x&7));
	sw=(VIS_FB_WIDTH(vis))>>3;
	adr=(__localfb);
	adr+=((VIS_GC_FGCOLOR(vis)&0xF) << 16);
	adr+=(x>>3)+y*sw;

	for (;height>0;height--,adr+=sw) {
		foo=*adr;
		*adr=bm;
	};
  
	return EOK;
}

/* Worst case - 1 page fault
 */
int _GGIdrawvline(ggi_visual *vis,int x,int y,int height)
{
	volatile unsigned char *adr;
	int sw,bm;
	unsigned char foo;

	bm=(0x80>>(x&7));
	sw=(VIS_FB_WIDTH(vis))>>3;
	adr=(__localfb);
	adr+=((VIS_GC_FGCOLOR(vis)&0xF) << 16);
	adr+=(x>>3)+y*sw;

	for (;height>0;height--,adr+=sw) {
		foo=*adr;
		*adr=bm;
	};
  
	return EOK;
}

/*
 * Worst case: 16(!) page faults
 */
int GGIputvline(ggi_visual *vis,int x,int y,int height,void *buffer)
{ 
	unsigned char *adr,*ptr,*buff=(unsigned char *)buffer;
	int mask,sw,i,shift,bm;
	volatile register unsigned char foo;

	/* Clipping */
	if (x< (VIS_GC(vis)->cliptl.x) ||
	    x>=(VIS_GC(vis)->clipbr.x)) return EOK;
	if (y< (VIS_GC(vis)->cliptl.y)) {
		int diff=(VIS_GC(vis)->cliptl.y)-y;
		y     +=diff;
		height-=diff;
		buffer =((char *)buffer)+(diff>>1);
		shift=diff&1;
	}
	if (y+height>(VIS_GC(vis)->clipbr.y)) {
		height=(VIS_GC(vis)->clipbr.y)-y;
	}

	sw=(VIS_FB_WIDTH(vis)>>3);
	adr=(__localfb);
	adr+=(x>>3)+y*sw;

	bm=(0x80>>(x&7));
	for (mask=0;mask<16;mask++) {
		ptr=adr+(mask<<16);
		buff=(unsigned char *)buffer;
		for (i=0;i<height;i++,ptr+=sw) {
			if ( (buff[i>>1] >> (i&1 ? 0 : 4)) == mask) {
				foo=*ptr;
				*ptr=bm;
			}
		}
	}
		
	return EOK;
}

/*
 * Worst case: 4 page faults
 */
int GGIgetvline(ggi_visual *vis,int x,int y,int height,void *buffer)
{ 
	unsigned char *adr,*ptr,*buff=(unsigned char *)buffer;
	int plane,sw,i,bm;

	if (x<0||y<0||x>=VIS_FB_WIDTH(vis)||y+height>VIS_FB_HEIGHT(vis)) return -1;

	sw=(VIS_FB_WIDTH(vis)>>3);
	adr=__localfb;
	adr+=(x>>3)+y*sw;

	memset(buffer,0,(height+1)>>1);

	bm=(0x80>>(x&7));
	for (plane=0;plane<4;plane++) {
		ptr=adr+(plane<<16);
		buff=(unsigned char *)buffer;
		for (i=0;i<height;i++,ptr+=sw)
		  buff[i>>1]|=((*ptr&bm)!=0)<<((((i&1)^1)<<2)+plane);
	}
		
	return EOK;
}
