/* 
   Graphics library for GGI. Horizontal lines.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>

#include "ggi-dl.h"
#include "modex.h"

/**********************************/
/* draw/get/put a horizontal line */
/**********************************/

/* GGIdrawhline
 *  - Worst case: 3 page faults
 */
int GGIdrawhline(ggi_visual *vis,int x,int y,int w)
{
	unsigned char color,*ptr;
	int i,j,mask;

	/* Clipping */
	if (y<(VIS_GC(vis)->cliptl.y) || y>=(VIS_GC(vis)->clipbr.y)) return 0;
	if (x< (VIS_GC(vis)->cliptl.x)) {
		int diff=(VIS_GC(vis)->cliptl.x)-x;
		x+=diff;
		w-=diff;
	}
	if (x+w>(VIS_GC(vis)->clipbr.x)) {
		w=(VIS_GC(vis)->clipbr.x)-x;
	}
	if (w<=0) return 0;
	
	color=VIS_GC_FGCOLOR(vis);
	ptr=__localfb+(y*(VIS_FB_WIDTH(vis)>>2));

	if ((i=x&3)) {
		if ((j=w+i-4)<=0) {
			mask=(0xf<<i)&(0xf>>-j);
			ptr[(x>>2)+(mask<<16)]=color;
			return EOK;
		} else {
			mask=(0xf<<i)&0xF;
			ptr[(x>>2)+(mask<<16)]=color;
			w=w+i-4;x=x+4-i;
		}
	}

	memset(ptr+(x>>2)+0xF0000,color,w>>2);

	if ((i=(x+w)&3)) {
		mask = (1<<i)-1;
		ptr[((x+w)>>2)+(mask<<16)]=color;
	}

	return EOK;
}

/* _GGIdrawhline
 *  - Worst case: 3 page faults
 */
int _GGIdrawhline(ggi_visual *vis,int x,int y,int w)
{
	unsigned char color,*ptr;
	int i,j,mask;

	color=VIS_GC_FGCOLOR(vis);
	ptr=__localfb+(y*(VIS_FB_WIDTH(vis)>>2));

	if ((i=x&3)) {
		if ((j=w+i-4)<=0) {
			mask=(0xf<<i)&(0xf>>-j);
			ptr[(x>>2)+(mask<<16)]=color;
			return EOK;
		} else {
			mask=(0xf<<i)&0xF;
			ptr[(x>>2)+(mask<<16)]=color;
			w=w+i-4;x=x+4-i;
		}
	}

	memset(ptr+(x>>2)+0xF0000,color,w>>2);

	if ((i=(x+w)&3)) {
		mask = (1<<i)-1;
		ptr[((x+w)>>2)+(mask<<16)]=color;
	}

	return EOK;
}

/* GGIputhline
 * Worst case: 4 page faults
 */
int GGIputhline(ggi_visual *vis,int x,int y,int w,void *buff)
{ 
	unsigned char *ptr,*buffer=(unsigned char *)buff;
	int sl,i,j;

	/* Clipping */
	if (y<(VIS_GC(vis)->cliptl.y) || y>=(VIS_GC(vis)->clipbr.y)) return 0;
	if (x< (VIS_GC(vis)->cliptl.x)) {
		int diff=(VIS_GC(vis)->cliptl.x)-x;
		x+=diff;
		w-=diff;
		buffer+=diff;
	}
	if (x+w>(VIS_GC(vis)->clipbr.x)) {
		w=(VIS_GC(vis)->clipbr.x)-x;
	}
	if (w<=0) return 0;

	sl=(x>>2)+(VIS_FB_WIDTH(vis)>>2)*y;
	for (i=0;i<4 && w>0;buffer++,i++,w--) {
		ptr=__localfb+(1<<(16+(x&3)))+sl;
		for (j=0;j<w;j+=4) 
			*(ptr++)=buffer[j];
		x++;
		if ((x&3) == 0) 
			sl++;
	}
	return 0;
}

/* GGIgethline
 * Worst case: 4 page faults
 */
int GGIgethline(ggi_visual *vis,int x,int y,int w,void *buff)
{ 
	unsigned char *ptr,*buffer=(unsigned char *)buff;
	int sl,i,j;

	if (x<0||y<0||x+w>VIS_FB_WIDTH(vis)||y>=VIS_FB_HEIGHT(vis)) return -1;

	sl=(x>>2)+(VIS_FB_WIDTH(vis)>>2)*y;
	for (i=0;i<4 && w>0;buffer++,i++,w--) {
		ptr=__localfb+((x&3)<<16)+sl;
		for (j=0;j<w;j+=4) 
			buffer[j]=*(ptr++);
		x++;
		if ((x&3) == 0) 
			sl++;
	}
	return 0;
}
