/* 
 * ------------------------------------------------------------------------
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU Library General Public License as 
 *	published by the Free Software Foundation; either version 2, or 
 *	(at your option) any later version.
 * 
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Library General Public License for more details.
 *
 *	You should have received a copy of the GNU Library General Public 
 *	License	along with this program; see the file COPYING.LGPL .  
 *	If not, write to the Free Software Foundation, 675 Mass Ave, 
 *	Cambridge, MA 02139, USA.
 *
 * ------------------------------------------------------------------------
 */ 
#ifndef _MYSTIQUE_8_15_16_32_H_
#define _MYSTIQUE_8_15_16_32_H_

#include <Matrox/mga_common_lib.h>
#include <Matrox/mga_common_regs.h>
#include <Matrox/mga1064_main.h>

#define ACCELIDX 1024
#define ACCELMSK 1023 /* !!! Is system dependent */
#define FIFO_FLUSH 7 /* when to flush the buffer to the accelerator */

/* Global variables local to this lib 
*/

extern uint32  *__pingpong; /* Accelbuffer returned by mmap() */
extern uint8   *pengaddr;   /* BFIFO address data */
extern uint32  *pengval;    /* BFIFO register data */
extern uint32  *pengcnt;    /* pointer in accelbuffer to counter */
extern ggi_uint pengcur;    /* index to active pingpong page */
extern ggi_graphtype __gtype;
extern int __blkmode;       /* ATYPE_BLK or ATYPE_RPL */
extern int __bpp;
extern ggi_common_plb_setup __plb_setup;
extern const char *__dbname;
extern ggi_uint (*__fallback_mapcolor)(ggi_visual  *, ggi_color *);

ggi_uint GGImapcolor8(ggi_visual *vis,ggi_color *color);
ggi_uint GGImapcolor16(ggi_visual *vis,ggi_color *color);

/* Accelbuffer fill commands.  Some are implemented as macros, because
** macros can be optimized better.
** 
** The dataformat is the same as used for the DMA window of the Mystique.
** In fact, the contenst is copied verbatim in the driver.
*/

static inline void ACC_SWAP(void)
{
	DPRINT("Swapping accel pages (%d words)\n", *pengcnt);
	pengcur ^= ACCELIDX;
	pengaddr = (uint8 *)(__pingpong + pengcur);
	pengcnt  = __pingpong + pengcur + ACCELMSK;
	pengval  = __pingpong + pengcur + 1;
	*pengcnt = 0; /* trigger pagefault */
}

static inline void ACC_ADDDATA(uint8 addr,uint32 val)
{
	*pengaddr++ = addr; /* !!! Honor endianess some day */
	*pengval++  = val;
	if (((int)pengaddr & 3) == 0) {
		pengaddr += 4 * sizeof(uint32);
		pengval  += 1;
		if (++(*pengcnt) >= FIFO_FLUSH) {
			ACC_SWAP();
		}
	}
}

static inline void ACC_FINISH(void) 
{
	int i;
	for (i=(int)pengaddr & 3; (i&3)!=0; i++) {
		*pengaddr++ = DMA_PAD;
		*pengval++  = 0;
	}
	*pengcnt += (i != 0);
	ACC_SWAP();
}

static inline int ACC_FLUSH(ggi_visual *vis)
{
	ACC_FINISH();
	return ioctl(VIS_FD(vis),GR_MATROX_WAITACCEL);
}

static inline int SYNC_FLUSH(ggi_visual *vis)
{
	if (VIS_FLAGS(vis) & GGIFLAG_ASYNC) {
		DPRINT("ASYNC call\n");
		return 0;
	} else {
		return ACC_FLUSH(vis);
	}
}

#endif /* _MYSTIQUE_8_15_16_32_H_ */
