/* 
 *  horizontal line drawing
 *
 *  Copyright (C) 1998 Thomas Tanner. See CREDITS for details.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "lib.h"
#include "util.h"

/*----------------------------------------------------------------------*/

void	GGINAME(drawHLine_NOOP)(ggi_visual_t vis, int x1, int x2, int y)
{
}

void	GGINAME(drawHLine_INVERT)(ggi_visual_t vis, int x1, int x2, int y)
{
	int	width;
	pixel	*ptr;
    
	CLIPHLINE(vis, x1, x2, y, width);
	ptr	= LFBADDR(vis, x1, y);
	width	= x2 - x1 + 1;
	while (width--) { *ptr = ~*ptr; ptr++; }
}

static void drawHLine(ggi_visual_t vis, int x1, int x2, int y, pixel color)
{
	int	width;
	pixel	*ptr;
	
	CLIPHLINE(vis, x1, x2, y, width);
	ptr	= LFBADDR(vis, x1, y);
	width	= x2 - x1 + 1;
	switch (GGI2D_GC_OP(vis)) {
	
#define	DRAWHLINE(OP,EXPR) \
	case GGI_ ## OP: while (width--) { *ptr = EXPR; ptr++; } break;

	case GGI_NOOP: 
	case GGI_INVERT: 
	case GGI_SET: memset(ptr, color, width); break;
	case GGI_SET_INVERTED: memset(ptr, ~color, width); break;
	DRAWHLINE(AND, *ptr & color )
	DRAWHLINE(NAND, ~(*ptr & color) )
	DRAWHLINE(AND_REVERSE, ~*ptr & color )
	DRAWHLINE(AND_INVERTED, *ptr & ~color )
	DRAWHLINE(OR, *ptr | color )
	DRAWHLINE(NOR, ~(*ptr | color) )
	DRAWHLINE(OR_REVERSE, ~*ptr | color )
	DRAWHLINE(OR_INVERTED, *ptr | ~color )
	DRAWHLINE(XOR, *ptr ^ color )
	DRAWHLINE(EQUIV, ~(*ptr ^ color) )
	DRAWHLINE(ADD, *ptr + color )
	DRAWHLINE(SUB, *ptr - color )
	}
}

static void drawHLine_a(ggi_visual_t vis, int x1, int x2, int y, pixel col)
{
	int	width;
	pixel	*ptr, color;
	ggi_alpha alpha;
	
	CLIPHLINE(vis, x1, x2, y, width);\
	ptr	= LFBADDR(vis, x1, y);
	alpha	= GGI2D_GC_ALPHA(vis);
	width	= x2 - x1 + 1;
	switch (GGI2D_GC_OP(vis)) {
	
#define	DRAWHLINE_a(OP,EXPR) \
	case GGI_ ## OP:\
		while (width--) {\
			color	= MIX(col, ptr, alpha);\
			*ptr	= EXPR; ptr++;\
		} break;

	case GGI_NOOP: 
	case GGI_INVERT:
	DRAWHLINE_a(SET, color )
	DRAWHLINE_a(SET_INVERTED, ~color )
	DRAWHLINE_a(AND, *ptr & color )
	DRAWHLINE_a(NAND, ~(*ptr & color) )
	DRAWHLINE_a(AND_REVERSE, ~*ptr & color )
	DRAWHLINE_a(AND_INVERTED, *ptr & ~color )
	DRAWHLINE_a(OR, *ptr | color )
	DRAWHLINE_a(NOR, ~(*ptr | color) )
	DRAWHLINE_a(OR_REVERSE, ~*ptr | color )
	DRAWHLINE_a(OR_INVERTED, *ptr | ~color )
	DRAWHLINE_a(XOR, *ptr ^ color )
	DRAWHLINE_a(EQUIV, ~(*ptr ^ color) )
	DRAWHLINE_a(ADD, *ptr + color )
	DRAWHLINE_a(SUB, *ptr - color )
	}
}

void	GGINAME(scanline)(ggi_visual_t vis, int x1, int x2, int y)
{
	drawHLine(vis, x1, x2, y, GGI2D_GC_FILLCOL(vis));
}

void	GGINAME(scanline_a)(ggi_visual_t vis, int x1, int x2, int y)
{
	drawHLine_a(vis, x1, x2, y, GGI2D_GC_FILLCOL(vis));
}

void	GGINAME(hLine)(ggi_visual_t vis, int x1, int x2, int y)
{
	drawHLine(vis, x1, x2, y, GGI2D_GC_DRAWCOL(vis));
}

void	GGINAME(hLine_a)(ggi_visual_t vis, int x1, int x2, int y)
{
	drawHLine_a(vis, x1, x2, y, GGI2D_GC_DRAWCOL(vis));
}

/*----------------------------------------------------------------------*/
