/* 
 *  pixel drawing
 *
 *  Copyright (C) 1998 Thomas Tanner. See CREDITS for details.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "lib.h"
#include "util.h"

/*----------------------------------------------------------------------*/

#define	PUTPIXEL(OP,EXPR) \
		case GGI_ ## OP: *ptr = EXPR; break;

static void	putPixel_NOOP(ggi_visual_t vis, int x, int y, ggi_pixel color)
{
}

static void	putPixel_INVERT(ggi_visual_t vis, int x, int y, ggi_pixel color)
{
	if (!CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		*ptr	= ~*ptr;
	}
}

static void	putPixel_SET(ggi_visual_t vis, int x, int y, ggi_pixel color)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		*ptr	= color;
	}
}

static void	putPixel_XOR(ggi_visual_t vis, int x, int y, ggi_pixel color)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		*ptr	^= color;
	}
}

static void	putPixel_GENERIC(ggi_visual_t vis, int x, int y, ggi_pixel color)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		switch (GGI2D_GC_OP(vis)) {
		case GGI_NOOP:
		case GGI_INVERT:
		case GGI_SET:
		PUTPIXEL(SET_INVERTED, ~color )
		PUTPIXEL(AND, *ptr & color )
		PUTPIXEL(NAND, ~(*ptr & color) )
		PUTPIXEL(AND_REVERSE, ~*ptr & color )
		PUTPIXEL(AND_INVERTED, *ptr & ~color )
		PUTPIXEL(OR, *ptr | color )
		PUTPIXEL(NOR, ~(*ptr | color) )
		PUTPIXEL(OR_REVERSE, ~*ptr | color )
		PUTPIXEL(OR_INVERTED, *ptr | ~color )
		case GGI_XOR:
		PUTPIXEL(EQUIV, ~(*ptr ^ color) )
		PUTPIXEL(ADD, *ptr + color )
		PUTPIXEL(SUB, *ptr - color )
		}
	}
}

static void	putPixel_a_SET(ggi_visual_t vis, int x, int y, ggi_pixel col)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		ggi_alpha alpha = GGI2D_GC_ALPHA(vis);
		*ptr	= MIX(col,ptr,alpha);
	}
}

static void	putPixel_a_XOR(ggi_visual_t vis, int x, int y, ggi_pixel col)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		ggi_alpha alpha = GGI2D_GC_ALPHA(vis);
		*ptr	^= MIX(col,ptr,alpha);
	}
}

static void	putPixel_a_GENERIC(ggi_visual_t vis, int x, int y, ggi_pixel col)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		ggi_alpha alpha = GGI2D_GC_ALPHA(vis);
		pixel	color = MIX(GGI2D_GC_DRAWCOL(vis),ptr,alpha);
		switch (GGI2D_GC_OP(vis)) {
		case GGI_NOOP:
		case GGI_INVERT:
		case GGI_SET:
		PUTPIXEL(SET_INVERTED, ~color )
		PUTPIXEL(AND, *ptr & color )
		PUTPIXEL(NAND, ~(*ptr & color) )
		PUTPIXEL(AND_REVERSE, ~*ptr & color )
		PUTPIXEL(AND_INVERTED, *ptr & ~color )
		PUTPIXEL(OR, *ptr | color )
		PUTPIXEL(NOR, ~(*ptr | color) )
		PUTPIXEL(OR_REVERSE, ~*ptr | color )
		PUTPIXEL(OR_INVERTED, *ptr | ~color )
		case GGI_XOR:
		PUTPIXEL(EQUIV, ~(*ptr ^ color) )
		PUTPIXEL(ADD, *ptr + color )
		PUTPIXEL(SUB, *ptr - color )
		}
	}
}

void	(*GGINAME(putPixel)[16])(ggi_visual_t vis, int x, int y, ggi_pixel color) =
{
    putPixel_NOOP,
    putPixel_INVERT,
    putPixel_SET,
    putPixel_GENERIC,
    putPixel_GENERIC,
    putPixel_GENERIC,
    putPixel_GENERIC,
    putPixel_GENERIC,
    putPixel_GENERIC,
    putPixel_GENERIC,
    putPixel_GENERIC,
    putPixel_GENERIC,
    putPixel_XOR,
    putPixel_GENERIC,
    putPixel_GENERIC,
    putPixel_GENERIC
};

void	(*GGINAME(putPixel_a)[16])(ggi_visual_t vis, int x, int y, ggi_pixel color) =
{
    putPixel_NOOP,
    putPixel_INVERT,
    putPixel_a_SET,
    putPixel_a_GENERIC,
    putPixel_a_GENERIC,
    putPixel_a_GENERIC,
    putPixel_a_GENERIC,
    putPixel_a_GENERIC,
    putPixel_a_GENERIC,
    putPixel_a_GENERIC,
    putPixel_a_GENERIC,
    putPixel_a_GENERIC,
    putPixel_a_XOR,
    putPixel_a_GENERIC,
    putPixel_a_GENERIC,
    putPixel_a_GENERIC
};

/*----------------------------------------------------------------------*/

static void	drawPixel_NOOP(ggi_visual_t vis, int x, int y)
{
}

static void	drawPixel_INVERT(ggi_visual_t vis, int x, int y)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		*ptr	= ~*ptr;
	}
}

static void	drawPixel_SET(ggi_visual_t vis, int x, int y)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		*ptr  = GGI2D_GC_DRAWCOL(vis);
	}
}

static void	drawPixel_XOR(ggi_visual_t vis, int x, int y)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		*ptr	^= GGI2D_GC_DRAWCOL(vis);
	}
}

static void	drawPixel_GENERIC(ggi_visual_t vis, int x, int y)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		pixel	color = GGI2D_GC_DRAWCOL(vis);
		switch (GGI2D_GC_OP(vis)) {
		case GGI_NOOP:
		case GGI_INVERT:
		case GGI_SET:
		PUTPIXEL(SET_INVERTED, ~color )
		PUTPIXEL(AND, *ptr & color )
		PUTPIXEL(NAND, ~(*ptr & color) )
		PUTPIXEL(AND_REVERSE, ~*ptr & color )
		PUTPIXEL(AND_INVERTED, *ptr & ~color )
		PUTPIXEL(OR, *ptr | color )
		PUTPIXEL(NOR, ~(*ptr | color) )
		PUTPIXEL(OR_REVERSE, ~*ptr | color )
		PUTPIXEL(OR_INVERTED, *ptr | ~color )
		case GGI_XOR:
		PUTPIXEL(EQUIV, ~(*ptr ^ color) )
		PUTPIXEL(ADD, *ptr + color )
		PUTPIXEL(SUB, *ptr - color )
		}
	}
}

static void	drawPixel_a_SET(ggi_visual_t vis, int x, int y)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		ggi_alpha alpha = GGI2D_GC_ALPHA(vis);
		*ptr	= MIX(GGI2D_GC_DRAWCOL(vis),ptr,alpha);
	}
}

static void	drawPixel_a_XOR(ggi_visual_t vis, int x, int y)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		ggi_alpha alpha = GGI2D_GC_ALPHA(vis);
		*ptr	^= MIX(GGI2D_GC_DRAWCOL(vis),ptr,alpha);
	}
}

static void	drawPixel_a_GENERIC(ggi_visual_t vis, int x, int y)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		ggi_alpha alpha = GGI2D_GC_ALPHA(vis);
		pixel	color = MIX(GGI2D_GC_DRAWCOL(vis),ptr,alpha);
		switch (GGI2D_GC_OP(vis)) {
		case GGI_NOOP:
		case GGI_INVERT:
		case GGI_SET:
		PUTPIXEL(SET_INVERTED, ~color )
		PUTPIXEL(AND, *ptr & color )
		PUTPIXEL(NAND, ~(*ptr & color) )
		PUTPIXEL(AND_REVERSE, ~*ptr & color )
		PUTPIXEL(AND_INVERTED, *ptr & ~color )
		PUTPIXEL(OR, *ptr | color )
		PUTPIXEL(NOR, ~(*ptr | color) )
		PUTPIXEL(OR_REVERSE, ~*ptr | color )
		PUTPIXEL(OR_INVERTED, *ptr | ~color )
		case GGI_XOR:
		PUTPIXEL(EQUIV, ~(*ptr ^ color) )
		PUTPIXEL(ADD, *ptr + color )
		PUTPIXEL(SUB, *ptr - color )
		}
	}
}

void	(*GGINAME(drawPixel)[16])(ggi_visual_t vis, int x, int y) =
{
    drawPixel_NOOP,
    drawPixel_INVERT,
    drawPixel_SET,
    drawPixel_GENERIC,
    drawPixel_GENERIC,
    drawPixel_GENERIC,
    drawPixel_GENERIC,
    drawPixel_GENERIC,
    drawPixel_GENERIC,
    drawPixel_GENERIC,
    drawPixel_GENERIC,
    drawPixel_GENERIC,
    drawPixel_XOR,
    drawPixel_GENERIC,
    drawPixel_GENERIC,
    drawPixel_GENERIC
};

void	(*GGINAME(drawPixel_a)[16])(ggi_visual_t vis, int x, int y) =
{
    drawPixel_NOOP,
    drawPixel_INVERT,
    drawPixel_a_SET,
    drawPixel_a_GENERIC,
    drawPixel_a_GENERIC,
    drawPixel_a_GENERIC,
    drawPixel_a_GENERIC,
    drawPixel_a_GENERIC,
    drawPixel_a_GENERIC,
    drawPixel_a_GENERIC,
    drawPixel_a_GENERIC,
    drawPixel_a_GENERIC,
    drawPixel_a_XOR,
    drawPixel_a_GENERIC,
    drawPixel_a_GENERIC,
    drawPixel_a_GENERIC
};

/*----------------------------------------------------------------------*/

static void	drawPixelA_NOOP(ggi_visual_t vis, int x, int y, ggi_alpha alpha)
{
}

static void	drawPixelA_INVERT(ggi_visual_t vis, int x, int y, ggi_alpha alpha)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		*ptr	= ~*ptr;
	}
}

static void	drawPixelA_SET(ggi_visual_t vis, int x, int y, ggi_alpha alpha)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		*ptr	= MIX(GGI2D_GC_DRAWCOL(vis),ptr,alpha);
	}
}

static void	drawPixelA_XOR(ggi_visual_t vis, int x, int y, ggi_alpha alpha)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		*ptr	^= MIX(GGI2D_GC_DRAWCOL(vis),ptr,alpha);
	}
}

static void	drawPixelA_GENERIC(ggi_visual_t vis, int x, int y, ggi_alpha alpha)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		pixel	color = MIX(GGI2D_GC_DRAWCOL(vis),ptr,alpha);
		switch (GGI2D_GC_OP(vis)) {
		case GGI_NOOP:
		case GGI_INVERT:
		case GGI_SET:
		PUTPIXEL(SET_INVERTED, ~color )
		PUTPIXEL(AND, *ptr & color )
		PUTPIXEL(NAND, ~(*ptr & color) )
		PUTPIXEL(AND_REVERSE, ~*ptr & color )
		PUTPIXEL(AND_INVERTED, *ptr & ~color )
		PUTPIXEL(OR, *ptr | color )
		PUTPIXEL(NOR, ~(*ptr | color) )
		PUTPIXEL(OR_REVERSE, ~*ptr | color )
		PUTPIXEL(OR_INVERTED, *ptr | ~color )
		case GGI_XOR:
		PUTPIXEL(EQUIV, ~(*ptr ^ color) )
		PUTPIXEL(ADD, *ptr + color )
		PUTPIXEL(SUB, *ptr - color )
		}
	}
}

static void	drawPixelA_a_SET(ggi_visual_t vis, int x, int y, ggi_alpha alpha2)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		ggi_alpha alpha = (GGI2D_GC_ALPHA(vis) + alpha2) / 2;
		*ptr	= MIX(GGI2D_GC_DRAWCOL(vis),ptr,alpha);
	}
}

static void	drawPixelA_a_XOR(ggi_visual_t vis, int x, int y, ggi_alpha alpha2)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		ggi_alpha alpha = (GGI2D_GC_ALPHA(vis) + alpha2) / 2;
		*ptr	^= MIX(GGI2D_GC_DRAWCOL(vis),ptr,alpha);
	}
}

static void	drawPixelA_a_GENERIC(ggi_visual_t vis, int x, int y, ggi_alpha alpha2)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		ggi_alpha alpha = (GGI2D_GC_ALPHA(vis) + alpha2) / 2;
		pixel	color = MIX(GGI2D_GC_DRAWCOL(vis),ptr,alpha);
		switch (GGI2D_GC_OP(vis)) {
		case GGI_NOOP:
		case GGI_INVERT:
		case GGI_SET:
		PUTPIXEL(SET_INVERTED, ~color )
		PUTPIXEL(AND, *ptr & color )
		PUTPIXEL(NAND, ~(*ptr & color) )
		PUTPIXEL(AND_REVERSE, ~*ptr & color )
		PUTPIXEL(AND_INVERTED, *ptr & ~color )
		PUTPIXEL(OR, *ptr | color )
		PUTPIXEL(NOR, ~(*ptr | color) )
		PUTPIXEL(OR_REVERSE, ~*ptr | color )
		PUTPIXEL(OR_INVERTED, *ptr | ~color )
		case GGI_XOR:
		PUTPIXEL(EQUIV, ~(*ptr ^ color) )
		PUTPIXEL(ADD, *ptr + color )
		PUTPIXEL(SUB, *ptr - color )
		}
	}
}

void	(*GGINAME(drawPixelA)[16])(ggi_visual_t vis, int x, int y, ggi_alpha alpha) =
{
    drawPixelA_NOOP,
    drawPixelA_INVERT,
    drawPixelA_SET,
    drawPixelA_GENERIC,
    drawPixelA_GENERIC,
    drawPixelA_GENERIC,
    drawPixelA_GENERIC,
    drawPixelA_GENERIC,
    drawPixelA_GENERIC,
    drawPixelA_GENERIC,
    drawPixelA_GENERIC,
    drawPixelA_GENERIC,
    drawPixelA_XOR,
    drawPixelA_GENERIC,
    drawPixelA_GENERIC,
    drawPixelA_GENERIC
};

void	(*GGINAME(drawPixelA_a)[16])(ggi_visual_t vis, int x, int y, ggi_alpha alpha) =
{
    drawPixelA_NOOP,
    drawPixelA_INVERT,
    drawPixelA_a_SET,
    drawPixelA_a_GENERIC,
    drawPixelA_a_GENERIC,
    drawPixelA_a_GENERIC,
    drawPixelA_a_GENERIC,
    drawPixelA_a_GENERIC,
    drawPixelA_a_GENERIC,
    drawPixelA_a_GENERIC,
    drawPixelA_a_GENERIC,
    drawPixelA_a_GENERIC,
    drawPixelA_a_XOR,
    drawPixelA_a_GENERIC,
    drawPixelA_a_GENERIC,
    drawPixelA_a_GENERIC
};

/*----------------------------------------------------------------------*/

void	GGINAME(fillPixelA)(ggi_visual_t vis, int x, int y, ggi_alpha alpha)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		pixel	color = MIX(GGI2D_GC_FILLCOL(vis),ptr,alpha);
		switch (GGI2D_GC_OP(vis)) {
		case GGI_NOOP:	break;
		PUTPIXEL(INVERT, ~*ptr )
		PUTPIXEL(SET, color )
		PUTPIXEL(SET_INVERTED, ~color )
		PUTPIXEL(AND, *ptr & color )
		PUTPIXEL(NAND, ~(*ptr & color) )
		PUTPIXEL(AND_REVERSE, ~*ptr & color )
		PUTPIXEL(AND_INVERTED, *ptr & ~color )
		PUTPIXEL(OR, *ptr | color )
		PUTPIXEL(NOR, ~(*ptr | color) )
		PUTPIXEL(OR_REVERSE, ~*ptr | color )
		PUTPIXEL(OR_INVERTED, *ptr | ~color )
		PUTPIXEL(XOR, *ptr ^ color )
		PUTPIXEL(EQUIV, ~(*ptr ^ color) )
		PUTPIXEL(ADD, *ptr + color )
		PUTPIXEL(SUB, *ptr - color )
		}
	}
}

void	GGINAME(fillPixelA_a)(ggi_visual_t vis, int x, int y, ggi_alpha alpha2)
{
	if (CLIP(vis, x, y)) {
		pixel	*ptr = LFBADDR(vis, x, y);
		ggi_alpha alpha = (GGI2D_GC_ALPHA(vis) + alpha2) / 2;
		pixel	color = MIX(GGI2D_GC_FILLCOL(vis),ptr,alpha);
		switch (GGI2D_GC_OP(vis)) {
		case GGI_NOOP:	break;
		PUTPIXEL(INVERT, ~*ptr )
		PUTPIXEL(SET, color )
		PUTPIXEL(SET_INVERTED, ~color )
		PUTPIXEL(AND, *ptr & color )
		PUTPIXEL(NAND, ~(*ptr & color) )
		PUTPIXEL(AND_REVERSE, ~*ptr & color )
		PUTPIXEL(AND_INVERTED, *ptr & ~color )
		PUTPIXEL(OR, *ptr | color )
		PUTPIXEL(NOR, ~(*ptr | color) )
		PUTPIXEL(OR_REVERSE, ~*ptr | color )
		PUTPIXEL(OR_INVERTED, *ptr | ~color )
		PUTPIXEL(XOR, *ptr ^ color )
		PUTPIXEL(EQUIV, ~(*ptr ^ color) )
		PUTPIXEL(ADD, *ptr + color )
		PUTPIXEL(SUB, *ptr - color )
		}
	}
}

/*----------------------------------------------------------------------*/

