/* 
 *  Internal declarations for loadable libraries
 *
 *  Copyright (C) 1998 by Thomas Tanner. See CREDITS for details.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
 
#ifndef _GGILIB_H_
#define _GGILIB_H_

#define _BUILDING_LIBGGI2D

#include "tmp.h" 
#include <ggi/ggi2d.h>

#include "debug.h"
#include <ggi/threads.h>

extern int ggi2d_offset;

typedef	void	(*updatefunc)(ggi_visual_t vis, int change);

/* mode changes: functions should be updated */
#define	GGI2D_UPD_ARCMODE	1	
#define	GGI2D_UPD_POLYMODE	2
#define	GGI2D_UPD_LINEMODE	4
#define	GGI2D_UPD_APPENDMODE	8	
#define	GGI2D_UPD_ANTIALIAS	16
#define	GGI2D_UPD_TEXTUREMODE	32
#define	GGI2D_UPD_ALPHAMODE	64
#define	GGI2D_UPD_OPERATOR	128

/* value changes: accelerator should be updated */
#define	GGI2D_UPD_LINE		256
#define	GGI2D_UPD_CLIP		512	
#define	GGI2D_UPD_DRAWCOL	1024	
#define	GGI2D_UPD_FILLCOL	2048	
#define	GGI2D_UPD_TEXTURE	4096
#define	GGI2D_UPD_ALPHA		8192

#define	GGI2D_UPD_ALL		(2*GGI2D_UPD_ALPHA-1)

#define ISVALUE(change)	(change > GGI_UPD_OPERATOR)

typedef struct {
	ggi_arcmode	arcmode;
	ggi_polymode	polymode;
	uint		*dash;
	uint		dashsize;
	ggi_bool	append;
	ggi_bool	antialias;
	ggi_pixel	drawcol;
	ggi_pixel	fillcol;
	ggi_bool	textured;
	ggi_visual_t	texture;
	ggi_coord	texref;
	ggi_operator	operator;
	ggi_alpha	alpha;
	int		accel_changes;	
} ggi2d_gc;

typedef struct {
	updatefunc update;
	int	change_mask;
} ggi2d_module; 

typedef struct {
/*------- Drawing functions (33) -------------*/
void	(*putPixel)(ggi_visual_t vis, int x, int y, ggi_pixel color);
void	(*drawPixel)(ggi_visual_t vis, int x, int y);
void	(*drawPixelA)(ggi_visual_t vis, int x, int y, ggi_alpha alpha);
void	(*fillPixelA)(ggi_visual_t vis, int x, int y, ggi_alpha alpha);
void	(*drawPixels)(ggi_visual_t vis, ggi_coord coords[], uint count);
void	(*scanline)(ggi_visual_t vis, int x1, int x2, int y);
void	(*scanlines)(ggi_visual_t vis, ggi_scanline scanlines[], uint count);
void	(*hLine)(ggi_visual_t vis, int x1, int x2, int y);
void	(*vLine)(ggi_visual_t vis, int x, int y1, int y2);
void	(*drawRect)(ggi_visual_t vis, int x1, int y1, int x2, int y2);
void	(*fillRect)(ggi_visual_t vis, int x1, int y1, int x2, int y2);
void	(*line)(ggi_visual_t vis, int x1, int y1, int x2, int y2);
void	(*linef)(ggi_visual_t vis, ggi_float x1, ggi_float y1,
			    ggi_float x2, ggi_float y2);
void	(*drawLines)(ggi_visual_t vis, ggi_line lines[], uint count);
void	(*drawCircle)(ggi_visual_t vis, int x, int y, uint r);
void	(*fillCircle)(ggi_visual_t vis, int x, int y, uint r);
void	(*drawEllipse)(ggi_visual_t vis, int x, int y, uint rx, uint ry);
void	(*fillEllipse)(ggi_visual_t vis, int x, int y, uint rx, uint ry);
void	(*drawArc)(ggi_visual_t vis, int x, int y, uint rx, uint ry,
	    	   ggi_float startangle, ggi_float endangle, ggi_bool close);
void	(*fillArc)(ggi_visual_t vis, int x, int y, uint rx, uint ry,
		   ggi_float startangle, ggi_float endangle);
void	(*bezier)(ggi_visual_t vis, int x1, int y1, int x2, int y2,
		     int x3, int y3, int x4, int y4);
void	(*trapezoid)(ggi_visual_t vis, 
		    int xl1, int xr1, int y1,
		    int xl2, int xr2, int y2);
void	(*triangle)(ggi_visual_t vis, 
		    int x1, int y1, int x2, int y2, int x3, int y3);
void	(*drawPoly)(ggi_visual_t vis, ggi_coord coords[], uint count);
void	(*fillPoly)(ggi_visual_t vis, ggi_coord coords[], uint count);
void	(*fillPolys)(ggi_visual_t vis, ggi_coord coords[], 
		uint counts[], uint count);
void	(*blit)(
	    ggi_visual_t dest, int dx, int dy,
	    ggi_visual_t src, int sx, int sy, 
	    int width, int height);	
void	(*blitOp)(
	    ggi_visual_t dest, int dx, int dy, 
	    ggi_visual_t src1, ggi_visual_t src2, int sx, int sy, 
	    int width, int height, ggi_operator op);
void	(*stretchBlit)(
	    ggi_visual_t dest, int dx, int dy, int dwidth, int dheight, 
	    ggi_visual_t src, int sx, int sy, int swidth, int sheight);
void	(*stretchBlitOp)(
	    ggi_visual_t dest, int dx, int dy, int dwidth, int dheight,
	    ggi_visual_t src1, ggi_visual_t src2,
	    int sx, int sy, int swidth, int sheight,
	    ggi_operator op);
} ggi2d_funcs;

#define	MAX_MODULES	4

typedef struct {
	ggi_uint	version;		
	int		modules;	/* number of loaded modules */
	int		change_mask;	/* general mask */
	ggi2d_module	module[MAX_MODULES];
	ggi2d_gc	gc;
	ggi2d_funcs	func;
} ggi2d_data;

#define GGI2D_DATA(vis)		((ggi2d_data *)((uchar *)vis+ggi2d_offset))
#define GGI2D_FUNC(vis)		(&(GGI2D_DATA(vis)->func))

#define GGI2D_GC(vis)		(GGI2D_DATA(vis)->gc)

#define GGI2D_GC_ARCMODE(vis) 	(GGI2D_GC(vis).arcmode)
#define GGI2D_GC_POLYMODE(vis)	(GGI2D_GC(vis).polymode)
#define GGI2D_GC_DASH(vis) 	(GGI2D_GC(vis).dash)
#define GGI2D_GC_DASHSIZE(vis)	(GGI2D_GC(vis).dashsize)
#define GGI2D_GC_APPEND(vis)	(GGI2D_GC(vis).append)
#define GGI2D_GC_ANTIALIAS(vis)	(GGI2D_GC(vis).antialias)
#define GGI2D_GC_CLIP_X1(vis)	(vis->gc->cliptl.x)
#define GGI2D_GC_CLIP_Y1(vis)	(vis->gc->cliptl.y)
#define GGI2D_GC_CLIP_X2(vis)	(vis->gc->clipbr.x)
#define GGI2D_GC_CLIP_Y2(vis)	(vis->gc->clipbr.y)
#define GGI2D_GC_DRAWCOL(vis)	(vis->gc->fg_color)
#define GGI2D_GC_FILLCOL(vis)	(vis->gc->bg_color)
#define GGI2D_GC_TEXTURED(vis)	(GGI2D_GC(vis).textured)
#define GGI2D_GC_TEXTURE(vis) 	(GGI2D_GC(vis).texture)
#define GGI2D_GC_TEXREF(vis)	(GGI2D_GC(vis).texref)
#define GGI2D_GC_ALPHA(vis)	(GGI2D_GC(vis).alpha)
#define GGI2D_GC_OP(vis) 	(GGI2D_GC(vis).operator)
#define GGI2D_ACCELCHANGES(vis)	(GGI2D_GC(vis).accel_changes)

#define GGI_FB_LINEAR(vis)	(vis->info.fb.linear)
#define GGI_FB_WIDTH(vis)	(vis->info.fb.width)
#define GGI_FB_BPL(vis)		(vis->info.fb.width)
#define GGI_FB_HEIGHT(vis)	(vis->info.fb.height)

void	GGI2DregisterUpdateFunc(ggi_visual_t vis, updatefunc f, int mask);

#endif /* _GGILIB_H_ */
