/* 
 *  GGI Utility library
 *
 *  Copyright (C) 1998 Uwe Maurer - uwe_maurer@t-online.de
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <ggi/libggi.h>
#include <ggi/libgbm.h>
#include <fcntl.h>
#include <malloc.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <internal.h>

int gguLoadImage(ggi_visual_t vis,char *file,ggi_color *pal)
{
	int ft,fd,rc;
	GBM gbm;
	GBMRGB gbmrgb[256];
	ggi_color col[256];
	int i;
	unsigned char *buffer,*temp;	
	int row;

	DPRINT("gguLoadImage: %s\n",file);	
	
	gbm_init();
	if (gbm_guess_filetype(file,&ft)!=GBM_ERR_OK) 
	{
		DPRINT("unknown filetype!\n");
		gbm_deinit();
		return -1;
	}
	if ((fd=gbm_io_open(file,O_RDONLY))==-1)
	{
		DPRINT("can't open file!\n");
		gbm_deinit();
		return -1;
	}
	if ((rc=gbm_read_header(file,fd,ft,&gbm,""))!=GBM_ERR_OK)
	{
		DPRINT("can't read header!\n");
		gbm_io_close(fd);
		gbm_deinit();
		return -1;
	}
	if (GGU(vis)->adjust==GGU_FALSE)
	{
		if (gbm.w>GGU_VIRT(vis).x || gbm.h>GGU_VIRT(vis).y ||
			gguBPP(GGU_GT(vis))!=gbm.bpp)
		{
			DPRINT("adjust==GGU_FALSE \n");
			gbm_io_close(fd);
			gbm_deinit();
			return -1;
		}
	}
	if (gbm.bpp<=8)
	{
		if ((rc=gbm_read_palette(fd,ft,&gbm,gbmrgb))!=GBM_ERR_OK)
		{
			DPRINT("can't read palette!\n");
			gbm_io_close(fd);
			gbm_deinit();
			return -1;
		}
	}

	row=(((gbm.w*gbm.bpp+7)/8+3)/4)*4;
	temp=buffer=malloc(row*gbm.h);
	if (buffer==NULL)
	{
		DPRINT("not enough memory!\n");
		gbm_io_close(fd);
		gbm_deinit();
		return -1;
	}
	rc=gbm_read_data(fd,ft,&gbm,buffer);
	gbm_io_close(fd);
	gbm_deinit();
	if (rc!=GBM_ERR_OK)
	{
		DPRINT("%s\n",gbm_err(rc));	
		free(buffer);
		return -1;
	}
	DPRINT("%ix%ix%i\n",gbm.w,gbm.h,gbm.bpp);

	if (GGU(vis)->adjust)
	{
		rc=ggiSetGraphMode(vis,gbm.w,gbm.h,gbm.w,gbm.h,
			gguGraphtype(gbm.bpp));
		if (rc==-1)
		{
			DPRINT("can't set grapmode!\n");
			free(buffer);
			return -1;	
		}
	}
	for(i=gbm.h-1;i;i--)
	{
		ggiPutHLine(vis,0,i,gbm.w,temp);
		temp+=row;
	}	
	
	free(buffer);

	if (gbm.bpp<=8)
	{
		if (pal==NULL) pal=col;

		for (i=0;i<(1<<gbm.bpp);i++)
		{
			pal[i].r=gbmrgb[i].r<<(GGI_COLOR_PRECISION-8);
			pal[i].g=gbmrgb[i].g<<(GGI_COLOR_PRECISION-8);
			pal[i].b=gbmrgb[i].b<<(GGI_COLOR_PRECISION-8);
		}
		ggiSetPaletteVec(vis,0,(1<<gbm.bpp),pal); 
	}

	ggiFlush(vis);
	return 0;
}

int gguSaveImage(ggi_visual_t vis,char *file,int x,int y,int w,int h,
		ggi_color *pal)
{
	int fd,ft,i;
	GBM gbm;
	GBMRGB gbmrgb[256];
	GBM_ERR rc;
	GBMFT gbmft;
	ggi_color col[256];
	unsigned char *buffer,*temp;
	int bytes_per_line;
	int oldbpp;
	ggi_visual_t convert;

	DPRINT("gguSaveImage: %s\n",file);	
	ggiFlush(vis);	
	gbm_init();
	if (gbm_guess_filetype(file,&ft)!=GBM_ERR_OK) 
	{
		DPRINT("unknown filetype!\n");
		gbm_deinit();
		return -1;
	}
	gbm_query_filetype(ft,&gbmft);
	gbm.w=w;
	gbm.h=h;
	oldbpp=gbm.bpp=gguBPP(GGU_GT(vis));

	switch(gbm.bpp)	
	{
		case 1: if (!(gbmft.flags & GBM_FT_W1)) 
			{
				gbm_deinit();
				return -1;
			} break;
		case 4: if (!(gbmft.flags & GBM_FT_W4))
			{
				gbm_deinit();
				return -1;
			} break;
		case 8: if (!(gbmft.flags & GBM_FT_W8))
			{
				gbm_deinit();
				return -1;
			} break;
		case 15:
		case 16:
		case 32:
		case 24: if (!(gbmft.flags & GBM_FT_W24))
			{
				gbm_deinit();
				return -1;
			} 
			gbm.bpp=24;
			break;
	} 
	if (oldbpp!=gbm.bpp)
	{

		convert=ggiOpen("display-memory",NULL);
		if (convert==NULL) 
		{
			gbm_deinit();
			return -1;
		}
		if (ggiSetGraphMode(convert,w,h,w,h,
			gguGraphtype(gbm.bpp))!=0)
		{
			gbm_deinit();
			return -1;
		}
		if (ggiCrossBlit(vis,x,y,w,h,convert,0,0,w,h))
		{
			gbm_deinit();
			return -1;
		}
		rc=gguSaveImage(convert,file,0,0,w,h,pal);
		ggiClose(convert);
		return rc;
	}
	bytes_per_line=(gbm.w*gbm.bpp+7)/8;
	temp=buffer=malloc(h*bytes_per_line);
	if (buffer==NULL)
	{
		DPRINT("not enough memory!\n");
		gbm_deinit();
		return -1;
	}

	if (gbm.bpp<=8)
	{
		if (!pal) 
		{
			if (ggiGetPaletteVec(vis,0,1<<gbm.bpp,col)==0) 
				pal=col;
		}

		if (pal) 
		{	
			for (i=0;i<(1<<gbm.bpp);i++)
			{
				gbmrgb[i].r=pal[i].r>>(GGI_COLOR_PRECISION-8);
				gbmrgb[i].g=pal[i].g>>(GGI_COLOR_PRECISION-8);
				gbmrgb[i].b=pal[i].b>>(GGI_COLOR_PRECISION-8);
			}
		}
		else
		{
			DPRINT("palette == NULL\n");
			free(buffer);
			gbm_deinit();
			return -1;
		}
	}

	if ((fd=gbm_io_create(file,O_RDWR))==-1)
	{
		DPRINT("can't open file!\n");
		free(buffer);
		gbm_deinit();
		return -1;
	}

	for(i=gbm.h-1;i;i--)
	{
		ggiGetHLine(vis,x,y+i,gbm.w,temp);
		temp+=bytes_per_line;
	}
	rc=gbm_write(file,fd,ft,&gbm,gbmrgb,buffer,"");
	free(buffer);
	gbm_io_close(fd);	
	gbm_deinit();
	if (rc!=GBM_ERR_OK)
	{
		DPRINT("%s\n",gbm_err(rc));	
		return -1;
	}
	return 0;
}

#define MAXSCREENSHOTS	10000

char *gguScreenShotFile="screen%.4i.";
char *gguIndexFormat="gif";
char *gguRGBFormat="tga";

int gguScreenShot(ggi_visual_t vis)
{

	struct stat buf;
	char filename[256];
	char buffer[256];
	int i;
	int w,h,bpp;

	DPRINT("gguScreenShot:\n");

	for (i=0;i<MAXSCREENSHOTS;i++)
	{	
		sprintf(filename,gguScreenShotFile,i);
		strcpy(buffer,filename);
		strcat(buffer,gguIndexFormat);
		if (stat(buffer,&buf)==0) continue; 
		strcpy(buffer,filename);
		strcat(buffer,gguRGBFormat);
		if (stat(buffer,&buf)==0) continue; 
		
		w=GGU_WIDTH(vis);
		h=GGU_HEIGHT(vis);
		bpp=gguBPP(GGU_GT(vis));
		strcat(filename, (bpp>8) ? gguRGBFormat : gguIndexFormat);
		DPRINT("saving %s\n",filename);
		return gguSaveImage(vis,filename,0,0,w,h,NULL);
	}
	return -1;
}


int gguAdjustVisual(ggi_visual_t vis,int adjust)
{
	GGU(vis)->adjust=(adjust!=GGU_FALSE);
	return 0;
}

int gguImageInfo(char *file,int *w,int *h,ggi_graphtype *gt)
{
	int ft,fd,rc;
	GBM gbm;
	gbm_init();
	if (gbm_guess_filetype(file,&ft)!=GBM_ERR_OK) 
	{
		DPRINT("unknown filetype!\n");
		gbm_deinit();
		return -1;
	}
	if ((fd=gbm_io_open(file,O_RDONLY))==-1)
	{
		DPRINT("can't open file!\n");
		gbm_deinit();
		return -1;
	}
	if ((rc=gbm_read_header(file,fd,ft,&gbm,""))!=GBM_ERR_OK)
	{
		DPRINT("can't read header!\n");
		gbm_io_close(fd);
		gbm_deinit();
		return -1;
	}
	gbm_io_close(fd);
	gbm_deinit();
	if (w) *w=gbm.w;
	if (h) *h=gbm.h;
	if (gt) *gt=gguGraphtype(gbm.bpp);
	return 0;

}
