/* -----------------------------------------------------------------------------
**	boot display code
** -----------------------------------------------------------------------------
**
**	Copyright (C) 1995-2000 by Steffen Seeger [seeger@physik.tu-chemnitz.de]
**	Copyright (C) 1997 by Geert Uytterhoeven
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
*/

#define	MAINTAINER	"Geert Uytterhoeven, [Geert.Uytterhoeven@cs.kuleuven.ac.be]"

#include <linux/config.h>
#include <linux/version.h>
#include <asm/io.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/ascii.h>
#include <linux/console.h>
#include <linux/kdev_t.h>
#include <ggi/kgi.h>

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)

	#define	CONFIG_VT_CONSOLE y
	extern register_console(void (*proc)(const char *));
#endif


/* -----------------------------------------------------------------------------
**	generic text16-framebuffer-only display driver
** -----------------------------------------------------------------------------
**
**	text16 framebuffer operations
**
**	These are the framebuffer operations for a framebuffer with text16
**	layout in physical memory. As we know ushort is 2 byte and uchar is
**	1 byte, most pointer arithmetic is done with hardcoded sizeof(ushort).
** -----------------------------------------------------------------------------
*/

#ifdef DEBUG
	/*	NOTE: the ASSERT lines should give you a ring if the framebuffer
	**	operations are used with inconsistent parameters. However, we
	**	rely on the OS to catch NULL pointer references!
	*/

	#define	FBSIZE	(fb->size.x*fb->size.y)
#endif

static void text16_move_mem(kgi_pixmap *fb, ggi_uint dst, ggi_uint src, ggi_uint cnt)
{
	register ushort *pbuf = (void *) fb->pbuf;

	ASSERT((src + cnt) <= FBSIZE);
	ASSERT((dst + cnt) <=  FBSIZE);

	memmove(pbuf + dst, pbuf + src, cnt + cnt);
}

static void text16_set_mem(kgi_pixmap *fb, ggi_uint src, ggi_pixel val, ggi_uint cnt)
{
	ushort *pbuf = ((ushort *) fb->pbuf) + src;

	ASSERT((src + cnt) <= FBSIZE);

	while (cnt--) {

		*(pbuf++) = val;
	}
}

static void text16_invert_mem(kgi_pixmap *fb, ggi_uint start, ggi_uint end)
{
	register ushort *pix = ((ushort *) fb->pbuf) + start;

	ASSERT((start < FBSIZE) && (end < FBSIZE));
	ASSERT(start <= end);

	while (start++ <= end) {

		*pix =     ((*pix) & 0x88FF) 
			| (((*pix)<<4) & 0x7000) | (((*pix>>4)) & 0x0700);
		pix++;
	}
}

static void text16_update_mem(kgi_pixmap *fb, ggi_uint pos, ggi_pixel val)
{
	ASSERT(pos < FBSIZE);

	*(((ushort*) fb->pbuf) + pos) = val;
}

static void text16_flush_mem(kgi_pixmap *fb)
{
}

static void text16_copyto_mem(kgi_pixmap *fb, void *buf, ggi_uint start, ggi_uint cnt)
{
	ASSERT((start + cnt) <= FBSIZE);

	memcpy(((ushort *) fb->pbuf) + start, buf, cnt + cnt);
}

static void text16_copyfrom_mem(kgi_pixmap *fb, void *buf, ggi_uint start, ggi_uint cnt)
{
	ASSERT((start + cnt) <= FBSIZE);

	memcpy(buf, ((ushort *) fb->pbuf) + start, cnt + cnt);
}


    /*
     *  This is ugly and terrible slow...
     */

struct consw *conswitchp;

extern u_int global_nr_cols;	/* defined in arch/m68k/console/fbcon.c */
extern u_int global_nr_rows;
extern u_int global_fontwidth;
extern u_int global_fontheight;
extern u_int global_can_do_color;

static void do_fbcon_putc(ggi_pixel val, ggi_uint pos)
{
    u_int y = pos/global_nr_cols;
    u_int x = pos%global_nr_cols;
    unsigned char attr = val >> 8;
    unsigned char c = val & 0xff;
    conswitchp->con_putc(c, y, x, attr);
}

static void do_fbcon_putcs(u_short *data, ggi_uint startpos, ggi_uint cnt)
{
    u_int y = startpos/global_nr_cols;
    u_int x = startpos%global_nr_cols;
    while (cnt--) {
	ggi_pixel val = *data++;
	unsigned char attr = val >> 8;
	unsigned char c = val & 0xff;
	conswitchp->con_putc(c, y, x, attr);
	x++;
	if (x == global_nr_cols) {
	    x = 0;
	    y++;
	}
    }
}

static void text16_move_io(kgi_pixmap *fb, ggi_uint dst, ggi_uint src, ggi_uint cnt)
{
	register ushort *pbuf = (void *) fb->pbuf;

	ASSERT((src + cnt) <= FBSIZE);
	ASSERT((dst + cnt) <=  FBSIZE);

	memmove(pbuf + dst, pbuf + src, cnt + cnt);
	do_fbcon_putcs(pbuf + dst, dst, cnt);
}

static void text16_set_io(kgi_pixmap *fb, ggi_uint src, ggi_pixel val, ggi_uint cnt)
{
	ushort *pbuf = ((ushort *) fb->pbuf) + src;

	ASSERT((src + cnt) <= FBSIZE);

	while (cnt--) {

		*(pbuf++) = val;
	}
	do_fbcon_putcs(((ushort *) fb->pbuf) + src, src, cnt);
}

static void text16_invert_io(kgi_pixmap *fb, ggi_uint start, ggi_uint end)
{
	register ushort *pix = ((ushort *) fb->pbuf) + start;

	ASSERT((start < FBSIZE) && (end < FBSIZE));
	ASSERT(start <= end);

	while (start++ <= end) {

		*pix =     ((*pix) & 0x88FF) 
			| (((*pix)<<4) & 0x7000) | (((*pix>>4)) & 0x0700);
		pix++;
	}
	do_fbcon_putcs(((ushort *) fb->pbuf) + start, start, end-start+1);
}

static void text16_update_io(kgi_pixmap *fb, ggi_uint pos, ggi_pixel val)
{
	ASSERT(pos < FBSIZE);

	*(((ushort*) fb->pbuf) + pos) = val;
	do_fbcon_putc(val, pos);
}

static void text16_flush_io(kgi_pixmap *fb)
{
}

static void text16_copyto_io(kgi_pixmap *fb, void *buf, ggi_uint start, ggi_uint cnt)
{
	ASSERT((start + cnt) <= FBSIZE);

	memcpy(((ushort *) fb->pbuf) + start, buf, cnt + cnt);
	do_fbcon_putcs(buf, start, cnt);
}

static void text16_copyfrom_io(kgi_pixmap *fb, void *buf, ggi_uint start, ggi_uint cnt)
{
	ASSERT((start + cnt) <= FBSIZE);

	memcpy(buf, ((ushort *) fb->pbuf) + start, cnt + cnt);
	do_fbcon_putcs(buf, start, cnt);
}


/*	color mode attribute to pixel value mapping. We care of the foreground
**	and background colors. The rest is just best effort.
*/
static ggi_pixel text16_atop_color(kgi_pixmap *fb, ggi_attr attr)
{
	register ggi_pixel
	result = (attr & ATTR_REVERSE)
			? (((attr & 0x0007) << 8) | ((attr & 0x0700) << 4))
			: (((attr & 0x0007) << 12) | (attr & 0x0700));

	if (attr & ATTR_BLINK) {

		result |= 0x8000;
	}

	if (attr & ATTR_BOLD) {

		result |= 0x0800;
	}

	return result;
}

static ggi_attr text16_ptoa_color(kgi_pixmap *fb, ggi_pixel val)
{
	return	  ((val & 0x8000) ? ATTR_BLINK : 0)
		| ((val & 0x0800) ? ATTR_BOLD  : 0)
		| ((val & 0x7000) >> 12) | (val & 0x0F00);
}

/*	monochrome attribute mapping. We care about the attributes possible
**	and ignore colors.
*/
static ggi_pixel text16_atop_mono(kgi_pixmap *fb, ggi_attr attr)
{
	register ggi_pixel result;

	if (attr & ATTR_UNDERLINE) {

		result = 0x0100;	/* MDA can't do reverse underline :-( */

	} else {

		result = (attr & ATTR_REVERSE) ? 0x7000 : 0x0700;
	}

	if (attr & ATTR_BLINK) {

		result |= 0x8000;
	}

	if (attr & ATTR_BOLD) {

		result |= 0x0800;
	}

	return result;
}

static ggi_attr text16_ptoa_mono(struct kgi_pixmap *fb, ggi_pixel val)
{
	register ggi_attr result;

	if ((val & 0x7700) == 0x0100) {

		result = ATTR_UNDERLINE;

	} else {

		result = (val & 0x7000) ? ATTR_REVERSE : ATTR_NORMAL;
	}

	if (val & 0x8000) {

		result |= ATTR_BLINK;
	}

	if (val & 0x0800) {

		result |= ATTR_BOLD;
	}

	return result;
}

/*	font postion to pixel mapping and vice versa. This cares about 256 
**	character fonts only (at least for now).
*/
static ggi_pixel text16_ftop(kgi_pixmap *fb, ggi_uint pos)
{
	return pos & 0x00FF;
}

static ggi_uint text16_ptof(kgi_pixmap *fb, ggi_pixel val)
{
	return val & 0x00FF;
}

/*	We have color and mono operations, both for I/O and normal memory; so
**	there are four framebuffer operations to provide.
*/

#define	TEXT16_FB_OPS(mode, mem)\
	text16_##mode##_##mem = {\
		text16_move_##mem,\
		text16_set_##mem,\
		text16_invert_##mem,\
		text16_update_##mem,\
		text16_flush_##mem,\
		text16_copyfrom_##mem,\
		text16_copyto_##mem,\
		\
		text16_atop_##mode,\
		text16_ptoa_##mode,\
		\
		text16_ftop,\
		text16_ptof\
	}

kgi_fb_ops TEXT16_FB_OPS(mono,  io);
kgi_fb_ops TEXT16_FB_OPS(mono,  mem);
kgi_fb_ops TEXT16_FB_OPS(color, io);
kgi_fb_ops TEXT16_FB_OPS(color, mem);

/* -----------------------------------------------------------------------------
**	display methods
** -----------------------------------------------------------------------------
*/

/*	check_mode() simply checks if the requested mode matches the one
**	needed for the framebuffer layout we were told or proposes it.
*/
static int text16_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
			   enum kgi_tm_cmd cmd)
{
	#define MATCH(x)	(mode->request.x == dpy->mode.request.x)

	if (   MATCH(graphtype) && MATCH(frames)
	    && MATCH(visible.x) && MATCH(visible.y)
	    && MATCH(text.x)    && MATCH(text.y)
	    && (mode->request.virt.x || mode->request.virt.y)) {

		if (mode->request.virt.x) {

			if (! MATCH(virt.x)) {

				return -E(CHIPSET, UNKNOWN);
			}
		}

		if (mode->request.virt.y) {

			if (! MATCH(virt.y)) {

				return -E(CHIPSET, UNKNOWN);
			}
		}

		if (cmd == CMD_PROPOSE) {

			memcpy(mode, &(dpy->mode), sizeof(*mode));
			return EOK;

		} else {

			return memcmp(mode, &(dpy->mode), sizeof(*mode))
				? EOK : -E(CHIPSET, MODE_CHECK);
		}

	} else {

		return -E(CHIPSET, MODE_CHECK);
	}

	#undef MATCH
}

/*	reset() simply clears the framebuffer.
*/
static void text16_reset(struct kgi_display *dpy)
{
	memset_io(dpy->fb.pbuf, 0, dpy->mode.fb_size); 
}

/*	as we have framebuffer only, we must use software gadgets. We store
**	the current position and saved character for the cursor and pointer,
**	as well as the xor-mask to apply in the driver specific fields
**	of dpy->mode.
*/
#define	MSKCUR		(dpy->mode.clock)
#define	MSKPTR		(dpy->mode.monitor)
#define	OLDCUR		(dpy->mode.chipset)
#define	OLDPTR		(dpy->mode.ramdac)

#define	TEXTX		(dpy->mode.request.text.x)
#define	TEXTY		(dpy->mode.request.text.y)
#define	VIRTUALX	(dpy->fb.size.x)
#define	VISIBLEX	(dpy->mode.request.visible.x)
#define	VISIBLEY	(dpy->mode.request.visible.y)

static void text16_show_pointer(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	ASSERT((x < VISIBLEX) && (y < VISIBLEY));

	conswitchp->con_cursor(x / TEXTX, y / TEXTY, CM_DRAW);
}

static void text16_hide_pointer(struct kgi_display *dpy)
{
	conswitchp->con_cursor(0, 0, CM_ERASE);
}

#define	text16_undo_pointer	text16_hide_pointer

#define text16_show_cursor	text16_show_pointer
#define text16_hide_cursor	text16_hide_pointer
#define	text16_undo_cursor	text16_hide_cursor

#undef	TEXTX
#undef	TEXTY
#undef	VIRTUALX
#undef	VISIBLEX
#undef	VISIBLEY

/* -----------------------------------------------------------------------------
**	kernel interface
** -----------------------------------------------------------------------------
**	text16_malloc_display() allocates and initializes a struct kgi_display
**	that handles the assigned area. It is given the following parameters:
**
**	fb		*virtual* address of the text16 framebuffer
**	x,y		size of the framebuffer in pixels
**	textx,texty	size of one character cell in dots
**	color		flag whether to use monochrome or color attiributes
** -----------------------------------------------------------------------------
*/
#define	COLOR	1
#define	MONO	0
struct kgi_display *text16_malloc_display(size_t fb, ggi_uint x, ggi_uint y,
	ggi_uint textx, ggi_uint texty, int color)
{
	struct kgi_display *dpy;

	ASSERT(textx && texty && x && y);

	if (! (dpy = kgi_malloc_small(sizeof(struct kgi_display),
		GFP_KERNEL))) {

		return NULL;
	}

	memset(dpy, 0, sizeof(struct kgi_display));

	/*	initialize generic fields
	*/
	dpy->id = -1;
	dpy->conflict_text = 0xFFFFFFFF;
	dpy->conflict_graph= 0xFFFFFFFF;
	dpy->conflict_setup= 0xFFFFFFFF;

	/* initialize function pointers
	*/
	dpy->reset = text16_reset;
	dpy->check_mode	= text16_check_mode;

	dpy->show_pointer	= text16_show_pointer;
	dpy->hide_pointer	= text16_hide_pointer;
	dpy->undo_pointer	= text16_undo_pointer;

	dpy->show_cursor	= text16_show_cursor;
	dpy->hide_cursor	= text16_hide_cursor;
	dpy->undo_cursor	= text16_undo_cursor;

	/*	initialize mode info
	*/
	dpy->defmode		= &(dpy->mode.request);
	#define	MODE	(dpy->mode.request)
		MODE.graphtype		= GT_TEXT16;
		MODE.frames		= 1;
		MODE.text.x		= textx;
		MODE.text.y		= texty;
		MODE.visible.x		= x*textx;
		MODE.visible.y		= y*texty;
		MODE.virt.x		= MODE.visible.x;
		MODE.virt.y		= MODE.visible.y;
	#undef MODE
	dpy->mode.fb	= fb;
	dpy->mode.fsize = x*y*sizeof(ushort);
	dpy->mode.fb_size = dpy->mode.fsize;
	dpy->mode.mem	= color ? &text16_color_mem : &text16_mono_mem;
	dpy->mode.io	= color ? &text16_color_io  : &text16_mono_io;

	/*	initialize display framebuffer
	*/
	dpy->fb.dpy	= dpy;
	dpy->fb.size.x	= x;
	dpy->fb.size.y	= y;
	dpy->fb.bpp	= MAP_16_BPP;
	dpy->fb.pbuf	= (ggi_pixel *)dpy->mode.fb;
	dpy->fb.ops	= dpy->mode.io;

	/*	initialize driver specific stuff
	*/
	if (color) {

		MSKCUR = 0x6F00;
		MSKPTR = 0x7F00;

	} else {

		MSKCUR = 0x7700;
		MSKPTR = 0x7700;
	}	

	text16_reset(dpy);
	return dpy;
}

void text16_free_display(struct kgi_display *dpy)
{
	ASSERT(dpy);
	ASSERT(!dpy->card);

	kgi_free_small(dpy);
}

#ifdef CONFIG_VT_CONSOLE
/* -----------------------------------------------------------------------------
**	boot printk for m68k/Amiga
** -----------------------------------------------------------------------------
*/

extern void ami_serial_console_write(const char *s, unsigned count);
extern void ami_serial_console_wait_key(void);

static void boot_console_write(const char *s, unsigned count)
{
	ami_serial_console_write(s, count);
}

static void boot_console_wait_key(void)
{
	ami_serial_console_wait_key();
}

static int boot_console_device(void)
{
	return MKDEV(TTY_MAJOR, 1);
}

static struct console boot_console={
	boot_console_printk,
	NULL,
	boot_console_wait_key,
	boot_console_device,
	NULL
};
#endif /* CONFIG_VT_CONSOLE */

/* -----------------------------------------------------------------------------
**	kernel interface
** -----------------------------------------------------------------------------
*/

/*	no_scroll() was introduced to disable hardware scrolling for some
**	braille lines. We assume a MDA-like framebuffer anyhow, so this does
**	simply nothing and has become obsolete. I think...
*/	
void no_scroll(char *str, int *ints)
{
}

/*	kgi_console_init() is called to setup a console_print() method
**	to display printk()-messages before the KGI manager is running.
**	There are no kernel services (PCI scan, kgi_malloc_*() working
**	when this is called!
*/
ulong kgi_console_init(ulong kmem_start, ulong kmem_end)
{
#ifdef __mc68000__
	const char *display_desc = "????";

#ifdef CONFIG_VT_CONSOLE
	#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)
		register_console(boot_console_printk);
	#else
		register_console(&boot_console);
	#endif
#endif /* CONFIG_VT_CONSOLE */

	kmem_start = conswitchp->con_startup(kmem_start, &display_desc);
#endif /* __mc68000__ */
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,31)
	kmem_start = tty_init_std_termios(kmem_start, kmem_end);
#endif

	return kmem_start;
}

/*	display_init() has to scan for displays the kernel can support.
**	All kernel services are accessible here, as kernel boot is already
**	complete. Later we may allow to link modules directly with the kernel,
**	for now this simply sets up a dumb framebuffer display for every
**	textmode card installed.
*/

int display_init(void)
{
	int dpys = 0;
	struct kgi_display *dpy;
	size_t dummy;

#warning This should probe all built-in frame buffer devices

	conswitchp->con_init();
	/* dummy frame buffer */
	dummy = (size_t)kmalloc(global_nr_cols*global_nr_rows*sizeof(u_short),
				GFP_KERNEL);

	if (global_can_do_color) {
		printk(KERN_INFO "color display detected.\n");

		dpy = text16_malloc_display(dummy, global_nr_cols,
					    global_nr_rows, global_fontwidth,
					    global_fontheight, COLOR);
		if (!dpy || kgi_register_display(dpy, -1))
			panic("Could not register color display.\n");
		dpys++;
	} else {
		printk(KERN_INFO "monochrome display detected.\n");

		dpy = text16_malloc_display(dummy, global_nr_cols,
					    global_nr_rows, global_fontwidth,
					    global_fontheight, MONO);
		if (!dpy || kgi_register_display(dpy, -1))
			panic("Could not register mono display.\n");
		dpys++;
	}

	printk(KERN_INFO "display_init(): %i displays detected.\n", dpys);

	return dpys;
}
