#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>

#include <ggi/serialmouse.h>

struct smouse_type {
	const char *op;
	const char *name;
	int	type;
	int	cflag;
} types[]={
	{"ms","Microsoft",MOUSE_TYPE_MICROSOFT,
		(CS7|CREAD|CLOCAL|HUPCL)},
	{"mousesys","MouseSystems",MOUSE_TYPE_MOUSESYS,
		(CS8|CSTOPB|CREAD|CLOCAL|HUPCL)},
	{"mmseries","MMSeries",MOUSE_TYPE_LOGITECH,
		(CS8|PARENB|PARODD|CREAD|CLOCAL|HUPCL)},
	{"mouseman","Logitech MouseMan",MOUSE_TYPE_LOGITECH,
		(CS7|CREAD|CLOCAL|HUPCL)},
	{"logi","Logitech",MOUSE_TYPE_LOGITECH,
		(CS8|CSTOPB|CREAD|CLOCAL|HUPCL)},
	{NULL,NULL,0,0}
};

int main(int argc,char **argv)
{
	int i,j,fd,err,pid,type,display,mindex=0;
	int registered=0;
	struct termios term;
	struct serialmouse_attr mattr;

	if (argc==1) {
		fprintf(stderr,"Usage:\n"
		               "%s [-t <type>] [-d <display>] </dev/ttySx> ...\n\n"
			       "Attaches /dev/ttySx to the N_MOUSE "
			       "tty line discipline.\n\n"
			       "Available mouse types:\n\n",argv[0]);
		for (i=0;types[i].op!=NULL;i++) {
			fprintf(stderr,"%-8s %-40s\n",
					types[i].op,
					types[i].name);
		}
		fprintf(stderr,"\nDefault mouse type: %s (%s)\n",
			       types[0].op,types[0].name);

		return 1;
	}

	display=0;
	for (i=1;i<argc;i++) {
		if (strcmp(argv[i],"-t")==0) {
			i++;
			for (j=0;types[j].op!=NULL;j++) 
				if (strcmp(types[j].op,argv[i])==0)
					break;
			
			if (types[j].op==NULL) {
				fprintf(stderr,"%s: Unknown mouse type '%s'\n",
					       argv[0],argv[i]);
				return 2;
			}
			mindex=j;
			continue;
		} else if (strcmp(argv[i],"-d")==0) {
			i++;
			display=atoi(argv[i]);
			continue;
		}
		fd=open(argv[i],O_RDWR);
		if (fd==-1) {
			perror(argv[i]);
			return 2;
		}
		type=N_MOUSE;
		mattr.type=types[mindex].type;
		mattr.display=display;

		/* Set up line conditions */
		err=tcgetattr(fd,&term);
		if (err) {perror("tcgetattr"); return 2;}
		cfmakeraw(&term);
		term.c_cflag = B1200 | types[mindex].cflag;
		err=tcsetattr(fd,TCSANOW,&term);
		if (err) {perror("tcsetattr"); return 2;}

		/* Set up mouse line discipline */
		err=ioctl(fd,TIOCSETD,&type);
		if (err<0) {perror("ioctl(TIOCSETD)"); return 2;}

		/* Set up mouse driver */
		err=ioctl(fd,MOUSESATTR,&mattr);
		if (err) {
			fprintf(stderr,"Could not register mouse"
				       " (%s:%s) on display %d\n",
				       argv[i],types[mindex].op,display);
			close(fd);
			continue;
		}
		fprintf(stderr,"%s Mouse (%s) on display %d\n",
				types[mindex].name,argv[i],display);
		display++;
		registered++;
	}

	if (registered==0) {
		fprintf(stderr,"No mice were registered\n");
		return 0;
	}

	if (pid=fork()) {
	  fprintf(stdout,"%d\n",pid);
	  return 0;
	} else {
	  close(0);
	  close(1);
	  close(2);
	  while (1) pause();
	}
}
